/***************************************************************************
                      delayline.cpp  -  description
                            -------------------
    begin                : Fri Jan 18 2002
    copyright            : (C) 2002 by Fabian Wenzel
    email                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#include "delayline.h"
#include <boost/checked_delete.hpp>
#include <algorithm>

////////////////////////////// Public Functions /////////////////////////////

BEGIN_CONFIG(delayLine,vistaLab::module)
    CONFIGINPUTPIN (m_inputImage  ,__("Input image"))
    CONFIGOUTPUTPIN(m_outputImage ,__("Output image"))
    CONFIGPARAMETER(m_bufferLength,__("Bufferlength"))
END_CONFIG()

/***************************************************************************/
/*!
 *  Constructor sets up default bufferLength
 *
 ***************************************************************************/
delayLine::delayLine() :
vistaLab::module(PROCESSING,__("Miscellaneous"),__("Delay Line"),
__("Delays another input channel"),
"Fabian Wenzel","wenzel@tu-harburg.de"),
m_bufferLength(1)
{
}


/***************************************************************************/
/*!
 *  The buffers is set to the correct size and the input- and
 *  output - pointers are initialized
 *
 ***************************************************************************/
bool delayLine::init()
{
    m_buffer.resize(m_bufferLength+1);

    // The input buffer pointer points to the head of the buffer
    m_currentImage  = m_buffer.begin();

    return true;
}


/***************************************************************************/
/*!
 *  The buffer gets cleared
 *
 ***************************************************************************/
void delayLine::done()
{
    for_each(m_buffer.begin(),m_buffer.end(),
             boost::checked_deleter< vigra::BRGBImage >());

    fill(m_buffer.begin(),m_buffer.end(), (vigra::BRGBImage *) 0);
}


/***************************************************************************/
/*!
 *  As the delay line might get called at any time (even before the source
 *  channel has fed the line) the first thing to do is to wait for the own
 *  sync which gets signaled in delayLine::addImageToLine() by another
 *  thread. Afterwards the m_currentImage pointer is just set to the current
 *  output image, and the current output image is set to zero (as it does not
 *  have to be deleted here anymore - it does not belong to the "buffer"
 *  anymore.
 *
 ***************************************************************************/
bool delayLine::process()
{
    *m_currentImage = new vigra::BRGBImage(*m_inputImage);

    ++m_currentImage;

    if(m_currentImage == m_buffer.end()){
        m_currentImage = m_buffer.begin();
    }

    if (*m_currentImage){
        m_outputImage = *m_currentImage;
        *m_currentImage = 0;
    } else{
        m_outputImage =
            new vigra::BRGBImage(m_inputImage->size());
    }

    return true;
}
