/***************************************************************************
               inputstreamimagesequence.cpp  -  description
                            -------------------
    begin                : Mon Feb 18 2001
    copyright            : (C) 2001 by Fabian Wenzel
    email                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include "inputstreamimagesequence.h"

#include <wx/image.h>

using namespace std;

//////////////////////////////// Public Functions ////////////////////////////////

BEGIN_CONFIG(inputStreamImageSequence,vistaLab::module)
    CONFIGOUTPUTPIN(m_outputImage      , __("Output Image")       )
    CONFIGPARAMETER(m_fileNames        , __("Images")             )
    CONFIGPARAMETER(m_loop             , __("Loop sequence")      )
    CONFIGPARAMETER(m_loopBackAndForth , __("Loop back and forth"))
    CONFIGVAL      (checkFileNameSize)
END_CONFIG()

/***************************************************************************/
/*!
 *  Initializes the list of provided colorspaced with GRAY and RGB. Default
 *  behavior is set not to loop.
 *
 ***************************************************************************/
inputStreamImageSequence::inputStreamImageSequence() :
vistaLab::module(INPUT,__("Input"),__("Image Sequence"),
__("Loads a sequence of Images from file"),
"Fabian Wenzel","wenzel@tu-harburg.de"),
m_loop            (false),
m_loopBackAndForth(false)
{
}


/***************************************************************************/
/*!
 *  Sets the current fileName - pointer to the first file and initializes
 *  flags for processing
 *
 ***************************************************************************/
bool inputStreamImageSequence::init()
{
    m_currentFile = m_fileNames.begin();

    m_endOfStream =             false;
    m_currentlyLoopingForward = true;

    return true;
}


/***************************************************************************/
/*!
 *  Does nothing
 *
 ***************************************************************************/
void inputStreamImageSequence::done()
{
}


/***************************************************************************/
/*!
 *   If new images are there, load one (similar to
 *   inputStreamImage::fetchImage() )
 *   The only thing which has to be done afterwards is to adjust the
 *   pointer to the fileName which will be used next, but the flags for
 *   looping have to be taken into account, too
 *
 ***************************************************************************/
bool inputStreamImageSequence::process()
{
    if(m_endOfStream){
        vistaLab::setError(_("No more images"));
        return false;
    }

    wxImage tmp(m_currentFile->c_str());


    m_outputImage = new vigra::BRGBImage(tmp.GetWidth(),tmp.GetHeight(),
                        (vigra::RGBValue< unsigned char > *) tmp.GetData());

    // If there is only one image, iterating can be complicated
    // So check this first
    if ( m_fileNames.size() == 1 ){
        m_endOfStream = (m_loop == false);
        return true;
    }

    if ( m_currentlyLoopingForward ){

        // We are looping forward
        ++m_currentFile;
        if ( m_currentFile == m_fileNames.end() ){
            if( ! m_loop ){
                // We don't loop
                m_endOfStream  = true;
                return true;
            } else {
                if ( m_loopBackAndForth ){
                    // We loop back and forth, so go back twice
                    m_currentlyLoopingForward = false;
                    --m_currentFile;
                    --m_currentFile;
                } else
                    // We start over
                    m_currentFile = m_fileNames.begin();
            }
        }


    } else {

        // We are looping backward (i.e. loop=true and
        // loopBackAndForth = true)
        --m_currentFile;
        if ( m_currentFile == m_fileNames.begin() )
            m_currentlyLoopingForward = true;

    }

    return true;
}


///////////////////////////// Private Functions /////////////////////////////



/***************************************************************************/
/*!
 *  checks whether there is at least one entry in the fileName list
 *
 ***************************************************************************/
const char *inputStreamImageSequence::checkFileNameSize()
{
    if ( m_fileNames.size() > 0 )
        return NULL;
    else
        return _("There must be at least one file in the image sequence!");
}
