/***************************************************************************
                         vlglwindow.cpp  -  description
                            -------------------
    begin                : Mon Mar 18 2002
    copyright            : (C) 2002 by Fabian Wenzel
    email                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include "vlglwindow.h"

#include <GL/glu.h>
#include <GL/gl.h>

#include <algorithm>
#include <locale.h>
#include <fstream>

extern "C" {
#include "trackball.h"
#include "gl2ps.h"
}

using namespace std;
using namespace vistaLab;

#define SPINUPDATE 20

enum{
    GLTIMERID = 10,
    GLSNAPSHOT,
    GLSNAPSHOTBACKGROUND,
    GLSNAPSHOTTRANSPARENCY,
    GLSNAPSHOTOPTIMIZATION,
    GLSNAPSHOTSUPPRESSTEXT,
    GLWRITEQUATS,
    GLREADQUATS
};


//////////////////////////////// Event table ////////////////////////////////


BEGIN_EVENT_TABLE(GLWindow, wxGLCanvas)
EVT_PAINT           (                         GLWindow::onPaint              )
EVT_MOUSE_EVENTS    (                         GLWindow::onMouse              )
EVT_SIZE            (                         GLWindow::onSize               )
EVT_ERASE_BACKGROUND(                         GLWindow::onEraseBackground    )
EVT_TIMER           ( GLTIMERID             , GLWindow::onTimer              )
EVT_MENU            ( GLSNAPSHOT            , GLWindow::onSnapshot           )
EVT_MENU            ( GLSNAPSHOTBACKGROUND  , GLWindow::onToggleBackground   )
EVT_MENU            ( GLSNAPSHOTTRANSPARENCY, GLWindow::onToggleTransparency )
EVT_MENU            ( GLSNAPSHOTOPTIMIZATION, GLWindow::onToggleOptimization )
EVT_MENU            ( GLSNAPSHOTSUPPRESSTEXT, GLWindow::onToggleSuppressText )
EVT_MENU            ( GLWRITEQUATS          , GLWindow::onWriteQuats         )
EVT_MENU            ( GLREADQUATS           , GLWindow::onReadQuats          )
END_EVENT_TABLE()


///////////////////////////// Public Functions //////////////////////////////


/***************************************************************************/
/*!
 *  Constructor sets up default zoom
 *
 ***************************************************************************/
GLWindow::GLWindow(wxWindow *parent, int id,
    const wxPoint &position, const wxSize &size) :
wxGLCanvas(parent,id,position,size,wxNO_BORDER),
m_zoom          (50.0),
m_initialized   (false),
m_drawBackground(true),
m_transparency  (false),
m_optimization  (false),
m_suppressText  (true),
m_kill          (false),
m_rightClickPos (-1,-1),
m_near          (10.0),
m_far           (1000.0),
m_wheelRot      (0),
m_backTrans     (0)
{
    m_trans[0]      = 0.0;
    m_trans[1]      = 0.0;
    m_trans[2]      = 0.0;
    m_mouseSpeed[0] = 0.0;
    m_mouseSpeed[1] = 0.0;
    m_spinTimer = new wxTimer(this,GLTIMERID);
}


/***************************************************************************/
/*!
 *  Destructor deletes object data
 *
 ***************************************************************************/
GLWindow::~GLWindow()
{
}


//////////////////////////////// Callbacks ////////////////////////////////


/***************************************************************************/
/*!
 *  Does nothing (copied from penguin sample)
 *
 ***************************************************************************/
void GLWindow::onEraseBackground(wxEraseEvent &event)
{
}


/***************************************************************************/
/*!
 *  Draws the histogram by stepping through the common openGL - process
 *  (clearing, setting up the coordinate system and drawing). The image
 *  is locked as it might be accessed by setImage in the meantime
 *
 ***************************************************************************/
void GLWindow::onPaint(wxPaintEvent &event)
{
    wxCriticalSectionLocker locker(m_openGLLock);

    // Has to be called as there may be other GL windows which would eat calls
    SetCurrent();

    if(!m_initialized)
        initializeGL();

    /* must always be here */
    wxPaintDC dc(this);

    // Clears the image
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

    // Initializes the coordinate system
    placeCamera();

    // Draw hook
    draw();

    // Done
    SwapBuffers();
}


void GLWindow::draw()
{
}

/***************************************************************************/
/*!
 *  Captures mouse events - they either rotate the modelview by using the
 *  trackball - functions (if left mousebutton is pressed),
 *  zoom back and forth (if right mousebutton is pressed) or reset the view
 *  if there has been a doubleclick
 *
 ***************************************************************************/
void GLWindow::onMouse(wxMouseEvent &event)
{
    m_mouseEvent = event;
}


/***************************************************************************/
/*!
 *  Adjusts the OpenGL - viewport to the new size - this is taken
 *  from the penguin example in wxWidgets
 *
 ***************************************************************************/
void GLWindow::onSize(wxSizeEvent &event)
{
    // this is also necessary to update the context on some platforms
    wxGLCanvas::OnSize(event);

    // set GL viewport (not called by wxGLCanvas::onSize on all platforms...)
    int w, h;
    GetClientSize(&w, &h);
#ifndef __WXMOTIF__
    if (GetContext())
#endif
    {
        SetCurrent();
        glViewport(0, 0, (GLint) w, (GLint) h);
    }
}



////////////////////////////// Private Functions ////////////////////////////


/***************************************************************************/
/*!
 *  Initializes the projection view based on the current zoom - factor
 *
 ***************************************************************************/
void GLWindow::initializeGL()
{
    trackball(m_trackballRot, 0.0, 0.0, 0.0, 0.0);
    glClearColor(0.0,0.0,0.0,0.0);
    m_initialized = true;
    m_spinTimer->Start(SPINUPDATE);
}


/***************************************************************************/
/*!
 *  Initializes the model view based on the current rotation (using the
 *  trackball - functions)
 *
 ***************************************************************************/
void GLWindow::placeCamera()
{
    // Set camera view and clipping planes
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    initializeCamera();

    // Translation given by mouse movement;
    glTranslatef( m_trans[0], m_trans[1], m_trans[2] );

    // Shift model back so that camera is outside of it
    GLfloat m[4][4];
    build_rotmatrix( m,m_trackballRot );
    glMultMatrixf( &m[0][0] );

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void GLWindow::setPlanes(float newNear, float newFar)
{
    m_near      = newNear;
    m_far       = newFar;
    m_backTrans = 2*m_near;
}

void GLWindow::setObjectSize(float size)
{
    m_near      = size;
    m_far       = 3 * size;
    m_backTrans = 2*m_near;
}

void GLWindow::done()
{
    show(false);
    if(m_spinTimer->IsRunning()){
        m_kill=true;
        m_spinTimer->Stop();
    } else {
        GetParent()->Destroy();
    }
}

void GLWindow::onTimer(wxTimerEvent   &event)
{
    static int counter=0;

    if(m_kill){
        GetParent()->Destroy();
        return;
    }
    bool needRefresh = false;

    m_wheelRot += m_mouseEvent.GetWheelRotation();
    int wheeldelta = m_mouseEvent.GetWheelDelta();
    if(wheeldelta == 0)
        wheeldelta = 120;

    if(abs(m_wheelRot) >= wheeldelta){
        m_mouseEvent.m_wheelRotation=0;
        needRefresh = true;
        if(m_wheelRot < 0){
            m_zoom+=0.1*m_zoom;
            m_zoom=min(m_zoom,180.0);
            m_wheelRot += wheeldelta;
        } else {
            m_zoom-=0.1*m_zoom;
            m_zoom=max(m_zoom,0.5);
            m_wheelRot -= wheeldelta;
        }
    }

    if(m_mouseEvent.RightDown()){
        m_rightClickPos=m_mouseEvent.GetPosition();
    }

    if(m_mouseEvent.RightUp() && m_rightClickPos == m_mouseEvent.GetPosition())
    {
        m_rightClickPos = wxPoint(-1,-1);
        wxMenu contextmenu(GetTitle());

        contextmenu.Append(GLWRITEQUATS,_("Write Position to File"),
                           _("Writes Trackball Rotation to File"));
        contextmenu.Append(GLREADQUATS,_("Read Position from File"),
                               _("Reads Trackball Rotation from File"));
        contextmenu.AppendSeparator();
        contextmenu.Append(GLSNAPSHOTBACKGROUND,_("Draw Background"),
            _("Draws Background"),wxITEM_CHECK);
        contextmenu.Append(GLSNAPSHOTTRANSPARENCY,_("Enable Transparency"),
            _("Enables Transparency"),wxITEM_CHECK);
        contextmenu.Append(GLSNAPSHOTOPTIMIZATION,_("Enable Optimization"),
            _("Enables Optimization"),wxITEM_CHECK);
        contextmenu.Append(GLSNAPSHOTSUPPRESSTEXT,_("Suppress Text Output"),
            _("Suppresses any Text Output"),wxITEM_CHECK);
        contextmenu.AppendSeparator();
        contextmenu.Append(GLSNAPSHOT,_("Take Snapshot"),
            _("Saves the current image to file"));
        contextmenu.Check(GLSNAPSHOTBACKGROUND,m_drawBackground);
        contextmenu.Check(GLSNAPSHOTTRANSPARENCY,m_transparency);
        contextmenu.Check(GLSNAPSHOTOPTIMIZATION,m_optimization);
        contextmenu.Check(GLSNAPSHOTSUPPRESSTEXT,m_suppressText);
        PopupMenu(&contextmenu,m_mouseEvent.GetX(),m_mouseEvent.GetY());

    }

    if(m_mouseEvent.Dragging()){
        needRefresh = true;
        if(m_mouseEvent.LeftIsDown()){
            if(m_mouseEvent.ShiftDown()){
                // Translates the modelview
                GLfloat currentmatrix[16];
                glMatrixMode(GL_MODELVIEW);
                glGetFloatv(GL_MODELVIEW_MATRIX,currentmatrix);
                m_trans[0] += (m_mouseEvent.GetX() - m_mousePos.x) * currentmatrix[0];
                m_trans[1] -= (m_mouseEvent.GetY() - m_mousePos.y) * currentmatrix[5];
                m_trans[2] += (m_mouseEvent.GetX() - m_mousePos.x) * currentmatrix[1];
            } else {
                wxSize sz(GetClientSize());
                // Rotates the modelview
                m_mouseSpeed[0] =  2*((m_mouseEvent.GetX() - m_mousePos.x)/(double) sz.x);
                m_mouseSpeed[1] = -2*((m_mouseEvent.GetY() - m_mousePos.y)/(double) sz.y);
            }
        } else if (m_mouseEvent.RightIsDown()){
            // Zooms back and forth
            m_backTrans+=0.5*(m_mousePos.y-m_mouseEvent.GetY());
            m_backTrans=max((float) 0.5,m_backTrans);
        }

    } else if (m_mouseEvent.LeftDClick()){
        // Resets the coordinate system
        m_zoom = 50.0;
        trackball(m_trackballRot,0.0,0.0,0.0,0.0);
        m_trans[0] = m_trans[1] = m_trans[2] = 0.0;
        m_wheelRot = 0;
        m_backTrans = 2*m_near;
        needRefresh = true;
    } else if (m_mouseEvent.LeftIsDown())
    {
        m_mouseSpeed[0] = 0.0;
        m_mouseSpeed[1] = 0.0;
    }

    if(m_mouseSpeed[0] != 0.0 || m_mouseSpeed[1] != 0.0)
    {
        float spinQuat[4];
        trackball(spinQuat,0.0,0.0,m_mouseSpeed[0],m_mouseSpeed[1]);
        add_quats( spinQuat, m_trackballRot, m_trackballRot );
        needRefresh=true;
    }

    if(needRefresh)
        Refresh(FALSE);

    // Tracks the current mouse position
    m_mousePos=m_mouseEvent.GetPosition();
}

void GLWindow::show(bool doShow)
{
    if(m_initialized && doShow)
        m_spinTimer->Start(SPINUPDATE);

    if(!doShow)
        m_spinTimer->Stop();

    m_mouseSpeed[0]=0.0;
    m_mouseSpeed[1]=0.0;
    GetParent()->Show(doShow);
}

void GLWindow::setTitle(const wxString &title)
{
    GetParent()->SetTitle(title);
}

void GLWindow::initializeCamera()
{
    wxSize size = GetClientSize();
    gluPerspective( m_zoom, (size.GetWidth() / (double) size.GetHeight()), m_near,m_far);
    glTranslatef( 0.0, 0.0,-m_backTrans);
}

void GLWindow::onToggleBackground(wxCommandEvent &event)
{
    m_drawBackground=!m_drawBackground;
}

void GLWindow::onToggleTransparency(wxCommandEvent &event)
{
    m_transparency=!m_transparency;
}

void GLWindow::onToggleOptimization(wxCommandEvent &event)
{
    m_optimization=!m_optimization;
}

void GLWindow::onToggleSuppressText(wxCommandEvent &event)
{
    m_suppressText=!m_suppressText;
}

void GLWindow::onSnapshot(wxCommandEvent &event)
{
    m_spinTimer->Stop();
    wxString fileName;

    wxFileDialog fileDialog( NULL, _("Save image as"),
        wxEmptyString, _("output.pdf"),
        _("\
PDF image (*.pdf)|*.pdf|\
EPS image (*.eps)|*.eps|\
Postscript image (*.ps)|*.ps|\
TeX file for gl2ps text (*.tex)|*.tex"),
        wxOVERWRITE_PROMPT|wxHIDE_READONLY|wxSAVE, wxDefaultPosition );

    if ( fileDialog.ShowModal() == wxID_OK )
    {
        fileName = fileDialog.GetPath();
        int wildcard = fileDialog.GetFilterIndex();

        int format = GL2PS_PDF;

        if (!fileName.Matches("*.pdf") && wildcard==0)
            fileName+=".pdf";
        else if(!fileName.Matches("*.eps")  && wildcard==1)
            fileName+=".eps";
        else if(!fileName.Matches("*.ps")   && wildcard==2)
            fileName+=".ps";
        else if(!fileName.Matches("*.tex")  && wildcard==3)
            fileName+=".tex";

        if(fileName.Matches("*.eps"))
            format = GL2PS_EPS;
        else if(fileName.Matches("*.ps"))
            format = GL2PS_PS;
        else if (fileName.Matches("*.tex"))
            format = GL2PS_TEX;


        FILE *fp;
        int state = GL2PS_OVERFLOW, buffSize = 0;

        fp = fopen(fileName.c_str(), "wb");

        if(!fp){
            wxMessageBox("Unable to open file for writing","Error",wxOK);
            m_spinTimer->Start(SPINUPDATE);
            return;
        }

        SetCurrent();

        const char *oldLocale = setlocale(LC_NUMERIC,"C");

        GLint options = GL2PS_USE_CURRENT_VIEWPORT|GL2PS_COMPRESS;
        GLint sorting = GL2PS_SIMPLE_SORT;

        if (m_drawBackground)
            options |= GL2PS_DRAW_BACKGROUND;

        if (!m_transparency){
            options |= GL2PS_NO_BLENDING;
            options |= GL2PS_OCCLUSION_CULL;
        }

        if (m_optimization){
            options |= GL2PS_BEST_ROOT;
            sorting  = GL2PS_BSP_SORT;
        }

        if (m_suppressText){
            options |= GL2PS_NO_TEXT;
        }

        gl2psPointSize(0.0);
        gl2psLineWidth(0.0);

        while(state == GL2PS_OVERFLOW){
            buffSize += 1024*1024;

            gl2psBeginPage(GetParent()->GetTitle().c_str(),
                _("VistaLab Image Processing Framework"), 0, format, sorting,
                options,
                GL_RGBA, 0, NULL, 0, 0, 0,
                buffSize, fp, fileName.c_str());
            // Clears the image
            glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

            // Initializes the coordinate system
            placeCamera();

            draw();
            state = gl2psEndPage();
        }

        setlocale(LC_NUMERIC,oldLocale);

        fclose(fp);

    }
    m_spinTimer->Start(SPINUPDATE);
}

void GLWindow::onReadQuats(wxCommandEvent &event)
{
    wxFileDialog fileDialog( NULL, _("Load position"),
        wxEmptyString, _("position.txt"),
        _("Text file (*.txt)|*.txt"),
        wxHIDE_READONLY|wxOPEN, wxDefaultPosition );

    if ( fileDialog.ShowModal() == wxID_OK )
    {
        const char *oldLocale = setlocale(LC_NUMERIC,"C");
        fstream rotationFile(fileDialog.GetPath().c_str(),ios::in);
        for(int i=0;i<4;++i)
            rotationFile >> m_trackballRot[i];

        for(int i=0;i<3;++i)
            rotationFile >> m_trans[i];

        setlocale(LC_NUMERIC,oldLocale);
    }

}

void GLWindow::onWriteQuats(wxCommandEvent &event)
{
    wxFileDialog fileDialog( NULL, _("Save position as"),
        wxEmptyString, _("position.txt"),
        _("Text file (*.txt)|*.txt"),
        wxOVERWRITE_PROMPT|wxHIDE_READONLY|wxSAVE, wxDefaultPosition );

    if ( fileDialog.ShowModal() == wxID_OK )
    {
        const char *oldLocale = setlocale(LC_NUMERIC,"C");
        fstream rotationFile(fileDialog.GetPath().c_str(),ios::out);
        for(int i=0;i<4;++i)
            rotationFile << m_trackballRot[i] << endl;

        for(int i=0;i<3;++i)
            rotationFile << m_trans[i] << endl;

        setlocale(LC_NUMERIC,oldLocale);
    }
}
