/***************************************************************************
                          vlapp.cpp  -  description
                            -------------------
    begin                : Fri Nov 16 2001
    copyright            : (C) 2001 by Fabian Wenzel
    email                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif


#include "vlapp.h"
#include "vlframe.h"

#include "vlprocessor.h"
#include "vlrepository.h"
#include "vldocument.h"

#include <wx/image.h>
#include <wx/splash.h>
#include <wx/bitmap.h>
#include <wx/config.h>
#include "resources/splash.xpm"

#include <cstdio>

using namespace vistaLab;

////////////////////////////// Public Functions //////////////////////////////

class splashScreen : public wxSplashScreen {
public:
    splashScreen(const wxBitmap &bitmap) :
      wxSplashScreen(bitmap, wxSPLASH_CENTRE_ON_SCREEN|wxSPLASH_TIMEOUT,
          2000,NULL,-1,wxDefaultPosition,wxDefaultSize,
          wxSIMPLE_BORDER|wxSTAY_ON_TOP)
      {
          Connect(-1,wxEVT_CLOSE_WINDOW,(wxObjectEventFunction) (wxCloseEventFunction) &splashScreen::OnClose);
      }

      void OnClose(wxCloseEvent &event)
      {
          wxGetApp().createMainWindow();
          event.Skip();
      }

};

/***************************************************************************/
/*!
 *  Initializer - function: Sets application name, initializes image
 *  handler (when saving snapshots), creates members and creates the window
 *
 ***************************************************************************/
bool app::OnInit()
{
    m_singleRun =false;
    m_showMainWindow = true;
    m_bootProject=wxEmptyString;

    if(!parseOptions()){
        showOptions();
        return false;
    }

    m_locale.Init();
    m_locale.AddCatalogLookupPathPrefix(getLocaleDir());

    if(showSplashScreen())
        new splashScreen(wxBitmap(splash_xpm));

    // initialize image handlers (for saving snapshots)
    wxInitAllImageHandlers();

    SetAppName(VLAPPNAME);

    m_locale.AddCatalog("vistalab");

#ifdef __WIN32__
    //Under MSW there is no default catalog lookup path, so
    //wxstd has not been loaded yet. Do so now.
    m_locale.AddCatalog("wxstd");
#endif

    // Create Members
    m_document   = new document  ();
    m_repository = new repository();
    m_processor  = new processor ();

    if(!showSplashScreen())
        createMainWindow();

    return true;
}


/***************************************************************************/
/*!
 *  Cleans up members
 *
 ***************************************************************************/
int app::OnExit()
{
    // First delete document as the module-functions have to be there
    // when deleting agent data of left images
    delete m_document;

    delete m_processor;

    // Now delete repository
    delete m_repository;
    wxConfig::Get()->Flush();
    return 0;
}


/***************************************************************************/
/*!
 *  Adds a catalog to localization
 *
 ***************************************************************************/
void app::addLocaleCatalog(const wxString &catalogname)
{
    m_locale.AddCatalog(catalogname);
}


/***************************************************************************/
/*!
 *  Returns the locale directory
 *
 ***************************************************************************/
const wxString app::getLocaleDir()
{
#ifndef WIN32
    return LOCALEDIR;
#else
    return wxPathOnly(m_fullProgramName)+wxString("\\locale");
#endif
}

/***************************************************************************/
/*!
 *  Returns the plugin path. This depends on the platform, the
 *  installation directory and the configuration (Debug/Release)
 *
 ***************************************************************************/
const wxString app::getPluginDir ()
{
    wxString pluginDirName;

#ifndef WIN32
    pluginDirName = wxString( PLUGINDIR );
#ifdef __WXDEBUG__
    pluginDirName+="debug/";
#else
    pluginDirName+="release/";
#endif
#else
    pluginDirName = wxPathOnly(m_fullProgramName)+"\\plugins";

#ifdef _MSC_VER
#ifdef __WXDEBUG__
    pluginDirName += "\\debug";
#else
    pluginDirName += "\\release";
#endif
#else
    // Workaround; we are assuming that we are using libtool
    // This is a hack as it only works with autotools, but
    // currently this is the only supported platform
    pluginDirName += "\\..\\bin";
#endif // _MSC_VER

#endif // WIN32
    return pluginDirName;
}


/***************************************************************************/
/*!
 *  The entry - point of the application object
 *
 ***************************************************************************/
void app::createMainWindow()
{
    // Finally, setup Frame and show it
    frame *topwindow = new frame( (wxWindow *) NULL, -1, GetAppName());

    SetTopWindow(topwindow);

    if(showMainWindow())
        topwindow->Show(true);

    // Load command line project
    if(bootProject() != wxEmptyString){
        topwindow->projectOpenFile(bootProject());
        if(singleRun()){
            wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED,PROCESS_START);
            topwindow->AddPendingEvent(event);
        }
    }

}

bool app::showSplashScreen()
{
    bool showSplash=true;
    return (showSplash && (bootProject() == wxEmptyString));
}

bool app::singleRun()
{
    return m_singleRun;
}

const wxString &app::bootProject() const
{
    return m_bootProject;
}


bool app::parseOptions()
{
    bool success=true;
    // Setup the full program name (for path for plugins)
    if( wxIsAbsolutePath(argv[0]) ){
        m_fullProgramName = argv[0];
    } else {
        m_fullProgramName = wxGetCwd() + argv[0];
    }

    for(int i=1;i<argc;++i)
    {
        if(argv[i] == wxString("-singlerun"))
            m_singleRun=true;
        else if(argv[i] == wxString("-hide"))
            m_showMainWindow=false;
        else if(!wxFileExists(argv[i])){
            m_singleRun=false;
            success=false;
            printf(_("File %s does not exist\n\n"),argv[i]);
        } else {
            m_bootProject=argv[i];
        }
    }

    if(!m_singleRun)
        m_showMainWindow=true;

    return success;
}

bool app::showMainWindow() const
{
    return m_showMainWindow;
}

void app::showOptions() const
{
    printf(_("Usage: vistalab [[-hide] [-singlerun] projectname]\n"));
}

/***************************************************************************/
/*!
 *  The entry - point of the application object
 *
 ***************************************************************************/
IMPLEMENT_APP(vistaLab::app)
