/***************************************************************************
                       vlchannelwindow.cpp  -  description
                            -------------------
    begin                : Wed Jan 2 2002
    copyright            : (C) 2002 by Fabian Wenzel
    email                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include <wx/image.h>
#include <wx/filedlg.h>
#include <wx/dcclient.h>
#include <wx/statbox.h>
#include <wx/stattext.h>
#include <wx/menu.h>
#include <wx/msgdlg.h>
#include <wx/sizer.h>
#include <wx/event.h>

#include "vlchannelwindow.h"
#include "vlapp.h"
#include "vlview.h"
#include "vlframe.h"
#include "vldocument.h"
#include "vlprocessor.h"
#include "modules/imagechannel.h"

#include <vigra/stdimage.hxx>


using namespace std;
using namespace vistaLab;


//////////////////// Functions for channelTarget ////////////////////

/***************************************************************************
 *
 *  Event table and its IDs
 *
 ***************************************************************************/
enum {
    SNAPSHOT = 40000,
};


/***************************************************************************
 *
 *  Event table
 *
 ***************************************************************************/
BEGIN_EVENT_TABLE(channelTarget,wxPanel)
    EVT_PAINT           (channelTarget::onPaint          )
    EVT_ERASE_BACKGROUND(channelTarget::onEraseBackground)
    EVT_RIGHT_DOWN      (channelTarget::onMouse          )
    EVT_RIGHT_UP        (channelTarget::onMouse          )
    EVT_MOTION          (channelTarget::onMouseMotion    )
END_EVENT_TABLE()


/***************************************************************************/
/*!
 *  Constructor initializes the parent and clears the target bitmap
 *
 ***************************************************************************/
channelTarget::channelTarget(wxWindow *parent) :
wxPanel(parent,-1,wxDefaultPosition,wxSize(320,240),
        wxNO_BORDER| wxNO_FULL_REPAINT_ON_RESIZE ),
m_x(-1),
m_y(-1)
{
    m_targetBitmap = 0;
    SetCursor(*wxCROSS_CURSOR);
}


/***************************************************************************/
/*!
 *  Destructor deletes the target bitmap (if there is one)
 *
 ***************************************************************************/
channelTarget::~channelTarget()
{
    if(m_targetBitmap)
        delete m_targetBitmap;
}


/***************************************************************************/
/*!
 *  Refresh function copies the target image to its bitmap
 *  and does a wxWindow-Refresh(). It has been bad to
 *  get a client-DC and draw into it as there was an effect similar to
 *  delayed processing / framedrops. So calling Refresh() ist the best
 *  solution.
 *
 ***************************************************************************/
void channelTarget::refreshFrame(const vigra::BRGBImage &targetImage)
{
    {
        wxCriticalSectionLocker locker(m_bitmapLock);
        if (m_targetBitmap)
            delete m_targetBitmap;
        m_targetBitmap = new wxBitmap(wxImage(targetImage.width(),
            targetImage.height(),
            (unsigned char *) &(*targetImage.begin()),true));
    }

    Refresh(FALSE);
}


/***************************************************************************/
/*!
 *  Clears the target image
 *
 ***************************************************************************/
void channelTarget::clear()
{
    {
        wxCriticalSectionLocker locker(m_bitmapLock);
        if(m_targetBitmap){
            delete m_targetBitmap;
            m_targetBitmap = 0;
        }
    }

    Refresh(true);
}


/***************************************************************************/
/*!
 *  Updates the client size and clears the image
 *
 ***************************************************************************/
void channelTarget::updateSize(int width, int height)
{
    SetClientSize(width,height);
    clear();
}


/***************************************************************************/
/*!
 *  Paint callback: If there is a bitmap, draw it, otherwise, clears the image
 *
 ***************************************************************************/
void channelTarget::onPaint(wxPaintEvent &event)
{
    wxCriticalSectionLocker locker(m_bitmapLock);
    wxPaintDC dc(this);
    dc.BeginDrawing();
    if (m_targetBitmap) {
        dc.DrawBitmap(*m_targetBitmap,0,0);
    } else {
        dc.SetBackground(*wxBLACK_BRUSH);
        dc.Clear();
    }
    dc.EndDrawing();
}


/***************************************************************************/
/*!
 *  Does nothing (i.e. background won't be cleared to avoid refresh-flickering)
 *
 ***************************************************************************/
void channelTarget::onEraseBackground(wxEraseEvent &e)
{
}


/***************************************************************************/
/*!
 *  Forwards the mouse click to its parent (which is the window)
 *
 ***************************************************************************/
void channelTarget::onMouse(wxMouseEvent &event)
{
     if( event.RightDown() ){
        m_rightMouseDown = true;
    } else {
        m_rightMouseDown = false;

        // Setup menu
        wxMenu contextMenu(((channelWindow *)GetParent())->
            m_connectedModule->getInstanceName().c_str());

        contextMenu.Append(SNAPSHOT,_("Take Snapshot"),
            _("Saves the current image to file"));
        PopupMenu(&contextMenu,event.GetX(),event.GetY());
    }
}


/***************************************************************************/
/*!
 *  Tells the parent to display image information at the current point
 *
 ***************************************************************************/
void channelTarget::onMouseMotion(wxMouseEvent &event)
{
    // Under MSW a mouse motion event is generated even if the mouse does not
    // have been moved but if the focus or something like that changes (e.g
    // by showPixelValues - see below - which caused an infinite loop)
    if(m_x != event.GetX() || m_y != event.GetY()){
        m_x=event.GetX();
        m_y=event.GetY();
        ((channelWindow *)GetParent())->showPixelValues(m_x,m_y);
    } else
        event.Skip();
}








///////////////////////// Functions for vistaWindow /////////////////////////


BEGIN_EVENT_TABLE(channelWindow,wxPanel)
EVT_RIGHT_DOWN(channelWindow::onMouse)
EVT_RIGHT_UP  (channelWindow::onMouse)
EVT_MENU      (SNAPSHOT ,channelWindow::onSnapshot)
END_EVENT_TABLE()


/***************************************************************************/
/*!
 *  Constructor: Initializes the GUI. The arguments have to be
 *  passed and cannot be fetched from the framework as this channelwindow
 *  is constructed in the constructor of wxInputThread
 *
 ***************************************************************************/
channelWindow::channelWindow(wxWindow *parent, imageChannel *module) :
wxPanel(parent,-1,wxDefaultPosition,wxDefaultSize,wxRAISED_BORDER),
m_connectedModule(module)
{
    // Initialize non-GUI mambers
    p_view   = ((vistaLab::frame *)wxGetApp().GetTopWindow())->getChannelView();

    // The childs of this window
    m_topSizer     = new wxBoxSizer(wxVERTICAL);
    m_renderTarget = new channelTarget(this);
    m_legend       = new wxStaticText(this,-1,module->getInstanceName().c_str());

    // If anybody rightclicks on the legent
    m_legend->SetEventHandler(this);

    // Composition of GUI elements
    m_topSizer->Add(0,0,1,wxEXPAND);
    m_topSizer->Add(m_renderTarget,wxSizerFlags(0).Align(wxALIGN_TOP|wxALIGN_CENTRE_HORIZONTAL).Border(wxALL,20));
    m_topSizer->Add(0,0,1,wxEXPAND);
    m_topSizer->Add(m_legend,wxSizerFlags(0).Align(wxALIGN_BOTTOM|wxALIGN_CENTRE).Border(wxALL,20));

    SetSizer(m_topSizer);
    m_topSizer->Fit(this);

    m_topSizer->SetSizeHints(this);
}


/***************************************************************************/
/*!
 *  Destructor does nothing
 *
 ***************************************************************************/
channelWindow::~channelWindow()
{
}

/***************************************************************************/
/*!
 *  Refreshes the target by grabbing the current frame and either
 *  clearing the channel or (after maybe resizing) displaying it
 *
 ***************************************************************************/
void channelWindow::eatImage(const vigra::BRGBImage &displayImage)
{
    // Copy image
    m_targetImage = displayImage;

    // If a frame is present, check its size and display it
    int width,height;
    m_renderTarget->GetClientSize(&width,&height);
    if( width!=m_targetImage.width() ||
        height!=m_targetImage.height()){

            // Update size and layout anew
            m_renderTarget->updateSize(m_targetImage.width(),
                m_targetImage.height());
            m_topSizer->SetItemMinSize(m_renderTarget,m_renderTarget->GetSize());
            m_topSizer->SetSizeHints(this);
            m_topSizer->Fit(this);
            m_topSizer->Layout();
            GetParent()->GetSizer()->SetItemMinSize(this,GetBestSize());
            p_view->updatePanelSize();
        }
        m_renderTarget->refreshFrame(m_targetImage);
}








//////////////////// Private Functions for vistaLabWindow ////////////////////



/***************************************************************************/
/*!
 *  Right mouseclick: Might display a popup menu (if mouseclick is released)
 *
 ***************************************************************************/
void channelWindow::onMouse(wxMouseEvent &e)
{
    if( e.RightDown() ){
        m_rightMouseDown = true;
    } else {
        m_rightMouseDown = false;

        // Setup menu
        wxMenu contextMenu(m_connectedModule->getInstanceName().c_str());

        contextMenu.Append(SNAPSHOT,_("Take Snapshot"),
            _("Saves the current image to file"));
        PopupMenu(&contextMenu,e.GetX(),e.GetY());
    }
}


/***************************************************************************/
/*!
 *  As the target image is not there anymore, the renderTarget uses
 *  its bitmap to save the image
 *
 ***************************************************************************/
void channelWindow::onSnapshot(wxCommandEvent &event)
{
    // First, get fileName

    // Save only valid fileNames
    if ( m_targetImage.width() && m_targetImage.height() ){
        wxString fileName = wxFileSelector(_("Save image as"),
            wxEmptyString,
            _("output.png"), "png",
            _("PNG image (*.png)|*.png"),
            wxOVERWRITE_PROMPT|wxHIDE_READONLY|wxSAVE);
        if (fileName!=wxEmptyString){
            if (!fileName.Matches("*.png"))
                fileName+=".png";

            wxImage tmpImage(m_targetImage.width(),m_targetImage.height(),
                (unsigned char *) (&(*m_targetImage.begin())),true);
            tmpImage.SaveFile(fileName.c_str());
        }
    }
}


/***************************************************************************/
/*!
 *  Displays the values at the current pixel
 *
 ***************************************************************************/
void channelWindow::showPixelValues (int x, int y)
{
    // First check if there is a frame and if the coordinates are
    // inside the image
    if ( x >= 0 && x < m_targetImage.width() &&
         y >= 0 && y < m_targetImage.height() ) {
            ((vistaLab::frame *)wxGetApp().GetTopWindow())->GetStatusBar()->
                SetStatusText(wxString::Format(_("%s: ( %03d, %03d )\
 -  [ %03d, %03d, %03d ]"),m_connectedModule->getInstanceName().c_str(),x,y,
                                                 m_targetImage(x,y).red(),
                                                 m_targetImage(x,y).green(),
                                                 m_targetImage(x,y).blue()));
        }
}


/***************************************************************************/
/*!
 *  Displays an empty string
 *
 ***************************************************************************/
void channelWindow::clearPixelValues()
{
    m_legend->SetLabel(m_connectedModule->getInstanceName().c_str());
    m_topSizer->SetItemMinSize(m_legend,m_legend->GetSize());
}


/***************************************************************************/
/*!
 *  Updates the GUI element with the current name
 *
 ***************************************************************************/
void channelWindow::updateInstanceName()
{
    clearPixelValues();
    m_topSizer->SetSizeHints(this);
    m_topSizer->Layout();
    Refresh();
}

