/***************************************************************************
                    vlconfigmanager.cpp  -  description
                            -------------------
    begin                : Mit Mai 22 2002
    copyright            : (C) 2002 by wenzel
    email                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#include "vlconfigmanager.h"
#include "vlprocessor.h"
#include "vlrepository.h"
#include "vlapp.h"
#include "vlpin.h"

#include <boost/checked_delete.hpp>
#include <algorithm>

using namespace std;
using namespace vistaLab;

////////////////////////// Misc Functions/Classes ///////////////////////////


////////////////////////////// Public Functions //////////////////////////////



/***************************************************************************/
/*!
 *  Static function which implements the singleton pattern
 *
 ***************************************************************************/
configManager &configManager::getManager()
{
    static configManager theManager;
    return theManager;
}

/***************************************************************************/
/*!
 *  Registers a new active module by assigning outputPins and renderPins
 *
 ***************************************************************************/
void configManager::addActiveModule(module *newModule)
{
    const TConfigItems *configItems = newModule->getConfigItems();
    TConfigItems::const_iterator currentItem;

    for(currentItem  = configItems->begin();
        currentItem != configItems->end();
        ++currentItem)
        switch(currentItem->m_type){

           case configItem::OUTPUTPIN:
               m_outputPins.push_back(new outputPin(newModule,
                   currentItem->m_selections.front(),
                   currentItem->m_description,m_outputPins.size()));
               *((int *) currentItem->m_target.data)
                   = m_outputPins.size()-1;
               break;

           case configItem::RENDEROUTPUTPIN:
               m_renderPins.push_back(new renderPin(newModule,
                   (configItem::TRenderFunc) currentItem->m_target.func,
                   currentItem->m_description));
               break;

           case configItem::LOGOUTPUTPIN:
               m_logPins.push_back(new logPin(newModule,
                   (configItem::TLogFunc) currentItem->m_target.func,
                   currentItem->m_description));
               break;

           default:
               break;
        }
}


/***************************************************************************/
/*!
 *  Gets all processor previous names which provide a datatype
 *
 ***************************************************************************/
configManager::TOutputPins
configManager::getMatchingOutputPins(const std::string &type,
                                     const configurableObject *module)
{
    //void *voidptr;

    TOutputPins matches;
    TOutputPins::const_iterator currentOutputPin;
    for(currentOutputPin  = m_outputPins.begin();
        currentOutputPin != m_outputPins.end() &&
        (*currentOutputPin)-> m_moduleInstance != module;
    ++currentOutputPin)
        if ((*currentOutputPin)->m_type == type)// || type == typeid(voidptr).name())
            matches.push_back(*currentOutputPin);

    return TOutputPins(matches);
}





////////////////////////// Private Functions ///////////////////////////


/***************************************************************************/
/*!
 *  Constructor register invalid queries for the configmap
 *
 ***************************************************************************/
configManager::configManager()
{
}


/***************************************************************************/
/*!
 *  Destructor clears all internal used pins
 *
 ***************************************************************************/
configManager::~configManager()
{
    clearPins(&m_outputPins,&m_renderPins,&m_logPins);
}


/***************************************************************************/
/*!
 *  Sets all incorrecly or not connected modules to "inactive"
 *
 ***************************************************************************/
void configManager::setValidModules()
{

    TActiveModules modules = wxGetApp().getProcessor()->getModules();
    TActiveModules::iterator currentModule;

    for(currentModule  = modules.begin();
        currentModule != modules.end();
        currentModule++)
        (*currentModule)->setValid(isValidModule(modules.begin(),
        currentModule));
}


/***************************************************************************/
/*!
 *  Helper function: Determines if module is valid or not by checking all
 *  its INPUT - pins
 *
 ***************************************************************************/
bool configManager::isValidModule(TActiveModules::iterator begin,
                                  TActiveModules::iterator currentModule)
{
    TConfigItems *moduleConfigItems = (*currentModule)->getConfigItems();
    TConfigItems::const_iterator currentConfigItem;

    bool valid=true;

    for(currentConfigItem  = moduleConfigItems->begin();
        currentConfigItem != moduleConfigItems->end();
        ++currentConfigItem)
        if(currentConfigItem->m_type == configItem::INPUTPIN){
            int outputPin = *((int *) currentConfigItem->m_target.data);
            // see if the pin is connected itself
            if(outputPin == -1)
                valid=false;
            else {
                bool foundconnectedmodule=false;

                // see if the pin points to an invalid and active module
                TActiveModules::iterator inputmodule;
                for(inputmodule=begin;
                    inputmodule!=currentModule;
                    ++inputmodule)
                    if((*inputmodule)->isValid() && (*inputmodule)->isActive() && 
                    outputPin < (int) m_outputPins.size() &&
                    m_outputPins[outputPin]->m_moduleInstance == *inputmodule)
                        foundconnectedmodule=true;

                if(!foundconnectedmodule)
                    // an existing active module could not be found for the inputpin
                    valid=false;
            }
        }

        // this module has no input pins
        return valid;
}


/***************************************************************************/
/*!
 *  Sorts Output/Input/Render/Renderinput/Log/Loginput - Pins according
 *  to the new order of active modules
 *
 ***************************************************************************/
void configManager::updatePins()
{
    // Make a copy of all pin-pointers
    TOutputPins oldoutputPins(m_outputPins);
    TRenderPins oldrenderPins(m_renderPins);
    TLogPins    oldlogPins   (m_logPins);

    // Clear the list (pins actually don't get deleted as they
    // are still stored in the old lists)
    m_outputPins.clear();
    m_renderPins.clear();
    m_logPins   .clear();

    // Make up new lists
    TActiveModules modules = wxGetApp().getProcessor()->getModules();
    TActiveModules::const_iterator currentModule;

    for(currentModule = modules.begin();
        currentModule != modules.end();
        ++currentModule)
        addActiveModule(*currentModule);

    // ReConnect pins
    for(currentModule = modules.begin();
        currentModule != modules.end();
        ++currentModule){

            TConfigItems *moduleConfigItems = (*currentModule)->getConfigItems();
            TConfigItems::iterator currentConfigItem;
            for(currentConfigItem  = moduleConfigItems->begin();
                currentConfigItem != moduleConfigItems->end();
                ++currentConfigItem)

                // Process input pins
                if(currentConfigItem->m_type == configItem::INPUTPIN){
                    int slot =*((int *) currentConfigItem->m_target.data);

                    if(slot != -1){
                        // It is a connected slot

                        // first disconnect module
                        *((int *) currentConfigItem->m_target.data)=-1;


                        // It may happen (e.g. when loading from file and a
                        // module has changed) that a module does not get loaded
                        // but other ones are still referring to it	  
                        if(slot < (int) oldoutputPins.size()){

                            outputPin *firstpin = oldoutputPins[slot];

                            // Now look in the new pins for the same one
                            bool found=false;
                            TOutputPins::iterator currentpin;
                            for( currentpin  = m_outputPins.begin();
                                currentpin != m_outputPins.end() && !found;
                                ++currentpin ){
                                    if((*currentpin)->m_moduleInstance ==
                                        firstpin->m_moduleInstance &&
                                        (*currentpin)->m_description    ==
                                        firstpin->m_description){
                                            // We found a pin with the same instance and
                                            // the same description, so this must be the new one:
                                            // update the slot
                                            found=true;		  
                                            *((int *) currentConfigItem->m_target.data) =
                                                (*currentpin)->m_slot;
                                        }
                                }
                        }
                    }
                } else if (currentConfigItem->m_type == configItem::RENDERINPUTPIN){
                    vector< int > *renderslots = ((vector< int > *)
                        currentConfigItem->m_target.data);
                    vector< int >::iterator currentrenderslot=renderslots->begin();
                    while(currentrenderslot != renderslots->end()){

                        TRenderPins::iterator newrenderpin;
                        bool found=false;
                        int  i;
                        for(i=0,newrenderpin  = m_renderPins.begin();
                            newrenderpin != m_renderPins.end() && !found;
                            ++newrenderpin,++i)
                            if(*currentrenderslot < (int) oldrenderPins.size() &&
                               (*newrenderpin)->m_moduleInstance ==
                                oldrenderPins[*currentrenderslot]->m_moduleInstance &&
                                (*newrenderpin)->m_description    ==
                                oldrenderPins[*currentrenderslot]->m_description){
                                    found=true;
                                    (*currentrenderslot)=i;
                                }

                                if(!found)
                                    currentrenderslot = renderslots->erase(currentrenderslot);
                                else
                                    ++currentrenderslot;
                    }
                } else if (currentConfigItem->m_type == configItem::LOGINPUTPIN){
                    vector< int > *logslots = ((vector< int > *)
                        currentConfigItem->m_target.data);
                    vector< int >::iterator currentlogslot=logslots->begin();
                    while(currentlogslot != logslots->end()){

                        TLogPins::iterator newlogpin;
                        bool found=false;
                        int  i;
                        for(i=0,newlogpin  = m_logPins.begin();
                            newlogpin != m_logPins.end() && !found;
                            ++newlogpin,++i)
                            if(*currentlogslot < (int) oldlogPins.size() &&
                                (*newlogpin)->m_moduleInstance ==
                                oldlogPins[*currentlogslot]->m_moduleInstance &&
                                (*newlogpin)->m_description    ==
                                oldlogPins[*currentlogslot]->m_description){
                                    found=true;
                                    (*currentlogslot)=i;
                                }

                                if(!found)
                                    currentlogslot = logslots->erase(currentlogslot);
                                else
                                    ++currentlogslot;
                    }
                }

        }

        clearPins(&oldoutputPins,&oldrenderPins,&oldlogPins);
}


/***************************************************************************/
/*!
 *  Clears a set of pins. This can be either the member ones (used
 *  in the destructor) or a backup (used in sortPins)
 *
 ***************************************************************************/
void configManager::clearPins(TOutputPins *outputPins,
                               TRenderPins *renderPins,
                               TLogPins    *logPins)
{
    for_each(outputPins->begin(),outputPins->end(),
        boost::checked_deleter<outputPin>());

    outputPins->clear();

    for_each(renderPins->begin(),renderPins->end(),
        boost::checked_deleter<renderPin>());

    renderPins->clear();


    for_each(logPins->begin(),logPins->end(),
        boost::checked_deleter<logPin>());

    logPins->clear();
}


/***************************************************************************/
/*!
 *  Gets the output pin matching a slot in a slice
 *
 ***************************************************************************/
outputPin *vistaLab::configManager::getOutputPin(const int slot)
{
    TOutputPins::iterator currentOutputPin;
    for(currentOutputPin  = m_outputPins.begin();
        currentOutputPin != m_outputPins.end();
        ++currentOutputPin)
        if((*currentOutputPin)->m_slot == slot)
            return *currentOutputPin;

    return 0;
}
