/***************************************************************************
                       vldocument.cpp  -  description
                            -------------------
    begin                : Fri Nov 16 2001
    copyright            : (C) 2001 by Fabian Wenzel
    email                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include "vldocument.h"
#include "vlapp.h"
#include "vlprocessor.h"
#include "vlframe.h"
#include "vlslice.h"
#include "vlconfigmanager.h"

#include <algorithm>
#include <boost/checked_delete.hpp>


////////////////////// Functions for the document //////////////////////

using namespace vistaLab;

BEGIN_CONFIG_TOPOBJECT(document)
    CONFIGPARAMETER(m_bufferLength,__("Bufferlength"))
    CONFIGPARAMETER(m_frameRate,   __("Framerate"))
    CONFIGPARAMETER(m_waitFlag,__("Don't use the frame rate as trigger interval"))
END_CONFIG()

/***************************************************************************/
/*!
 *  Constructor initializes the document with default values and
 *  sets up the pointers
 *
 ***************************************************************************/
document::document() :
configurableObject(__("Global Settings")),
m_bufferLength      (10),
m_waitFlag          (false),
m_frameRate         (2.0),
m_sliceIsFull       (true),
m_buffer            (m_bufferLength+1,(slice *) 0),
m_preBuffer         (0),
m_currentInputSlice (m_buffer.begin()),
m_currentOutputSlice(m_buffer.begin())
{
    _initConfig(true);
}


/***************************************************************************/
/*!
 *  Deletes all remaining slices
 *
 ***************************************************************************/
document::~document()
{
    resetBuffers();
}



/***************************************************************************/
/*!
 *  Adds an image to the preBuffer. If the slice
 *  is full, the full preBuffer is copied to the current input position
 *  of the main buffer - with overwriting and deleting present ones.
 *
 ***************************************************************************/
void document::addData(module *stream, slice *data)
{
    wxCriticalSectionLocker lock(m_preBufferLock);

    // If the slice is full, switch to the other one
    if(m_sliceIsFull){
        m_preBuffer = new slice();
    }

    if (m_preBuffer->containsData(*data))
        // There has been data in the current slot, so delete it!
        stream->cleanup(*m_preBuffer);
    else
        m_preBuffer->increaseValidEntries();

    m_preBuffer->merge(*data);

    m_sliceIsFull = ( m_preBuffer->getValidEntries() ==
        wxGetApp().getProcessor()->getNumberOfInputThreads() );

    // If there is no more free frame, notify the processor
    if(m_sliceIsFull){

        processCore::prepareForProcessing();

        delete *m_currentInputSlice;
        *m_currentInputSlice = m_preBuffer;

        m_preBuffer = 0;
        processCore::continueProcessing();
    }
}


/***************************************************************************/
/*!
 *  If the process core has successfuly detected that there is new
 *  data at the buffer's current input position it shifts the buffer
 *  in a ringbuffer-like fashion. The slice at the new ringbuffer's input
 *  position is deleted and set to zero such that following checks for new
 *  data don't succeed until it is added by document::addImage()
 *
 ***************************************************************************/
void document::shiftBuffer()
{
    m_currentOutputSlice = m_currentInputSlice;

    // Implements ringbuffer behavior
    ++m_currentInputSlice;

    if(m_currentInputSlice==m_buffer.end()){
        m_currentInputSlice = m_buffer.begin();
    }

    delete *m_currentInputSlice;
    *m_currentInputSlice = 0;

    // If there are no input threads, the slice has not been generated
    // at all, so generate it now
    if (!*m_currentOutputSlice)
        *m_currentOutputSlice = new slice();
}

/***************************************************************************/
/*!
 *  Removes all remaining buffers and resets iterators and the full-flag
 *
 ***************************************************************************/
void document::resetBuffers()
{
    std::for_each(m_buffer.begin(),m_buffer.end(),
        boost::checked_deleter<slice>());

    m_buffer.assign(m_buffer.size(), 0);

    delete m_preBuffer;
    m_preBuffer = 0;

    m_sliceIsFull       = true;
    m_currentInputSlice = m_buffer.begin();
    m_currentOutputSlice= m_buffer.begin();
}


/***************************************************************************/
/*!
 *  The output frame - pointer is totally independent of the input frame
 *  pointer when not running (otherwise is gets shifted in
 *  document::addImage(). This function gets called when the slider in
 *  the GUI has been moved
 *
 ***************************************************************************/
void document::shiftSlicePointer(bool downwards)
{
    if(downwards){
        if (m_currentOutputSlice==m_buffer.begin())
            m_currentOutputSlice=m_buffer.end();
        --m_currentOutputSlice;
    } else {
        ++m_currentOutputSlice;
        if (m_currentOutputSlice==m_buffer.end())
            m_currentOutputSlice=m_buffer.begin();
    }
}


/***************************************************************************/
/*!
 *  Gets called when the bufferLength might have changed
 *
 ***************************************************************************/
void document::postConfig()
{
    // Buffer has to keep an empty spot for new images
    m_buffer.resize(m_bufferLength+1);


    // Even though the input- and outputSlice has been set
    // to begin in resetBuffers() a few lines above, the
    // buffer memory may have shifted after resize(), so
    // to be safe, get new pointers
    m_currentInputSlice = m_buffer.begin();
    m_currentOutputSlice= m_buffer.begin();
    ((frame *)wxGetApp().GetTopWindow())->updateSlider();
}
