/***************************************************************************
                      vlframecontrols.cpp  -  description
                            -------------------
    begin                : Thu Aug 15 2001
    copyright            : (C) 2002 by Fabian Wenzel
    email                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include <algorithm>
#include <wx/dnd.h>
#include <wx/imaglist.h>
#include <wx/utils.h>
#include <wx/filename.h>

#include "vlframecontrols.h"

// For common IDs
#include "vlframe.h"

// For GUI classes
#include "vlapp.h"
#include "vlrepository.h"
#include "vlprocessor.h"
#include "vlmoduleconfigdialog.h"
#include "vlconfigmanager.h"
#include "vlpin.h"

// ICONS
#include "resources/modulecategory.xpm"
#include "resources/openmodulecategory.xpm"
#include "resources/moduleicon.xpm"
#include "resources/moduleiconnotconnected.xpm"
#include "resources/moduleiconinactive.xpm"

#ifndef WIN32
#include "resources/moduleiconcursor.xpm"
#endif

#include "resources/treeparamboolfalse.xpm"
#include "resources/treeparambooltrue.xpm"
#include "resources/treeparamstring.xpm"
#include "resources/treeparamint.xpm"
#include "resources/treeparamdouble.xpm"
#include "resources/treeparamsinglechoice.xpm"
#include "resources/treeparaminputpin.xpm"
#include "resources/treeparamfilename.xpm"

#include <boost/checked_delete.hpp>
#include <boost/bind.hpp>

#define GETVISTALABDATA(a) ((treeItemData *) GetItemData(a))

using namespace std;
using namespace vistaLab;

enum{
    ACTIVATE_MODULE=10000,
    INACTIVATE_MODULE
};


///////////////////////////////// DropTarget ////////////////////////////////

namespace vistaLab {

/***************************************************************************/
//! Drop Target handles DND Drops for the modulesList
/*!
 *
 * \author  Fabian Wenzel
 *
 ***************************************************************************/
class moduleDropTarget : public wxTextDropTarget{
public:
    moduleDropTarget(activeProjectTree *activeprojecttree) :
      p_activeprojecttree(activeprojecttree) {}
private:
    bool OnDropText(int x, int y, const wxString &data);
    activeProjectTree *p_activeprojecttree;
};


/***************************************************************************/
/*!
 *  Callback function which may trigger the modulesList for adding
 *  a module
 *
 ***************************************************************************/
bool moduleDropTarget::OnDropText(int x, int y, const wxString &data)
{
    if(!wxGetApp().getProcessor()->isRunning()){
        p_activeprojecttree->addModule(x,y,data);
        return true;
    } else {
        wxMessageBox(_("Stop processing first"),_("Notification"),wxOK|wxCENTRE);
        return false;
    }
}

}


//////////////////////////// repositoryTree ///////////////////////////


/***************************************************************************
 *
 *  Event table
 *
 ***************************************************************************/
BEGIN_EVENT_TABLE(repositoryTree,wxTreeCtrl)
EVT_TREE_ITEM_ACTIVATED(ID_REPOSITORYTREE,repositoryTree::onAddModule)
EVT_TREE_BEGIN_DRAG    (ID_REPOSITORYTREE,repositoryTree::onDrag)
EVT_MOUSE_EVENTS       (repositoryTree::onMouse)
END_EVENT_TABLE()


/***************************************************************************/
/*!
 *  Constructor initializes images for wxWidgets (Icons) and places all
 *  modules the repository has as tree items
 *
 ***************************************************************************/
repositoryTree::repositoryTree(wxWindow *parent, const wxWindowID id) :
wxTreeCtrl(parent,id,wxDefaultPosition,wxDefaultSize,
           wxTR_DEFAULT_STYLE|wxSUNKEN_BORDER|wxTR_SINGLE|wxTR_HIDE_ROOT)
{
    SetBackgroundColour(wxSystemSettings::GetColour(wxSYS_COLOUR_MENU));
    // Add images
    wxImageList *images = new wxImageList(16,16,true);

    images->Add(wxIcon(modulecategory_xpm));
    images->Add(wxIcon(openmodulecategory_xpm));
    images->Add(wxIcon(moduleicon_xpm));

    AssignImageList(images);

    // Root node (too bad that there can only be one)
    AddRoot(_("Modules"),repositoryTree::moduleCategory);

    initRepository();
}

void repositoryTree::cleanRepository()
{
    DeleteChildren(GetRootItem());
    m_categories.clear();
}

void repositoryTree::initRepository()
{
    // Loop through all modules and add them to the menu
    TModuleRepository modules = wxGetApp().getRepository()->getModules();
    TModuleRepository::iterator currentModule;

    for (currentModule  = modules.begin();
        currentModule != modules.end();
        currentModule++){

            // Look if there's already a category with that name
            wxString categoryString = (*currentModule)->getModuleCategory().c_str();
            list< wxTreeItemId >::iterator currentCategory = m_categories.begin();
            while ( currentCategory != m_categories.end() &&
                GetItemText(*currentCategory) != categoryString &&
                GetItemText(*currentCategory) != _(categoryString))
                ++currentCategory;

            if(currentCategory == m_categories.end()){
                // A new category is needed
                wxTreeItemId newCategory = AppendItem(GetRootItem(),
                    _(categoryString));
                SetItemImage(newCategory,repositoryTree::moduleCategory,wxTreeItemIcon_Normal);
                SetItemImage(newCategory,repositoryTree::openModuleCategory,wxTreeItemIcon_Expanded);

                m_categories.push_back(newCategory);

                AppendItem(newCategory,_((*currentModule)->getName().c_str()),
                    repositoryTree::moduleIcon);
            } else
                // An existing category is there
                AppendItem(*currentCategory,_((*currentModule)->getName().c_str()),
                repositoryTree::moduleIcon);
        }
}


/***************************************************************************/
/*!
 *  When double clicking on a module, it also gets appended to the active
 *  modules
 *
 ***************************************************************************/
void repositoryTree::onAddModule(wxTreeEvent &event)
{
    // It has to be checked if it is a module we are clicking on.
    // This is done via the item image as it is the easiest hack.

    if(GetItemImage(event.GetItem()) != repositoryTree::moduleIcon){
        event.Skip();
        return;
    }

    moduleConfigDialog config;
    ((frame *) wxGetApp().GetTopWindow())->
        addModule(GetItemText(event.GetItem()),config);
}


/***************************************************************************/
/*!
 *  Starts Drag & Drop if it is a child node
 *
 ***************************************************************************/
void repositoryTree::onDrag(wxTreeEvent &event)
{
    if( find(m_categories.begin(),
        m_categories.end(),
        event.GetItem()) == m_categories.end()){
            // A Child node has been found
            wxTextDataObject modulename_object(GetItemText(event.GetItem()));
            wxDropSource dragSource(modulename_object,this,wxDROP_ICON(moduleiconcursor),
                wxDROP_ICON(moduleiconcursor),
                wxDROP_ICON(moduleiconcursor));

            dragSource.DoDragDrop(true);
        }
}


/***************************************************************************/
/*!
 *  Displays information in the status bar as soon as a target is hit
 *
 ***************************************************************************/
void repositoryTree::onMouse(wxMouseEvent &event)
{
    int hitTestFlags = wxTREE_HITTEST_ONITEMBUTTON | wxTREE_HITTEST_ONITEMICON | wxTREE_HITTEST_ONITEMLABEL;
    wxTreeItemId index = HitTest(event.GetPosition(), hitTestFlags);
    if(index.IsOk() && index != currentId){

        TModuleRepository modules = wxGetApp().getRepository()->getModules();
        TModuleRepository::iterator currentModule;

        wxString HelpString;
        for (currentModule  = modules.begin();
            currentModule != modules.end() && HelpString == wxEmptyString;
            currentModule++){
                if(wxString(_((*currentModule)->getName().c_str())) == GetItemText(index))
                    HelpString = (*currentModule)->getHelpString().c_str();
            }
            ((frame *)(wxGetApp().GetTopWindow()))->GetStatusBar()->SetStatusText(HelpString);

            currentId = index;
    }
    event.Skip();
}



////////////////////////////    modulesList   ///////////////////////////
extern const wxEventType wxEVT_VISTALAB_UPDATEPROJECTTREE;

const wxEventType wxEVT_VISTALAB_UPDATEPROJECTTREE = wxNewEventType();

#define EVT_VISTALAB_UPDATEPROJECTTREE(id, fn) \
    DECLARE_EVENT_TABLE_ENTRY( \
    wxEVT_VISTALAB_UPDATEPROJECTTREE, id, -1, \
    (wxObjectEventFunction)(wxEventFunction)(wxCommandEventFunction)&fn, \
    (wxObject *) NULL \
    ),


namespace vistaParameterItems {

class treeItemData: public wxTreeItemData{
public:

    virtual bool isEditable()
    {
        return false;
    }

    virtual void doubleClick()
    {
    }

    virtual void rightClick(const wxPoint &position)
    {
    }

    virtual const wxString &getToolTip()
    {
        return m_toolTip;
    }
    virtual bool eatChar(int keyCode)
    {
        return false;
    }

    virtual void update()=0;
protected:
    activeProjectTree *p_ctrl;
    wxString           m_toolTip;
};


class parameterItemBase: public treeItemData, public parameterChangeBase
{
public:
    parameterItemBase(const configItem *configItem,
        configurableObject *module):
    parameterChangeBase(configItem,module)
    {
    }
    virtual const char *doneWithEditing(const wxString &text)
    {
        return NULL;
    }
};


template< class T > class parameterItem: public parameterChange<T, parameterItemBase>{
public:
    parameterItem(const configItem *configItem,
        activeProjectTree *ctrl,
        configurableObject *module):
    parameterChange<T,parameterItemBase>(configItem,module),
        p_ctrl(ctrl)
    {
        this->m_toolTip = configItem->m_localizedDescription.c_str();
        this->m_value   = this->target();
    }
protected:
    activeProjectTree *p_ctrl;
    T m_value;
};

template< class T > class editableTreeItem: public parameterItem<T>{
public:
    editableTreeItem(const configItem *configItem,
        activeProjectTree *ctrl,
        configurableObject *module):
    parameterItem<T>(configItem,ctrl,module)
    {
    }

    virtual bool isEditable()
    {
        return true;
    }

    virtual void doubleClick()
    {
        this->p_ctrl->EditLabel(this->GetId());
    }

    virtual bool eatChar(int keyCode)
    {
        if(keyCode == WXK_F2){
            this->p_ctrl->EditLabel(this->GetId());
            return true;
        } else
            return false;
    }
};

class rootItemData: public treeItemData{
public:
    virtual const wxString &getToolTip()
    {
        return ((frame *)wxGetApp().GetTopWindow())->getCurrentProject();
    }
    virtual void update()
    {
    }
};

class paramIntItemData: public editableTreeItem<int>{
public:
    paramIntItemData(const configItem *configItem,
        activeProjectTree *ctrl,
        configurableObject *module):
    editableTreeItem<int>(configItem,ctrl,module)
    {
    }

    virtual const char *doneWithEditing(const wxString &text)
    {
        long tmp;
        if (text.ToLong(&tmp)){
            m_value = tmp;
            p_ctrl->addPendingParameterChange(this);
            return NULL;
        } else {
            return _("This should be an integer!");
        }
    }
    virtual void update()
    {
        p_ctrl->SetItemText (GetId(),wxString::Format("%d",target()));
        p_ctrl->SetItemImage(GetId(),m_iconId,wxTreeItemIcon_Normal);
        p_ctrl->SetItemImage(GetId(),m_iconId,wxTreeItemIcon_Selected);
    }
    virtual bool eatChar(int keyCode)
    {
        if(editableTreeItem<int>::eatChar(keyCode))
            return true;

        if(keyCode == '+'){
            m_value = target()+1;
            p_ctrl->addPendingParameterChange(this);
            return true;
        } else if (keyCode == '-')
        {
            m_value = target()-1;
            p_ctrl->addPendingParameterChange(this);
            return true;
        }
        return false;
    }

    void save()
    {
        store(m_value);
    }

    static int m_iconId;
};

int paramIntItemData::m_iconId;

class fileNameItemData : public parameterItem<std::string>{
public:
    fileNameItemData(const configItem *configItem,
        activeProjectTree *ctrl,
        configurableObject *module):
    parameterItem<std::string>(configItem,ctrl,module),
    m_path(wxPathOnly(target().c_str()))
    {
    }

    virtual void doubleClick()
    {
        wxString fileName =  wxFileSelector(_("File"),m_path,
           wxEmptyString,wxEmptyString,"*.*");
        if(fileName != wxEmptyString){
            m_value = fileName;
            p_ctrl->addPendingParameterChange(this);
            if(p_ctrl->tryApplyChanges())
                update();
            m_path = wxPathOnly(fileName);
        }
    }

    virtual void update()
    {
        wxString fileName;
        wxString extension;
        wxFileName::SplitPath(target().c_str(),NULL,&fileName,&extension);
        p_ctrl->SetItemText(GetId(),fileName+wxString(".")+extension);
        p_ctrl->SetItemImage(GetId(),m_iconId,wxTreeItemIcon_Normal);
        p_ctrl->SetItemImage(GetId(),m_iconId,wxTreeItemIcon_Selected);
    }

    void save()
    {
        store(m_value);
    }

    static int m_iconId;
private:
    wxString m_path;
};

int fileNameItemData::m_iconId;

class paramBoolItemData: public wxEvtHandler, public parameterItem<bool>{
public:
    paramBoolItemData(const configItem *configItem,
        activeProjectTree *ctrl,
        configurableObject *module):
    parameterItem<bool>(configItem,ctrl,module)
    {
        m_window = new wxWindow(ctrl,-1);
        m_window->Show(FALSE);
        m_window->PushEventHandler(this);

        // Make context menu
        m_contextMenu = new wxMenu(p_item->m_localizedDescription.c_str());
        int id=wxNewId();

        Connect(id,wxEVT_COMMAND_MENU_SELECTED,
            (wxObjectEventFunction) (wxCommandEventFunction)
            &paramBoolItemData::onMenuTrue);
        m_contextMenu->Append(id,m_yesString);

        id = wxNewId();
        Connect(id,wxEVT_COMMAND_MENU_SELECTED,
            (wxObjectEventFunction) (wxCommandEventFunction)
            &paramBoolItemData::onMenuFalse);
        m_contextMenu->Append(id,m_noString);
    }

    virtual ~paramBoolItemData()
    {
        m_window->Destroy();
        delete m_contextMenu;
    }

    void updateToTrue(){
        p_ctrl->SetItemText(wxTreeItemData::GetId(),m_yesString);
        p_ctrl->SetItemImage(wxTreeItemData::GetId(),m_iconYesId,wxTreeItemIcon_Normal);
        p_ctrl->SetItemImage(wxTreeItemData::GetId(),m_iconYesId,wxTreeItemIcon_Selected);
    }

    void updateToFalse(){
        p_ctrl->SetItemText(wxTreeItemData::GetId(),m_noString);
        p_ctrl->SetItemImage(wxTreeItemData::GetId(),m_iconNoId,wxTreeItemIcon_Normal);
        p_ctrl->SetItemImage(wxTreeItemData::GetId(),m_iconNoId,wxTreeItemIcon_Selected);
    }

    virtual void update()
    {
        if(target())
            updateToTrue();
        else
            updateToFalse();
    }

    virtual bool eatChar(int keyCode)
    {
        if(keyCode == '+' && !target()){
            m_value = true;
            p_ctrl->addPendingParameterChange(this);
            return true;
        } else if (keyCode == '-' && target()){
            m_value = false;
            p_ctrl->addPendingParameterChange(this);
            return true;
        }
        return false;
    }

    virtual void rightClick(const wxPoint &position)
    {
        m_window->PopupMenu(m_contextMenu,position);
    }

    void onMenuTrue(wxCommandEvent &event){
        if(!target()){
            m_value = true;
            p_ctrl->addPendingParameterChange(this);
            if(p_ctrl->tryApplyChanges())
                update();
        }
    }

    void onMenuFalse(wxCommandEvent &event)
    {
        if(target()){
            m_value = false;
            p_ctrl->addPendingParameterChange(this);
            if(p_ctrl->tryApplyChanges())
                update();
        }
    }

    void save()
    {
        store(m_value);
    }

    static int m_iconYesId;
    static int m_iconNoId;
    static wxString m_yesString;
    static wxString m_noString;
private:
    wxWindow *m_window;
    wxMenu *m_contextMenu;
    bool    m_value;
};
int paramBoolItemData::m_iconYesId;
int paramBoolItemData::m_iconNoId;
wxString paramBoolItemData::m_yesString;
wxString paramBoolItemData::m_noString;

class paramStringItemData: public editableTreeItem<std::string>{
public:
    paramStringItemData(const configItem *configItem,
        activeProjectTree *ctrl,
        configurableObject *module):
    editableTreeItem<std::string>(configItem,ctrl,module)
    {
    }

    virtual const char *doneWithEditing(const wxString &text)
    {
        m_value = text.c_str();
        p_ctrl->addPendingParameterChange(this);
        return NULL;
    }

    virtual void update()
    {
        p_ctrl->SetItemText(GetId(),target().c_str());
        p_ctrl->SetItemImage(GetId(),m_iconId,wxTreeItemIcon_Normal);
        p_ctrl->SetItemImage(GetId(),m_iconId,wxTreeItemIcon_Selected);
    }

    void save()
    {
        store(m_value);
    }
    static int m_iconId;
};

int paramStringItemData::m_iconId;

class paramDoubleItemData: public editableTreeItem<double>{
public:
    paramDoubleItemData(const configItem *configItem,
        activeProjectTree *ctrl,
        configurableObject *module):
    editableTreeItem<double>(configItem,ctrl,module)
    {
    }
    virtual const char *doneWithEditing(const wxString &text)
    {
        double tmp;
        if (text.ToDouble(&tmp)){
            m_value = tmp;
            p_ctrl->addPendingParameterChange(this);
            return NULL;
        } else {
            return _("This should be a double value!");
        }
    }

    virtual void update()
    {
        p_ctrl->SetItemText(GetId(),wxString::Format("%f",target()));
        p_ctrl->SetItemImage(GetId(),m_iconId,wxTreeItemIcon_Normal);
        p_ctrl->SetItemImage(GetId(),m_iconId,wxTreeItemIcon_Selected);
    }

    virtual bool eatChar(int keyCode)
    {
        if(editableTreeItem<double>::eatChar(keyCode))
            return true;

        if(keyCode == '+'){
            m_value = target()+1.0;
            p_ctrl->addPendingParameterChange(this);
            return true;
        } else if (keyCode == '-')
        {
            m_value = target()-1.0;
            p_ctrl->addPendingParameterChange(this);
            return true;
        }
        return false;
    }

    void save()
    {
        store(m_value);
    }
    static int m_iconId;
};

int paramDoubleItemData::m_iconId;

class singleChoiceItemData: public wxEvtHandler, public parameterItem<int>{
public:
    singleChoiceItemData(const configItem *configItem,
        activeProjectTree *ctrl,
        configurableObject *module):
    parameterItem<int>(configItem,ctrl,module)
    {
        m_window = new wxWindow(ctrl,-1);
        m_window->Show(FALSE);
        m_window->PushEventHandler(this);

        // Make context menu
        m_contextMenu = new wxMenu(p_item->m_localizedDescription.c_str());

        for(unsigned int i=0;i<configItem->m_selections.size();++i){
            m_ids.push_back(wxNewId());
            m_contextMenu->Append(m_ids[i],p_item->m_selections[i].c_str());
            Connect(m_ids[i],wxEVT_COMMAND_MENU_SELECTED,
                (wxObjectEventFunction) (wxCommandEventFunction)
                &singleChoiceItemData::onMenu);
        }
    }

    virtual ~singleChoiceItemData()
    {
        m_window->Destroy();
        delete m_contextMenu;
    }

    virtual void update()
    {
        p_ctrl->SetItemText(wxTreeItemData::GetId(),p_item->m_selections[target()].c_str());
        p_ctrl->SetItemImage(wxTreeItemData::GetId(),m_iconId,wxTreeItemIcon_Normal);
        p_ctrl->SetItemImage(wxTreeItemData::GetId(),m_iconId,wxTreeItemIcon_Selected);
    }

    virtual bool eatChar(int keyCode){
        if(keyCode == '+'){
            unsigned int nextSelection = target()+1;
            if(nextSelection < p_item->m_selections.size()){
                m_value=nextSelection;
                p_ctrl->addPendingParameterChange(this);
                return true;
            }
        } else if (keyCode == '-')
        {
            unsigned int nextSelection = target()-1;
            if(target()>0){
                m_value=nextSelection;
                p_ctrl->addPendingParameterChange(this);
                return true;
            }
        }
        return false; 
    }

    virtual void rightClick(const wxPoint &position)
    {
        m_window->PopupMenu(m_contextMenu,position);
    }

    void onMenu(wxCommandEvent &event){
        for(unsigned int i=0;i<p_item->m_selections.size();++i)
            if(m_ids[i] == event.GetId()){
                m_value=i;
                p_ctrl->addPendingParameterChange(this);
                if(p_ctrl->tryApplyChanges())
                    update();
            }
    }

    void save()
    {
        store(m_value);
    }

    static int m_iconId;

private:
    std::vector< int > m_ids;
    wxWindow *m_window;
    wxMenu *m_contextMenu;
};

int singleChoiceItemData::m_iconId;

class inputPinItemData: public wxEvtHandler, public parameterItem<int>{
public:
    inputPinItemData(const configItem *configItem,
        activeProjectTree *ctrl,
        configurableObject *module):
    parameterItem<int>(configItem,ctrl,module),
    m_contextMenu(0)
    {
        m_window = new wxWindow(ctrl,-1);
        m_window->Show(FALSE);
        m_window->PushEventHandler(this);

    }

    virtual ~inputPinItemData()
    {
        m_window->Destroy();
        delete m_contextMenu;
    }

    virtual void update()
    {
        wxString itemText=_("Not connected");
        if(target() != -1){
            m_outputPins = configManager::getManager().
                getMatchingOutputPins(p_item->m_selections.front(),(module *)getModule());

            configManager::TOutputPins::iterator currentPin=m_outputPins.begin();
            while(currentPin != m_outputPins.end())
            {
                if((*currentPin)->m_slot == target())
                    itemText = wxString::Format("%s of %s",
                    _((*currentPin)->m_description.c_str()),
                    (*currentPin)->m_moduleInstance->getInstanceName().c_str());
                ++currentPin;
            }

        }

        p_ctrl->SetItemText(wxTreeItemData::GetId(),itemText);
        p_ctrl->SetItemImage(wxTreeItemData::GetId(),m_iconId,wxTreeItemIcon_Normal);
        p_ctrl->SetItemImage(wxTreeItemData::GetId(),m_iconId,wxTreeItemIcon_Selected);
    }

    virtual void rightClick(const wxPoint &position)
    {
        if(!wxGetApp().getProcessor()->isRunning()){
            m_ids.clear();

            // Make context menu
            delete m_contextMenu;
            m_contextMenu = new wxMenu(p_item->m_localizedDescription.c_str());

            m_outputPins = configManager::getManager().
                getMatchingOutputPins(p_item->m_selections.front(),(module *)getModule());

            configManager::TOutputPins::iterator currentPin=m_outputPins.begin();

            // First pin: "Not connected"
            m_ids.push_back(wxNewId());
            m_contextMenu->Append(m_ids[0],_("Not connected"));
            Connect(m_ids[0],wxEVT_COMMAND_MENU_SELECTED,
                (wxObjectEventFunction) (wxCommandEventFunction)
                    &inputPinItemData::onMenu);

            for(unsigned int i=1;i<=m_outputPins.size();++i){
                m_ids.push_back(wxNewId());
                m_contextMenu->Append(m_ids[i],wxString::Format("%s of %s",
                        _((*currentPin)->m_description.c_str()),
                        (*currentPin)->m_moduleInstance->getInstanceName().c_str()));
                Connect(m_ids[i],wxEVT_COMMAND_MENU_SELECTED,
                    (wxObjectEventFunction) (wxCommandEventFunction)
                    &inputPinItemData::onMenu);
                ++currentPin;
            }

            m_window->PopupMenu(m_contextMenu,position);
        }
    }

    void onMenu(wxCommandEvent &event){
        for(unsigned int i=0;i<m_ids.size();++i){
            if(m_ids[i] == event.GetId()){
                if ( i <= 0 )
                    m_value=-1;
                else if (i > 0)
                    m_value=m_outputPins[i-1]->m_slot;

                p_ctrl->addPendingParameterChange(this);

                // The "if" is not necessary: In rightClick we check if
                // the program is processing, so this function will
                // only be called if we are not, i.e. changes will apply
                // immediately
                if(p_ctrl->tryApplyChanges())
                    update();

                wxCommandEvent eventCustom(wxEVT_VISTALAB_UPDATEPROJECTTREE,p_ctrl->GetId());
                wxPostEvent(p_ctrl, eventCustom);
            }
        }
    }

    void save()
    {
        store(m_value);
    }

    static int m_iconId;
private:
    wxWindow *m_window;
    wxMenu   *m_contextMenu;
    std::vector< int > m_ids;
    configManager::TOutputPins m_outputPins;
};

int inputPinItemData::m_iconId;

class moduleItemData: public wxEvtHandler, public paramStringItemData{
public:
    moduleItemData(const configItem *configItem,
        activeProjectTree *ctrl,
        configurableObject *module) :
    paramStringItemData(configItem,ctrl,module)
    {
        m_window = new wxWindow(ctrl,-1);
        m_window->Show(FALSE);
        m_window->PushEventHandler(this);

        m_contextMenu = new wxMenu();

        // Set up Event IDs
        int id = wxNewId();
        m_contextMenu->Append(id,_("Configure"),
            _("Configures the current module"));
        Connect(id,wxEVT_COMMAND_MENU_SELECTED,
            (wxObjectEventFunction) (wxCommandEventFunction)
            &moduleItemData::onConfigure);

        id = wxNewId();
        m_contextMenu->Append(id,_("Remove"),
            _("Removes the current module"));
        Connect(id,wxEVT_COMMAND_MENU_SELECTED,
            (wxObjectEventFunction) (wxCommandEventFunction)
            &moduleItemData::onRemove);

        id = wxNewId();
        m_contextMenu->Append(id,_("Set Active"),
            _("Sets the module active"));
        Connect(id,wxEVT_COMMAND_MENU_SELECTED,
            (wxObjectEventFunction) (wxCommandEventFunction)
            &moduleItemData::SetActive);

        id = wxNewId();
        m_contextMenu->Append(id,_("Set Inactive"),
            _("Sets the module inactive"));
        Connect(id,wxEVT_COMMAND_MENU_SELECTED,
            (wxObjectEventFunction) (wxCommandEventFunction)
            &moduleItemData::SetInactive);
    }

    virtual void doubleClick()
    {
        configure();
    }

    virtual const char *doneWithEditing(const wxString &text)
    {
        m_value = text.c_str();
        p_ctrl->addPendingParameterChange(this);
        wxCommandEvent eventCustom(wxEVT_VISTALAB_UPDATEPROJECTTREE,p_ctrl->GetId());
        wxPostEvent(p_ctrl, eventCustom);
        return NULL;
    }

    void configure()
    {
        moduleConfigDialog config;
        config.attachModule(getModule());
        config.configure();

        wxCommandEvent eventCustom(wxEVT_VISTALAB_UPDATEPROJECTTREE,p_ctrl->GetId());
        wxPostEvent(p_ctrl, eventCustom);
    }

    virtual bool isEditable()
    {
        return !wxGetApp().getProcessor()->isRunning();
    }

    virtual ~moduleItemData()
    {
        m_window->PopEventHandler();
        m_window->Destroy();
        delete m_contextMenu;
    }

    virtual void update()
    {
        paramStringItemData::update();
        int iconId=m_iconOkayId;
        if(!((module *)getModule())->isValid())
            iconId=m_iconInvalidId;
        if(!((module *)getModule())->isActive())
            iconId=m_iconInactiveId;

        p_ctrl->SetItemImage(wxTreeItemData::GetId(),iconId,wxTreeItemIcon_Normal);
        p_ctrl->SetItemImage(wxTreeItemData::GetId(),iconId,wxTreeItemIcon_Selected);
    }

    virtual void rightClick(const wxPoint &position)
    {
        // Make context menu

        m_contextMenu->SetTitle(target().c_str());
        m_window->PopupMenu(m_contextMenu,position);
    }

    void onConfigure(wxCommandEvent &event)
    {
        configure();
    }

    void onRemove(wxCommandEvent &event)
    {
        wxGetApp().getProcessor()->removeModule((module *)getModule());
        wxCommandEvent eventCustom(wxEVT_VISTALAB_UPDATEPROJECTTREE,p_ctrl->GetId());
        wxPostEvent(p_ctrl, eventCustom);
    }

    void SetActive(wxCommandEvent &event)
    {
        ((module *)getModule())->setActive(true);
        wxGetApp().getProcessor()->needsSorting();
        wxCommandEvent eventCustom(wxEVT_VISTALAB_UPDATEPROJECTTREE,p_ctrl->GetId());
        wxPostEvent(p_ctrl, eventCustom);
    }

    void SetInactive(wxCommandEvent &event)
    {
        ((module *)getModule())->setActive(false);
        wxGetApp().getProcessor()->needsSorting();
        wxCommandEvent eventCustom(wxEVT_VISTALAB_UPDATEPROJECTTREE,p_ctrl->GetId());
        wxPostEvent(p_ctrl, eventCustom);
    }

    void addParameters()
    {
        update();

        const TConfigItems *configItems = getModule()->getConfigItems();
        TConfigItems::const_iterator currentItem=configItems->begin();

        while(currentItem != configItems->end()){
            treeItemData *newItem=0;

            switch(currentItem->m_type){

                // Special case: The instance name does not have to be displayed
case configItem::STRING :
    if(currentItem->m_description != __("Instancename"))
        newItem = new paramStringItemData(&(*currentItem),p_ctrl,getModule());
    break;

case configItem::INT:
    newItem = new paramIntItemData(&(*currentItem),p_ctrl,getModule());
    break;

case configItem::BOOL:
    newItem = new paramBoolItemData(&(*currentItem),p_ctrl,getModule());
    break;

case configItem::DOUBLE:
    newItem = new paramDoubleItemData(&(*currentItem),p_ctrl,getModule());
    break;

case configItem::SINGLECHOICE:
    newItem = new singleChoiceItemData(&(*currentItem),p_ctrl,getModule());
    break;

case configItem::INPUTPIN:
    newItem = new inputPinItemData(&(*currentItem),p_ctrl,getModule());
    break;

case configItem::FILENAME:
    newItem = new fileNameItemData(&(*currentItem),p_ctrl,getModule());
    break;

default:
    break;
            }

            if(newItem){
                p_ctrl->AppendItem(wxTreeItemData::GetId(),wxEmptyString,0,-1,newItem);
                newItem->update();
            }

            ++currentItem;
        }

    }

    static const configItem *getConfigItem(module *module){

        // We have to look for the instance name to
        // attach it to the treedata.
        TConfigItems *configItems = module->getConfigItems();
        TConfigItems::iterator currentItem = configItems->begin();

        while((*currentItem).m_description != std::string("Instancename")){
            ++currentItem;
        }

        return &(*currentItem);
    }

    void save()
    {
        store(m_value);
        wxGetApp().getProcessor()->checkInstanceName((module *)getModule());
        // Update item if the module name has changed
        if(m_value != ((module *)getModule())->getInstanceName()){
            m_value = ((module *)getModule())->getInstanceName();
            wxCommandEvent eventCustom(wxEVT_VISTALAB_UPDATEPROJECTTREE,p_ctrl->GetId());
            wxPostEvent(p_ctrl, eventCustom);
        }
    }

    static int m_iconInvalidId;
    static int m_iconOkayId;
    static int m_iconInactiveId;

private:
    wxWindow *m_window;
    wxMenu   *m_contextMenu;
};

int moduleItemData::m_iconInvalidId;
int moduleItemData::m_iconOkayId;
int moduleItemData::m_iconInactiveId;

}

using namespace vistaParameterItems;

/***************************************************************************
 *
 *  Event table
 *
 ***************************************************************************/

BEGIN_EVENT_TABLE(activeProjectTree,wxTreeCtrl)
EVT_CHAR           (activeProjectTree::onChar)
EVT_MOTION   (activeProjectTree::onMouseMotion)
EVT_TREE_BEGIN_DRAG(ID_ACTIVEMODULELIST,activeProjectTree::onBeginDrag)
EVT_TREE_END_DRAG  (ID_ACTIVEMODULELIST,activeProjectTree::onEndDrag)
EVT_TREE_ITEM_RIGHT_CLICK(ID_ACTIVEMODULELIST,activeProjectTree::onRightClick)
EVT_TREE_BEGIN_LABEL_EDIT(ID_ACTIVEMODULELIST,activeProjectTree::onBeginLabelEdit)
EVT_TREE_END_LABEL_EDIT(ID_ACTIVEMODULELIST,activeProjectTree::onEndLabelEdit)
EVT_TREE_ITEM_ACTIVATED(ID_ACTIVEMODULELIST,
                        activeProjectTree::onConfigureDoubleClick)
EVT_VISTALAB_UPDATEPROJECTTREE(ID_ACTIVEMODULELIST,activeProjectTree::onUpdateModules)
END_EVENT_TABLE()


/***************************************************************************/
/*!
 *  Constructor sets up the drop target and loads two icons for modules
 *  (valid and invalid)
 *
 ***************************************************************************/
activeProjectTree::activeProjectTree(wxWindow *parent, const wxWindowID id) :
wxTreeCtrl(parent,id,wxDefaultPosition,wxDefaultSize,
           wxSUNKEN_BORDER|wxTR_DEFAULT_STYLE|wxTR_MULTIPLE|wxTR_EDIT_LABELS),
           m_draggedItem(-1),
           m_changed(false),
           m_pendingParameterChanges(),
           m_markedItemColor(255,226,176)
{
    // Load icons
    wxImageList *imagelist = new wxImageList(16,16,true);

    int rootIconId         = imagelist->Add(wxIcon(modulecategory_xpm));
    int rootIconExpandedId = imagelist->Add(wxIcon(openmodulecategory_xpm));

    moduleItemData::m_iconInvalidId  = imagelist->Add(wxIcon(moduleiconnotconnected_xpm));
    moduleItemData::m_iconOkayId     = imagelist->Add(wxIcon(moduleicon_xpm));
    moduleItemData::m_iconInactiveId = imagelist->Add(wxIcon(moduleiconinactive_xpm));
    paramBoolItemData::m_iconNoId    = imagelist->Add(wxIcon(treeparamboolfalse_xpm));
    paramBoolItemData::m_iconYesId   = imagelist->Add(wxIcon(treeparambooltrue_xpm));
    paramStringItemData ::m_iconId   = imagelist->Add(wxIcon(treeparamstring_xpm));
    paramDoubleItemData ::m_iconId   = imagelist->Add(wxIcon(treeparamdouble_xpm));
    paramIntItemData    ::m_iconId   = imagelist->Add(wxIcon(treeparamint_xpm));
    singleChoiceItemData::m_iconId   = imagelist->Add(wxIcon(treeparamsinglechoice_xpm));
    inputPinItemData::m_iconId       = imagelist->Add(wxIcon(treeparaminputpin_xpm));
    fileNameItemData::m_iconId       = imagelist->Add(wxIcon(treeparamfilename_xpm));

    paramBoolItemData::m_yesString = _("Yes");
    paramBoolItemData::m_noString = _("No");

    AssignImageList(imagelist);

    // Make up dropTarget
    dropTarget = new moduleDropTarget(this);
    SetDropTarget(dropTarget);
    AddRoot(_("Active Project"),rootIconId,rootIconId,new rootItemData());
    SetItemImage(GetRootItem(),rootIconExpandedId,wxTreeItemIcon_Expanded);
    m_unmarkedItemColor = GetBackgroundColour();
}


/***************************************************************************/
/*!
 *  Restores all modules after letting the configManager checking their
 *  validity
 *
 ***************************************************************************/
void activeProjectTree::updateModules()
{
    // Save expanded states for later addition
    typedef map< string, bool> TExpandMap;
    TExpandMap expandedmap;

    wxTreeItemIdValue cookie=0;
    wxTreeItemId module = GetFirstChild(GetRootItem(),cookie);
    while(module.IsOk()){
        bool tmp = IsExpanded(module);
        expandedmap.insert(TExpandMap::value_type(GetItemText(module).c_str(),tmp));
        module=GetNextChild(GetRootItem(),cookie);
    }

    DeleteChildren(GetRootItem());
    m_mouseOverId=wxTreeItemId();
    configManager::getManager().setValidModules();

    TActiveModules modules = wxGetApp().getProcessor()->getModules();
    TActiveModules::iterator currentModule;

    wxString *moduleInstances = new wxString[modules.size()];
    wxString *currentModuleInstance = moduleInstances;

    for (currentModule =modules.begin();
        currentModule!=modules.end();
        currentModule++,currentModuleInstance++){

            moduleItemData *moduleData = new moduleItemData(moduleItemData::getConfigItem(*currentModule),this,*currentModule);
            wxTreeItemId newItem = AppendItem(GetRootItem(),(*currentModule)->getInstanceName().c_str(),
                -1,-1,moduleData);

            TExpandMap::iterator pos = expandedmap.find((*currentModule)->getInstanceName());

            moduleData->addParameters();

            if( pos != expandedmap.end() && pos->second )
                Expand(newItem);
        }

        delete [] moduleInstances;
        Expand(GetRootItem());

    ((frame *) wxGetApp().GetTopWindow())->updateStatusCount();
}

/***************************************************************************/
/*!
 *  When double clicking on a module, it gets configured
 *
 ***************************************************************************/
void activeProjectTree::onConfigureDoubleClick(wxTreeEvent &event)
{
    if(!wxGetApp().getProcessor()->isRunning())
        GETVISTALABDATA(event.GetItem())->doubleClick();
}


/***************************************************************************/
/*!
 *  Delete selected modules
 *
 ***************************************************************************/
void activeProjectTree::onChar(wxKeyEvent &event)
{
    wxArrayTreeItemIds selections;
    int selectionCount = GetSelections(selections);

    if (event.GetKeyCode() == WXK_DELETE)
    {
        std::list<module *> selectedModules;

        for(int currentSelection=selectionCount-1;currentSelection>=0;--currentSelection){

            moduleItemData *moduleData = dynamic_cast<moduleItemData *>
                (GETVISTALABDATA(selections[currentSelection]));

            if(moduleData)
                selectedModules.push_back((module *)moduleData->getModule());
        }

        std::list<module *>::iterator currentSelectedModule=selectedModules.begin();

        while(currentSelectedModule!=selectedModules.end()){
            wxGetApp().getProcessor()->removeModule(*currentSelectedModule);
            ++currentSelectedModule;
        }
        updateModules();
        setToBeSaved(true);

    } else if(selectionCount == 1){
        // See if selected module eats char. If so, treat the
        // pending changes appropriately
        if(GETVISTALABDATA(selections[0])->eatChar(event.GetKeyCode()) && tryApplyChanges())
            GETVISTALABDATA(selections[0])->update();



    }
//    event.Skip();
}

/***************************************************************************/
/*!
 *  Callback from DND - Target
 *
 ***************************************************************************/
void activeProjectTree::addModule(int x, int y, const wxString &data)
{
    // Get Index from position first
    int hitTestFlags = wxTREE_HITTEST_ONITEM;
    wxTreeItemId index = HitTest(wxPoint(x,y), hitTestFlags);
    module *predecessor = 0;
    if(index.IsOk()){
        moduleItemData *moduleData = dynamic_cast<moduleItemData *> GETVISTALABDATA(index);

        // If target item is no module, it may be either a parameter or the root module
        if(!moduleData && index != GetRootItem() ){
            moduleData = dynamic_cast<moduleItemData *> GETVISTALABDATA(GetItemParent(index));
        }

        if(moduleData)
            predecessor = (module *) moduleData->getModule();
    }

    moduleConfigDialog config;
    ((frame *)wxGetApp().GetTopWindow())->addModule(data.c_str(),
        config,predecessor);
}

/***************************************************************************/
/*!
 *  Checks if editing labels is okay
 *
 ***************************************************************************/
void activeProjectTree::onBeginLabelEdit(wxTreeEvent &event)
{
    if(!GETVISTALABDATA(event.GetItem())->isEditable())
        event.Veto();
}

/***************************************************************************/
/*!
 *  Checks if editing labels is okay
 *
 ***************************************************************************/
void activeProjectTree::onEndLabelEdit(wxTreeEvent &event)
{
    // For some strange reason, events are fired twice - the first time
    // with the new label and then another one which is empty
    if(event.IsEditCancelled())
        return;

    const char *errorText;

    if(event.GetLabel() == wxEmptyString)
    {
        event.Veto();
        errorText = _("Please enter a value");
        return;
    }

    errorText = ((parameterItemBase *)
        GETVISTALABDATA(event.GetItem()))->doneWithEditing(event.GetLabel());

    if(errorText != NULL){
        event.Veto();
        //Cannot fire a message box, as wxWidgets crashes!!!
        //wxMessageBox(errorText,_("Error"),wxOK|wxCENTRE);
        ((frame *)wxGetApp().GetTopWindow())->GetStatusBar()->SetStatusText(errorText);
        wxBell();
    }

    if(!tryApplyChanges())
        event.Veto();
}

void activeProjectTree::addPendingParameterChange(parameterItemBase *change)
{
    wxCriticalSectionLocker locker(m_pendingParameterLock);

    // Look if there is a pending change already for this parameter;
    // If not, add one
    TParameterChanges::iterator currentChange =
        find(m_pendingParameterChanges.begin(),
        m_pendingParameterChanges.end(),change);

    if(currentChange == m_pendingParameterChanges.end())
        m_pendingParameterChanges.push_back(change);
}

/***************************************************************************/
/*!
 *  Displays information in the status bar as soon as a target is hit
 *
 ***************************************************************************/
void activeProjectTree::onMouseMotion(wxMouseEvent &event)
{
    int hitTestFlags = wxTREE_HITTEST_ONITEM;
    wxTreeItemId index = HitTest(event.GetPosition(), hitTestFlags);

    if(index.IsOk()){
        if (index != m_mouseOverId){
            ((frame *)(wxGetApp().GetTopWindow()))->GetStatusBar()->
                SetStatusText(GETVISTALABDATA(index)->getToolTip());
            m_mouseOverId = index;
        }
    } else {
        ((frame *)(wxGetApp().GetTopWindow()))->GetStatusBar()->
            SetStatusText(wxEmptyString);
    }
    event.Skip();
}

const wxString activeProjectTree::applyPendingChanges(){
    wxCriticalSectionLocker locker(m_pendingParameterLock);

    wxString validationErrors;

    if(!m_pendingParameterChanges.empty()){

        // Save values in targetmodule so that validator
        // functions can begin validating
        for_each(m_pendingParameterChanges.begin(),m_pendingParameterChanges.end(),
            boost::bind(&parameterChangeBase::save,_1));

        // Encapsulate changed modules
        typedef std::vector< configurableObject * > TChangedModules;
        TChangedModules changedModules;

        TParameterChanges::iterator currentParameter=m_pendingParameterChanges.begin();
        while(currentParameter!=m_pendingParameterChanges.end()){

            if(find(changedModules.begin(),changedModules.end(),
                (*currentParameter)->getModule()) == changedModules.end())
                changedModules.push_back((*currentParameter)->getModule());
            ++currentParameter;
        }

        for_each(changedModules.begin(),changedModules.end(),
            boost::bind(&module::initChoices,_1));

        TChangedModules::iterator currentChangedModule=changedModules.begin();

        while(currentChangedModule != changedModules.end())
        {
            const TConfigItems *configItems = (*currentChangedModule)->getConfigItems();

            TConfigItems::const_iterator currentItem;

            bool okay = true;

            for(currentItem  = configItems->begin();
                currentItem != configItems->end();
                ++currentItem)
                if(currentItem->m_type == configItem::VALIDATORFUNC){
                    // We have a validator function here
                    const char *result = ((*currentChangedModule)->*((configItem::TValidatorFunc)
                        currentItem->m_target.func))();
                    if(result != NULL){
                        validationErrors+=wxString::Format(_("Error while changing parameters of %s:\n\n%s\n"),
                            ((module *)(*currentChangedModule))->getInstanceName().c_str(),
                            result);
                        okay=false;

                    }
                }

                if(!okay){
                    ++currentChangedModule;
                } else {
                    // This module can be delted from the list
                    currentChangedModule=changedModules.erase(currentChangedModule);
                }
        }

        // There are still modules in the list, i.e.
        // errors have occured.
        if(!changedModules.empty()){
            // Restore errors of modules
            currentChangedModule=changedModules.begin();
            while(currentChangedModule!=changedModules.end()){
                for(currentParameter=m_pendingParameterChanges.begin();
                    currentParameter!=m_pendingParameterChanges.end();
                    ++currentParameter)
                    if ((*currentParameter)->getModule() == (*currentChangedModule)){
                        (*currentParameter)->restore();
                    }
                    ++currentChangedModule;
            }
        }
    }

    markParameterChanges(false);
    m_pendingParameterChanges.clear();

    if(validationErrors == wxEmptyString)
        setToBeSaved(true);

    return validationErrors;
}

void activeProjectTree::updateModule(const wxTreeItemId &id)
{
    wxTreeItemIdValue cookie=0;
    GETVISTALABDATA(id)->update();
    wxTreeItemId child=GetFirstChild(id,cookie);

    while(child.IsOk()){
        GETVISTALABDATA(child)->update();
        child=GetNextChild(id,cookie);
    }
}

bool activeProjectTree::tryApplyChanges(){
    if(wxGetApp().getProcessor()->isRunning()){
        markParameterChanges(true);
    } else {
        // Only apply changes now if we are not running
        // Otherwise the processorCore will do it as soon
        // as there is time
        wxString errors = applyPendingChanges();

        if(errors != wxEmptyString){
            wxMessageBox(errors,_("Error"),wxOK|wxCENTRE);
            return false;
        }
    }
    return true;
}

void activeProjectTree::markParameterChanges(bool doMark)
{
    if(!doMark && !m_parameterChangesAreMarked)
        return;

    m_parameterChangesAreMarked=doMark;
    wxColour markcolor = doMark?m_markedItemColor:m_unmarkedItemColor;

    TParameterChanges::iterator currentParameter=m_pendingParameterChanges.begin();
    while(currentParameter!=m_pendingParameterChanges.end()){
        SetItemBackgroundColour((*currentParameter)->GetId(),markcolor);
        ++currentParameter;
    }
}

bool activeProjectTree::hasParameterChanges()
{
    return !m_pendingParameterChanges.empty();
}

void activeProjectTree::onRightClick(wxTreeEvent &event)
{
    if(!wxGetApp().getProcessor()->isRunning()){
        // Right mouseclick
        m_rightClickItem=event.GetItem();
        GETVISTALABDATA(m_rightClickItem)->rightClick(event.GetPoint());
    }
}


void activeProjectTree::onBeginDrag(wxTreeEvent &event)
{
    // need to explicitly allow drag
    if ( event.GetItem() != GetRootItem() && dynamic_cast< moduleItemData *> GETVISTALABDATA(event.GetItem()))
    {
        m_draggedItem = event.GetItem();
        event.Allow();
    }
}

void activeProjectTree::onEndDrag(wxTreeEvent &event)
{
    // Swap modules if first index is module and second index is either
    // module or not okay (which is the case if dropped below the bottom one)
    moduleItemData *draggedData = dynamic_cast< moduleItemData *> GETVISTALABDATA(m_draggedItem);
    moduleItemData *indexData   = 0;
    if(event.GetItem().IsOk())
        indexData = dynamic_cast< moduleItemData *> GETVISTALABDATA(event.GetItem());

    if( (!event.GetItem().IsOk() || indexData)){
        module *firstModule = (module *) draggedData->getModule();
        module *secondModule=0;

        if(event.GetItem().IsOk())
            secondModule = (module *) indexData->getModule();

        if( event.GetItem() != m_draggedItem ){
            wxGetApp().getProcessor()->swapModules(firstModule,secondModule);
            updateModules();
            setToBeSaved(true);
        }
    }
}

void activeProjectTree::onUpdateModules(wxCommandEvent &event)
{
    updateModules();
}
