/***************************************************************************
                 vlmoduleConfigdialog.cpp  -  description
                            -------------------
    begin                : Mon Feb 4 2002
    copyright            : (C) 2002 by Fabian Wenzel
    email                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include <algorithm>
#include <functional>
#include <boost/checked_delete.hpp>
#include <boost/bind.hpp>

#include "vlmoduleconfigdialog.h"
#include "vlapp.h"
#include "vlprocessor.h"

#include "vlmodule.h"
#include "vlconfigmanager.h"
#include "vlpin.h"

#include <wx/dialog.h>
#include <wx/msgdlg.h>
#include <wx/sizer.h>
#include <wx/button.h>
#include <wx/gdicmn.h>
#include <wx/intl.h>
#include <wx/textctrl.h>
#include <wx/valtext.h>
#include <wx/statbox.h>
#include <wx/filedlg.h>
#include <wx/checkbox.h>
#include <wx/combobox.h>
#include <wx/listbox.h>
#include <wx/utils.h>


enum{
    VALIDATE_ENTRIES=3500,
    SHOW_INFO
};

using namespace std;
using namespace vistaLab;


//////////////////// Baseclass for the config-dialog items //////////////////

// ConfigDialogItems get an own namespace for better sorting in KDevelop
namespace vistaLabDialogItems{

class configDialogItemBase : public parameterChangeBase {
public:
    configDialogItemBase(const configItem *item,
        configurableObject *targetModule):
    parameterChangeBase(item,targetModule),
    m_guiControl(0),
    m_panel(0),
    m_sizer(0)
    {
    }

    virtual void assemble()=0;

    wxPanel *getPanel()
    {
        return m_panel;
    }
protected:
    wxWindow *m_guiControl;
    wxPanel  *m_panel;
    wxSizer  *m_sizer;
};

template< class T >class parameterChangeForDialog : public parameterChange<T,configDialogItemBase>{
public:
    parameterChangeForDialog(const configItem *item,
        configurableObject *targetModule):
    parameterChange<T,configDialogItemBase>(item,targetModule)
    {
    }
};

/***************************************************************************/
//! Another middle-layer class which puts gui elements in a static box
/*!
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
template< class T >class configDialogItemWithStaticBox :
public parameterChangeForDialog< T >{
public:
    using parameterChangeForDialog< T >::m_guiControl;

    //! Assembles the gui element and the static box with the dialog sizer
    virtual void assemble()
    {
        this->m_sizer->Add(m_guiControl,wxSizerFlags(3).Center().Border(wxALL,10).Expand());
        this->m_sizer->Fit(this->m_panel);
        this->m_sizer->Layout();
    }

protected:
    //! Initializes the static box sizer
    configDialogItemWithStaticBox(const configItem *item,
        configurableObject *targetModule,
        wxWindow *parent,
        wxOrientation orientation = wxHORIZONTAL) :
    parameterChangeForDialog< T >(item,targetModule)
    {
        this->m_panel = new wxPanel(parent,-1);
        this->m_sizer = new wxStaticBoxSizer(new wxStaticBox(this->m_panel,-1,
            this->p_item->m_localizedDescription.c_str()),orientation);
        this->m_panel->SetSizer(this->m_sizer);
    }

};

//////////////////// child classes for gui - access ////////////////////



/***************************************************************************/
//! Child class for configurable string items
/*!
 *  A wxTextCtrl is used as the GUI element.
 *
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class configDialogString : public configDialogItemWithStaticBox< string >{
public:
    using parameterChangeForDialog< string >::m_guiControl;
    configDialogString(const configItem *item,
        configurableObject *targetModule,
        wxWindow *parent) :
    configDialogItemWithStaticBox< string >(item,targetModule,parent)
    {
        m_guiControl = new wxTextCtrl(m_panel, -1, target().c_str());
    }
    virtual void save()
    {
        store(((wxTextCtrl *) m_guiControl)->GetValue().c_str());
    }
};


/***************************************************************************/
//! Child class for configurable double-valued items
/*!
 *  A wxTextCtrl is used as the GUI element. A validator for numeric
 *  input is used as input filter, but the target string is not used
 *  as it gets updated during destruction of the GUI element; this is
 *  too late as it has to take place before validating.
 *
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class configDialogDouble : public configDialogItemWithStaticBox< double >{
public:
    using configDialogItemWithStaticBox< double >::m_guiControl;
    configDialogDouble(const configItem *item,
        configurableObject *targetModule,
        wxWindow *parent) :
    configDialogItemWithStaticBox< double >(item,targetModule,parent)
    {
        m_uselessString = wxString::Format("%2.9f",target());

        m_guiControl = new wxTextCtrl(m_panel, -1, m_uselessString,
            wxDefaultPosition, wxDefaultSize,0,
            wxTextValidator(wxFILTER_NUMERIC,&m_uselessString));
    }
    virtual void save()
    {
        wxString result = ((wxTextCtrl *) m_guiControl)->GetValue();
        double tmp;
        result.ToDouble(&tmp);
        store(tmp);
    }
private:
    wxString m_uselessString;
};


/***************************************************************************/
//! Child class for configurable integer-valued items
/*!
 *  A wxTextCtrl is used as the GUI element. A validator for numeric
 *  input is used as input filter, but the target string is not used
 *  as it gets updated during destruction of the GUI element; this is
 *  too late as it has to take place before validating.
 *
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class configDialogInt : public configDialogItemWithStaticBox< int >{
public:
    using configDialogItemWithStaticBox< int >::m_guiControl;
    configDialogInt(const configItem *item,
        configurableObject *targetModule,
        wxWindow *parent) :
    configDialogItemWithStaticBox< int >(item, targetModule,parent)
    {
        m_uselessString = wxString::Format("%d",target());
        m_guiControl = new wxTextCtrl(m_panel, -1, m_uselessString,
            wxDefaultPosition, wxDefaultSize, 0,
            wxTextValidator(wxFILTER_NUMERIC,&m_uselessString));
    }
    virtual void save()
    {
        wxString result = ((wxTextCtrl *) m_guiControl)->GetValue();
        long tmp;
        result.ToLong(&tmp);
        store(tmp);
    }
private:
    wxString m_uselessString;
};


/***************************************************************************/
//! Child class for configurable bool-valued items
/*!
 *  A wxCheckBox is used as the GUI element.
 *
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class configDialogBool : public parameterChangeForDialog< bool >{
public:
    using parameterChangeForDialog< bool >::m_guiControl;
    configDialogBool(const configItem *item,
        configurableObject *targetModule,
        wxWindow *parent) :
    parameterChangeForDialog< bool >(item, targetModule)
    {
        m_panel = new wxPanel(parent,-1);
        m_sizer = new wxBoxSizer(wxVERTICAL);
        m_panel->SetSizer(m_sizer);
        m_guiControl =
            new wxCheckBox(m_panel,-1,item->m_localizedDescription.c_str());

        ((wxCheckBox *)m_guiControl)->SetValue(target());
    }

    virtual void assemble()
    {
        this->m_sizer->Add(m_guiControl,wxSizerFlags(3).Center().Border(wxALL,10).Expand());
    }

    virtual void save()
    {
        store(((wxCheckBox *)m_guiControl)->GetValue());
    }
};


/***************************************************************************/
//! Child class for configurable fileNames
/*!
 *  The difference to configDialogString is that a Browsebutton is added
 *  to the GUI which may be used for file selection. This is why an event
 *  handler has to be used for event processing. A path string is used
 *  for keeping a starting point for successive Browse-clicks
 *
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class configDialogFilename : public wxEvtHandler,
    public configDialogString{
public:
    configDialogFilename(const configItem *item,
        configurableObject *targetModule,
        wxWindow *parent, bool save) :
    configDialogString(item, targetModule,parent),
    m_save(save)
    {
        parent->PushEventHandler(this);
        m_browseButton = new wxButton(m_panel,wxNewId(),
            _("Browse"));
        m_path = wxPathOnly(target().c_str());
    }
    virtual void assemble()
    {
        configDialogString::assemble();
        Connect(m_browseButton->GetId(),wxEVT_COMMAND_BUTTON_CLICKED,
            (wxObjectEventFunction) (wxCommandEventFunction)
            &configDialogFilename::FileDialog);
        m_sizer->Add(m_browseButton,wxSizerFlags(1).Right().Border(wxALL,10));
    }

private:
    wxButton *m_browseButton;
    void FileDialog (wxCommandEvent &event);
    wxString  m_path;
    bool      m_save;
};

//! Callback for browse-clicks
void configDialogFilename::FileDialog(wxCommandEvent &event)
{
    // We set up a wxSAVE File selector as it is not sure
    // if we are saving (entering filename possible) or
    // loading
    wxString fileName =  wxFileSelector(_("File"),m_path,
        wxEmptyString,wxEmptyString,"*.*",m_save?wxSAVE:wxOPEN);
    if(fileName != wxEmptyString){
        ((wxTextCtrl *)m_guiControl)->SetValue(fileName);
        m_path = wxPathOnly(fileName);
    }
}


/***************************************************************************/
//! Child class for configurable multiple fileNames
/*!
 *  A wxListBox is used as GUI element - and several buttons enable
 *  adding, removing and moving of fileNames
 *
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class configDialogMultipleFilenames :
    public wxEvtHandler,
    public configDialogItemWithStaticBox< vector<string> >{
public:
    configDialogMultipleFilenames(const configItem *item,
        configurableObject *targetModule,
        wxWindow *parent) :
    configDialogItemWithStaticBox< vector< string > >(item,
        targetModule,parent,wxVERTICAL)
    {
        parent->PushEventHandler(this);
        unsigned int numberOfEntries = target().size();

        wxString  *listBoxEntries = new wxString[numberOfEntries];

        for(unsigned int i=0;i<numberOfEntries;i++)
            listBoxEntries[i]=target()[i].c_str();

        m_guiControl = new wxListBox(m_panel,-1, wxDefaultPosition,
            wxSize(450,200), numberOfEntries,
            listBoxEntries,
            wxLB_NEEDED_SB | wxLB_EXTENDED );
        delete [] listBoxEntries;
        m_add    = new wxButton(m_panel,wxNewId(),_("Add file"));
        m_remove = new wxButton(m_panel,wxNewId(),_("Remove file"));
        m_up     = new wxButton(m_panel,wxNewId(),_("Move up"));
        m_down   = new wxButton(m_panel,wxNewId(),_("Move down"));
    }

    virtual void assemble(){
        m_sizer->Add(m_guiControl,wxSizerFlags(1).Center().Border(wxALL,10));

        wxBoxSizer *addRemoveSizer = new wxBoxSizer(wxHORIZONTAL);
        // Add buttons
        addRemoveSizer->Add(m_add,   wxSizerFlags(0).Center().Border(wxALL,5));
        addRemoveSizer->Add(m_remove,wxSizerFlags(0).Center().Border(wxALL,5));
        addRemoveSizer->Add(m_up,    wxSizerFlags(0).Center().Border(wxALL,5));
        addRemoveSizer->Add(m_down,  wxSizerFlags(0).Center().Border(wxALL,5));

        Connect(m_add   ->GetId(),wxEVT_COMMAND_BUTTON_CLICKED,
            (wxObjectEventFunction) (wxCommandEventFunction)
            &configDialogMultipleFilenames::onButtonAdd);

        Connect(m_remove->GetId(),wxEVT_COMMAND_BUTTON_CLICKED,
            (wxObjectEventFunction) (wxCommandEventFunction)
            &configDialogMultipleFilenames::onButtonRemove);

        Connect(m_up    ->GetId(),wxEVT_COMMAND_BUTTON_CLICKED,
            (wxObjectEventFunction) (wxCommandEventFunction)
            &configDialogMultipleFilenames::onButtonUp);

        Connect(m_down  ->GetId(),wxEVT_COMMAND_BUTTON_CLICKED,
            (wxObjectEventFunction) (wxCommandEventFunction)
            &configDialogMultipleFilenames::onButtonDown);

        m_sizer->Add(addRemoveSizer,wxSizerFlags(0).Center().Border(wxALL,5));
    }
    virtual void save()
    {
        vector< string > files;
        files.clear();
        int i;
        for(i=0;i<((wxListBox *) m_guiControl)->GetCount();i++)
            files.push_back(((wxListBox *) m_guiControl)->GetString(i).c_str());
        store(files);
    }

private:
    void onButtonAdd(wxCommandEvent &event);
    void onButtonRemove(wxCommandEvent &event);
    void onButtonUp(wxCommandEvent &event);
    void onButtonDown(wxCommandEvent &event);
    wxButton *m_add;
    wxButton *m_remove;
    wxButton *m_up;
    wxButton *m_down;
    wxString  m_path;
};


//! Callback for add-clicks
void configDialogMultipleFilenames::onButtonAdd(wxCommandEvent &event)
{
    wxFileDialog dialog(NULL, _T("Open File(s)"),
        m_path, wxEmptyString, wxFileSelectorDefaultWildcardStr,
        wxMULTIPLE|wxOPEN);
    if (dialog.ShowModal() == wxID_OK)
    {
        wxArrayString paths;

        dialog.GetPaths(paths);

        size_t count = paths.GetCount();

        if(count > 0){
            m_path = wxPathOnly(paths[0]);
            for ( size_t n = 0; n < count; n++ )
                ((wxListBox *) m_guiControl)->Append(paths[n]);
        }

    }
}

//! Callback for remove-clicks
void configDialogMultipleFilenames::onButtonRemove(wxCommandEvent &event)
{
    wxArrayInt selections;
    ((wxListBox *) m_guiControl)->GetSelections(selections);

    for(unsigned int i=0;i<selections.GetCount();++i)
        ((wxListBox *) m_guiControl)->Delete(selections[i]-i);
}

//! Callback for up-clicks
void configDialogMultipleFilenames::onButtonUp(wxCommandEvent &event)
{
    wxArrayInt selections;
    ((wxListBox *) m_guiControl)->GetSelections(selections);
    if(selections.GetCount() > 1)
    {
        wxMessageBox(_("Can only move one file at a time"));
        return;
    }
    int selectionIndex = selections[0];
    if ( selectionIndex > 0 ){

        wxString selection = ((wxListBox *) m_guiControl)->GetString(selectionIndex);
        ((wxListBox *) m_guiControl)->InsertItems(1,& selection,selectionIndex-1);
        ((wxListBox *) m_guiControl)->Delete(selectionIndex+1);
        ((wxListBox *) m_guiControl)->SetSelection(selectionIndex-1);
    }
}

//! Callback for down-clicks
void configDialogMultipleFilenames::onButtonDown(wxCommandEvent &event)
{
    wxArrayInt selections;
    ((wxListBox *) m_guiControl)->GetSelections(selections);
    if(selections.GetCount() > 1)
    {
        wxMessageBox(_("Can only move one file at a time"));
        return;
    }
    int selectionIndex = selections[0];
    if ( selectionIndex < ((wxListBox *) m_guiControl)->GetCount() -1){
        wxString selection = ((wxListBox *) m_guiControl)->GetString(selectionIndex);
        ((wxListBox *) m_guiControl)->InsertItems(1,&selection,selectionIndex+2 );
        ((wxListBox *) m_guiControl)->Delete(selectionIndex);
        ((wxListBox *) m_guiControl)->SetSelection(selectionIndex+1);
    }
}


/***************************************************************************/
//! Child class for configurable single choices
/*!
 *  A wxComboBox is used as GUI element
 *
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class configDialogSingleChoice :
    public configDialogItemWithStaticBox< int >{
public:
    configDialogSingleChoice(const configItem *item,
        configurableObject *targetModule,
        wxWindow *parent) :
    configDialogItemWithStaticBox< int >(item, targetModule,parent)
    {
        int numberOfSelections = item->m_selections.size();
        wxString *selectionStrings = new wxString[numberOfSelections];
        wxString *startSelection=0;
        for(unsigned int i=0;i < (unsigned int) numberOfSelections;i++){
            selectionStrings[i]=item->m_selections[i].c_str();
            if((unsigned int) target()==i)
                startSelection=&selectionStrings[i];
        }

        // Target contains an invalid choice, so take the first one
        if(!startSelection)
            startSelection = &selectionStrings[0];

        m_guiControl = new wxComboBox(m_panel, -1, *startSelection,
            wxDefaultPosition, wxDefaultSize, numberOfSelections,
            selectionStrings, wxCB_READONLY);

        delete [] selectionStrings;
    }
    virtual void save()
    {
        int selection = ((wxComboBox *) m_guiControl)->GetSelection();
        if ( selection != -1 )
            store(selection);
    }
    };


/***************************************************************************/
//! Child class for configurable multiple choices
/*!
 *  A wxCheckListBox is used as GUI element
 *
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class configDialogMultipleChoices :
    public configDialogItemWithStaticBox< vector< int > >{
public:
    configDialogMultipleChoices(const configItem *item,
        configurableObject *targetModule,
        wxWindow *parent) :
    configDialogItemWithStaticBox< vector< int > >(item,
        targetModule,parent,wxVERTICAL)
    {
        unsigned int numberOfSelections = item->m_selections.size();
        wxString *selectionStrings = new wxString[numberOfSelections];
        for( unsigned int i=0;i<numberOfSelections;i++)
            selectionStrings[i] = item->m_selections[i].c_str();

        m_guiControl = new wxCheckListBox(m_panel,-1, wxDefaultPosition,wxDefaultSize,
            numberOfSelections, selectionStrings,
            wxLB_NEEDED_SB);

        vector< int >::const_iterator currentItem;

        for(currentItem  = target().begin();
            currentItem!= target().end();
            ++currentItem){
            if((unsigned int) *currentItem < item->m_selections.size())
                ((wxCheckListBox *) m_guiControl)->Check(*currentItem);
        }

        delete [] selectionStrings;
    }
    void save()
    {
        vector< int > results;

        for(int i=0;i<((wxCheckListBox *) m_guiControl)->GetCount();++i)
        {
            if(((wxCheckListBox *) m_guiControl)->IsChecked(i))
                results.push_back(i);
        }

        store(results);
    }
    };


/***************************************************************************/
//! Child class for a custom single choice
/*!
 *  A non-read - only wxComboBox is used as GUI element
 *
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class configDialogCustomChoice :
    public configDialogItemWithStaticBox< string > {
public:
    configDialogCustomChoice(const configItem *item,
        configurableObject *targetModule,
        wxWindow *parent) :
    configDialogItemWithStaticBox< string >(item,targetModule,parent)
    {
        int numberOfSelections = item->m_selections.size();
        wxString *selections = new wxString[numberOfSelections];

        for(unsigned int i=0;i < (unsigned int) numberOfSelections;i++)
            selections[i]=item->m_selections[i].c_str();

        m_guiControl = new wxComboBox(m_panel, -1, target().c_str(),
            wxDefaultPosition, wxDefaultSize,
            numberOfSelections, selections);
        delete [] selections;
    }
    void save()
    {
        store( ((wxComboBox * ) m_guiControl)->GetValue().c_str() );
    }
    };


/***************************************************************************/
//! Child class for an input pin
/*!
 *  A read-only wxComboBox is used as GUI element
 *
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class configDialogInputPin : public configDialogItemWithStaticBox< int > {
public:
    configDialogInputPin(const configItem *item,
        configurableObject *targetModule,
        wxWindow *parent) :
    configDialogItemWithStaticBox< int >(item,targetModule,parent)
    {
        m_outputPins = configManager::getManager().
            getMatchingOutputPins(item->m_selections.front(),targetModule);
        wxString *selections = new wxString[m_outputPins.size()+1];

        configManager::TOutputPins::iterator currentPin;

        selections[0]=_("Not connected");

        int choice=m_outputPins.size();
        unsigned int i;

        for(i=1,currentPin=m_outputPins.begin();
            currentPin != m_outputPins.end();
            ++currentPin,++i){
                selections[i]=wxString::Format("%s of %s",
                    _((*currentPin)->m_description.c_str()),
                    (*currentPin)->m_moduleInstance->getInstanceName().c_str());

                if((*currentPin)->m_slot == target())
                    choice = i;
            }

            m_guiControl = new wxComboBox(m_panel, -1, wxEmptyString,
                wxDefaultPosition, wxDefaultSize,
                m_outputPins.size()+1, selections, wxCB_READONLY);

            ((wxComboBox *)m_guiControl)->SetSelection(choice);

            delete [] selections;
    }
    void save()
    {
        int selection = ((wxComboBox *) m_guiControl)->GetSelection();
        if ( selection <= 0 )
            store(-1);
        else if (selection > 0){
            store(m_outputPins[selection-1]->m_slot);
        }
    }
private:
    configManager::TOutputPins m_outputPins;
};


/***************************************************************************/
//! Child class for a render input pin
/*!
 *  A wxCheckListBox is used as GUI element as this is a multiple choice
 *
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class configDialogRenderInputPin :
    public configDialogItemWithStaticBox< vector< int > > {
public:
    configDialogRenderInputPin(const configItem *item,
        configurableObject *targetModule,
        wxWindow *parent) :
    configDialogItemWithStaticBox< vector< int > >(item,targetModule,parent)
    {
        m_renderPins = configManager::getManager().getRenderPins();

        wxString *selections = new wxString[m_renderPins.size()];

        configManager::TRenderPins::iterator currentPin;
        unsigned int i;

        for(i=0,currentPin=m_renderPins.begin();
            currentPin != m_renderPins.end();
            ++currentPin,++i)
            selections[i]=wxString::Format("%s of %s",
            _((*currentPin)->m_description.c_str()),
            (*currentPin)->m_moduleInstance->getInstanceName().c_str());

        m_guiControl = new wxCheckListBox(m_panel,-1, wxDefaultPosition,wxDefaultSize,
            m_renderPins.size(),selections,wxLB_NEEDED_SB);

        vector< int >::const_iterator currentItem;

        for(currentItem  = target().begin();
            currentItem != target().end();
            ++currentItem){
            ((wxCheckListBox *) m_guiControl)->Check(*currentItem);
            ++i;
        }

        delete [] selections;
    }
    void save()
    {
        vector< int > results;

        for(int i=0;i<((wxCheckListBox *) m_guiControl)->GetCount();++i)
        {
            if(((wxCheckListBox *) m_guiControl)->IsChecked(i))
                results.push_back(i);
        }

        store(results);
    }
private:
    configManager::TRenderPins m_renderPins;
};


/***************************************************************************/
//! Child class for a log input pin
/*!
 *  A wxCheckListBox is used as GUI element as this is a multiple choice
 *
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class configDialogLogInputPin :
    public configDialogItemWithStaticBox< vector< int > > {
public:
    configDialogLogInputPin(const configItem *item,
        configurableObject *targetModule,
        wxWindow *parent) :
    configDialogItemWithStaticBox< vector< int > >(item,targetModule,parent)
    {
        m_logPins = configManager::getManager().getLogPins();
        wxString *selections = new wxString[m_logPins.size()];

        configManager::TLogPins::iterator currentPin;
        unsigned int i;

        for(i=0,currentPin=m_logPins.begin();
            currentPin != m_logPins.end();
            ++currentPin,++i)
            selections[i]=wxString::Format("%s of %s",
            _((*currentPin)->m_description.c_str()),
            (*currentPin)->m_moduleInstance->getInstanceName().c_str());

        m_guiControl = new wxCheckListBox(m_panel,-1, wxDefaultPosition,wxDefaultSize,
            m_logPins.size(),selections,wxLB_NEEDED_SB);

        vector< int >::const_iterator currentItem;
        for(currentItem  = target().begin();
            currentItem != target().end();
            ++currentItem)
            ((wxCheckListBox *) m_guiControl)->Check(*currentItem);

        delete [] selections;
    }
    void save()
    {
        vector< int > results;

        for(int i=0;i<((wxCheckListBox *) m_guiControl)->GetCount();++i)
        {
            if(((wxCheckListBox *) m_guiControl)->IsChecked(i))
                results.push_back(i);
        }

        store(results);
    }
private:
    configManager::TLogPins m_logPins;
};

}


using namespace vistaLabDialogItems;

//////////////////// Public methods for the class ///////////////////////////



/***************************************************************************/
/*!
 *  Constructor sets up sizer into which new items will be added later.
 *  Also the id-counter for items which need a wxWindow-ID gets initialized.
 *  The dialog is a member of this implementation class
 *
 ***************************************************************************/
moduleConfigDialog::moduleConfigDialog() :
m_hasPins    (false)
{
}

void moduleConfigDialog::attachModule(configurableObject *module)
{
    moduleConfig::attachModule(module);
    m_dialog = new wxDialog(wxGetApp().GetTopWindow(), -1,
        _(module->getName().c_str()), wxDefaultPosition,
        wxSize(400,400), wxDEFAULT_DIALOG_STYLE | wxDIALOG_MODAL);

    m_dialogSizer = new wxBoxSizer(wxVERTICAL);
    m_dialog->PushEventHandler(this);
    Connect(-1,wxEVT_CHAR_HOOK,(wxObjectEventFunction) (wxCharEventFunction)
        &moduleConfigDialog::onCharHook);
    m_dialogSizer->Add(0,20,0,wxEXPAND);
    m_notebook = new wxNotebook(m_dialog,-1);
}

/***************************************************************************/
/*!
 *  Destructor deletes all the members of all the lists which have
 *  accumulated so far
 *
 ***************************************************************************/
moduleConfigDialog::~moduleConfigDialog()
{
    delete m_dialog;
    for_each(m_dialogItems.begin(),m_dialogItems.end(),
        boost::checked_deleter< parameterChangeBase >());
}


/***************************************************************************/
/*!
 *  Finished setting up the dialog and shows it. The OK - Button will
 *  not send a wxID_OKAY - event but it will lead to validating the
 *  changed contents (using save() before) and eventually restoring
 *  back the original data using restore()
 *
 ***************************************************************************/
bool moduleConfigDialog::configure()
{
    bool returnValue = false;

    addConfigItems();

    // Check if there is only one panel
    if(m_dialogPanels.size() > 1){
        //Reparent panels to notebook

        int counter=1;
        std::vector< wxPanel * >::iterator currentPanel;
        for(currentPanel = m_dialogPanels.begin();
            currentPanel != m_dialogPanels.end();
            ++currentPanel){
                m_notebook->AddPage(*currentPanel,wxString::Format(_("Parameters %d"),counter++));
            }
            m_dialogSizer->Add(m_notebook,wxSizerFlags(0).Center().Border(wxALL,10).Expand());
    } else {
        m_dialogSizer->Add(m_dialogPanels.front(),wxSizerFlags(0).Center().Border(wxALL,10).Expand());
        m_notebook->Destroy();
    }

    // Finish dialog by adding OK, CANCEL and "OUTPUTPIN" - buttons
    wxBoxSizer *buttonSizer = new wxBoxSizer(wxHORIZONTAL);
    m_wantsFocus = new wxButton (m_dialog, VALIDATE_ENTRIES, _("OK"));

    buttonSizer->Add(m_wantsFocus,wxSizerFlags(0).Border(wxALL,10));
    Connect(VALIDATE_ENTRIES,wxEVT_COMMAND_BUTTON_CLICKED,
        (wxObjectEventFunction) (wxCommandEventFunction) &moduleConfigDialog::validate);
    buttonSizer->Add(new wxButton (m_dialog, wxID_CANCEL, _("Cancel")),
        wxSizerFlags(0).Border(wxALL,10));
    if(m_hasPins)
        buttonSizer->Add(new wxButton (m_dialog, SHOW_INFO, _("Module Info")),
        wxSizerFlags(0).Border(wxALL,10));

    Connect(SHOW_INFO,wxEVT_COMMAND_BUTTON_CLICKED,
        (wxObjectEventFunction) (wxCommandEventFunction) &moduleConfigDialog::showInfo);

    Connect(-1,wxEVT_IDLE,
        (wxObjectEventFunction) (wxIdleEventFunction) &moduleConfigDialog::onIdle);

    m_dialogSizer->Add(buttonSizer,wxSizerFlags(0).Center().Border(wxALL,10));

    // Set up layout stuff
    m_dialog->SetSizer(m_dialogSizer);
    m_dialogSizer->SetSizeHints(m_dialog);
    m_dialogSizer->Fit(m_dialog);

    m_dialogSizer->Layout();

    m_dialog->Centre();

    // Show it
    if(m_dialog->ShowModal() == wxID_OK)
        returnValue=true;
    else
        restore();

    m_targetModule->postConfig();
    return returnValue;
}



/***************************************************************************/
/*!
 *  Restores all data using the backup - fields in the members
 *
 ***************************************************************************/
void moduleConfigDialog::restore()
{
    for_each(m_dialogItems.begin(),m_dialogItems.end(),
        boost::bind(&parameterChangeBase::restore,_1));
    m_targetModule->initChoices();
}


/***************************************************************************/
/*!
 *  Validates inputs by calling the validator functions. If validation
 *  succeeds, a wxID_OK - event is emitted
 *
 ***************************************************************************/
void moduleConfigDialog::validate(wxCommandEvent &event)
{

    // Save values in targetModule so that validator
    // functions can begin validating
    for_each(m_dialogItems.begin(),m_dialogItems.end(),
        boost::bind(&parameterChangeBase::save,_1));

    m_targetModule->initChoices();

    const char *result=0;

    const TConfigItems *configItems = m_targetModule->getConfigItems();

    TConfigItems::const_iterator currentItem;

    for(currentItem  = configItems->begin();
        currentItem != configItems->end() && result == 0;
        ++currentItem)
        if(currentItem->m_type == configItem::VALIDATORFUNC)
            // We have a validator function here
            result = (m_targetModule->*((configItem::TValidatorFunc)
            currentItem->m_target.func))();

    if(result != 0){
        // show error and do nothing
        wxLogError(result);
    } else {
        // check instance name and send okay event if we have a module
        vistaLab::module *module = dynamic_cast< vistaLab::module *> (m_targetModule);
        if(!module || wxGetApp().getProcessor()->checkInstanceName(module)){
            wxCommandEvent event(wxEVT_COMMAND_BUTTON_CLICKED,wxID_OK);
            m_dialog->GetEventHandler()->AddPendingEvent(event);
        }
    }
}


/***************************************************************************/
/*!
 *  If Return has been pressed, simulate OK-Button-Click
 *
 ***************************************************************************/
void moduleConfigDialog::onCharHook  (wxKeyEvent    &event)
{
    if(event.GetKeyCode() == WXK_RETURN){
        wxCommandEvent event(wxEVT_COMMAND_BUTTON_CLICKED,VALIDATE_ENTRIES);
        m_dialog->GetEventHandler()->AddPendingEvent(event);
    } else
        event.Skip();
}


/***************************************************************************/
/*!
 *  For setting the Focus to the OK button (which has to be done in IDLE 
 *  time according to usenet)
 *
 ***************************************************************************/
void moduleConfigDialog::onIdle       (wxIdleEvent   &event)
{
    if(m_wantsFocus){
        m_wantsFocus->SetFocus();
        m_wantsFocus=0;
    }
}


/***************************************************************************/
/*!
 *  adds all configurable items
 *
 ***************************************************************************/
void moduleConfigDialog::addConfigItems()
{
    int disp_x,disp_y;
    wxDisplaySize(&disp_x,&disp_y);

    m_dialogPanels.clear();
    wxWindow *parent = m_dialog;

    wxPanel    *currentPanel = new wxPanel(parent,-1);
    wxBoxSizer *currentSizer = new wxBoxSizer(wxVERTICAL);
    currentPanel->SetSizer(currentSizer);

    m_dialogPanels.push_back(currentPanel);

    // Fetch configurable items for this module
    const TConfigItems *configItems = m_targetModule->getConfigItems();

    TConfigItems::const_iterator currentItem;

    for(currentItem  = configItems->begin();
        currentItem != configItems->end();
        ++currentItem){
            vistaLabDialogItems::configDialogItemBase *newItem=0;

            switch(currentItem->m_type){

           case     configItem::STRING :
               newItem = new configDialogString(&*currentItem,
                   m_targetModule, currentPanel);
               break;

           case configItem::FILENAME:
               newItem = new configDialogFilename(&*currentItem,
                   m_targetModule, currentPanel,false);
               break;

           case configItem::SAVEFILENAME:
               newItem = new configDialogFilename(&*currentItem,
                   m_targetModule, currentPanel,true);
               break;

           case configItem::INT:
               newItem = new configDialogInt(&*currentItem,
                   m_targetModule, currentPanel);
               break;

           case configItem::MULTIPLEFILENAMES:
               newItem = new configDialogMultipleFilenames(
                   &*currentItem,m_targetModule, currentPanel);
               break;

           case configItem::BOOL:
               newItem = new configDialogBool(&*currentItem,
                   m_targetModule,currentPanel);
               break;

           case configItem::DOUBLE:
               newItem = new configDialogDouble(&*currentItem,
                   m_targetModule,currentPanel);
               break;

           case configItem::SINGLECHOICE:
               newItem = new configDialogSingleChoice(&*currentItem,
                   m_targetModule,currentPanel);
               break;

           case configItem::MULTIPLECHOICE:
               newItem = new configDialogMultipleChoices(&*currentItem,
                   m_targetModule,currentPanel);
               break;

           case configItem::CUSTOMCHOICE:
               newItem = new configDialogCustomChoice(&*currentItem,
                   m_targetModule,currentPanel);
               break;

           case configItem::INPUTPIN:
               newItem = new configDialogInputPin(&*currentItem,
                   m_targetModule,currentPanel);
               m_hasPins = true;
               break;

           case configItem::RENDERINPUTPIN:
               newItem = new configDialogRenderInputPin(&*currentItem,
                   m_targetModule,currentPanel);
               m_hasPins = true;
               break;

           case configItem::LOGINPUTPIN:
               newItem = new configDialogLogInputPin(&*currentItem,
                   m_targetModule,currentPanel);
               m_hasPins = true;
               break;

           case configItem::OUTPUTPIN:
               m_hasPins         = true;

           default:
               break;
            }

            if(newItem){
                // Add new item
                newItem->assemble();
                currentSizer->Add(newItem->getPanel(),wxSizerFlags(0).Center().Border(wxALL,10).Expand());

                //Test if size is still adequate
                currentSizer->Fit(currentPanel);
                currentSizer->SetSizeHints(currentPanel);
                currentSizer->Layout();

                // If not, make new panel and place current item on it
                if (currentSizer->GetSize().y > 0.6*disp_y){
                    // GTK seems to have problem with Reparent(), so this hack
                    // rewinds the whole story and creates everything anew
                    if(m_dialogPanels.size() == 1 && m_dialogPanels.front()->GetParent() == m_dialog){
                        m_dialogPanels.front()->Destroy();
                        m_dialogPanels.clear();

                        parent = m_notebook;

                        // Throw away all dialog items
                        for_each(m_dialogItems.begin(),m_dialogItems.end(),
                            boost::checked_deleter< vistaLabDialogItems::configDialogItemBase >());
                        m_dialogItems.clear();

                        currentItem  = configItems->begin();
                    } else {
                        currentSizer->Detach(newItem->getPanel());
                        newItem->getPanel()->Destroy();
                    }

                    currentPanel = new wxPanel(parent,-1);
                    currentSizer = new wxBoxSizer(wxVERTICAL);
                    currentPanel->SetSizer(currentSizer);

                    m_dialogPanels.push_back(currentPanel);
                    --currentItem;
                } else {
                    //This time, the item is placed well, so we can store the item
                    //as it won't be deleted
                    m_dialogItems.push_back(newItem);
                }
            }
        }
}


/***************************************************************************/
/*!
 *  Shows pins of the current module
 *
 ***************************************************************************/
void moduleConfigDialog::showInfo(wxCommandEvent &event)
{
    // Fetch configurable items for this module
    const TConfigItems *configItems = m_targetModule->getConfigItems();

    TConfigItems::const_iterator currentItem;

    bool foundPin=false;

    wxString outputString=_("This module has the following Pins:");

    // First: Input Items
    for(currentItem  = configItems->begin();
        currentItem != configItems->end();
        ++currentItem)
        if(currentItem->m_type == configItem::INPUTPIN){
            if(!foundPin){
                foundPin=true;
                outputString+=_("\n\nInput-Pins\n==========\n");
            }

            outputString+=wxString::Format("%s (type %s)\n",
                currentItem->m_localizedDescription.c_str(),
                currentItem->m_selections.front().c_str());
        }

    // First: Input render
    for(currentItem  = configItems->begin();
        currentItem != configItems->end();
        ++currentItem)
        if(currentItem->m_type == configItem::RENDERINPUTPIN){
            if(!foundPin){
                foundPin=true;
                outputString+=_("\n\nRenderinput-Pins\n==========\n");
            }

            outputString+=wxString::Format("%s \n",
                currentItem->m_localizedDescription.c_str());
        }

    // First: Log Input
    for(currentItem  = configItems->begin();
        currentItem != configItems->end();
        ++currentItem)
        if(currentItem->m_type == configItem::LOGINPUTPIN){
            if(!foundPin){
                foundPin=true;
                outputString+=_("\n\nLoginput-Pins\n==========\n");
            }

            outputString+=wxString::Format("%s \n",
                currentItem->m_localizedDescription.c_str());
        }


    foundPin=false;

    // Now: Output Pins
    for(currentItem  = configItems->begin();
        currentItem != configItems->end();
        ++currentItem)
        if(currentItem->m_type == configItem::OUTPUTPIN){
            if(!foundPin){
                foundPin=true;
                outputString+=_("\n\nOutput-Pins\n===========\n");
            }
            outputString+=wxString::Format("%s (type %s)\n",
                currentItem->m_localizedDescription.c_str(),
                currentItem->m_selections.front().c_str());
        }

    foundPin=false;

    // Now: Render Pins
    for(currentItem  = configItems->begin();
        currentItem != configItems->end();
        ++currentItem)
        if(currentItem->m_type == configItem::RENDEROUTPUTPIN){
            if(!foundPin){
                foundPin=true;
                outputString+=_("\n\nRenderoutput-Pins\n===========\n");
            }
            outputString+=wxString::Format("%s \n",
                currentItem->m_localizedDescription.c_str());
        }

    foundPin=false;

    // Now: Log Pins
    for(currentItem  = configItems->begin();
        currentItem != configItems->end();
        ++currentItem)
        if(currentItem->m_type == configItem::LOGOUTPUTPIN){
            if(!foundPin){
                foundPin=true;
                outputString+=_("\n\nLogoutput-Pins\n===========\n");
            }
            outputString+=wxString::Format("%s",
                currentItem->m_localizedDescription.c_str());
        }

    outputString+="\n\n";
    outputString+=_("Author: ")+wxString(((vistaLab::module *)m_targetModule)->getAuthorInfo().c_str());

    wxMessageBox(outputString,_("Module info"),
        wxICON_INFORMATION|wxCENTRE|wxOK);
}
