/***************************************************************************
                     vlmoduleconfigfile.cpp  -  description
                            -------------------
    begin                : Mon Feb 4 2002
    copyright            : (C) 2002 by Fabian Wenzel
    email                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include <wx/config.h>
#include <wx/fileconf.h>
#include <wx/tokenzr.h>

#include "vlmoduleconfigfile.h"
#include "vlconfigmanager.h"
#include "vlapp.h"
#include "vlprocessor.h"
#include "vlconfigitem.h"
#include "vlpin.h"

#include <algorithm>
#include <boost/checked_delete.hpp>

using namespace std;
using namespace vistaLab;


//////////////////// Baseclass for the config-dialog items //////////////////

// ConfigDialogItems get an own namespace for better sorting in KDevelop
namespace vistaLabFileItems{


    /***************************************************************************/
    //! Abstract Base File Item Class
    /*!
     *  \author  Fabian Wenzel
     *
     ***************************************************************************/
    class configFileItemBase{
    public:
        virtual void save()=0;
        virtual bool read()=0;
        virtual ~configFileItemBase() {}

    protected:
        configFileItemBase(const configItem *item,
            configurableObject *targetModule,
            wxFileConfig *config) :
        m_item(item),
            m_targetModule(targetModule),
            p_config(config)
        {
        }

        const configItem              *m_item;
        const configurableObject *m_targetModule;
        wxFileConfig                   *p_config;
    };




    /***************************************************************************/
    /*
     *  Template Traits for converting int to long such that configurable ints
     *  can also be used with templates
     *
     ***************************************************************************/
    template< class T > struct configTraits{
        typedef T    TConfigReturn;
    };

    template <> struct configTraits< int >{
        typedef long TConfigReturn;
    };

    /***************************************************************************/
    //! Template Child file Item middle-layer for matching wxConfig-calls
    /*!
     *  \author  Fabian Wenzel
     *
     ***************************************************************************/
    template< class T > class configFileItem : public configFileItemBase{
    public:
        configFileItem(const configItem *item,
            configurableObject *targetModule,
            wxFileConfig *config) :
        configFileItemBase(item,targetModule,config) {}
    protected:
        T &target() const
        {
            return *((T *) m_item->m_target.data);
        }
        virtual void save()
        {
            typename configTraits<T>::TConfigReturn tmp = target();
            p_config->Write(m_item->m_description.c_str(), tmp);
        }
        virtual bool read()
        {
            typename configTraits<T>::TConfigReturn tmp;
            bool readSuccess = p_config->Read(m_item->m_description.c_str(), &tmp);
            if( readSuccess )
                target() = tmp;
            return readSuccess;
        }
    };


    /***************************************************************************/
    //! Child for configurable string items.
    /*!
     *  They cannot be used with templates as there is no implicit conversion
     *  between std::string and wxString
     *
     *  \author  Fabian Wenzel
     *
     ***************************************************************************/
    class configFileString : public configFileItemBase{
    public:
        configFileString(const configItem *item,
            configurableObject *targetModule,
            wxFileConfig *config) :
        configFileItemBase(item,targetModule,config) {}

        virtual void save()
        {
            p_config->Write(m_item->m_description.c_str(), target().c_str());
        }
        virtual bool read()
        {
            wxString tmp;
            bool readSuccess = p_config->Read(m_item->m_description.c_str(), &tmp);
            if( readSuccess )
                target() = tmp.c_str();
            return readSuccess;
        }
    private:
        string &target() const
        {
            return *((string *) m_item->m_target.data);
        }
    };


    /***************************************************************************/
    //! Child for configurable input pins
    /*!
     *
     *  \author  Fabian Wenzel
     *
     ***************************************************************************/
    class configFileInputPin : public configFileItemBase{
    public:
        configFileInputPin(const configItem *item,
            configurableObject *targetModule,
            wxFileConfig *config) :
        configFileItemBase(item,targetModule,config) {}

        virtual void save()
        {
            outputPin *connectedPin =
                configManager::getManager().getOutputPin(*( (int *) m_item->m_target.data));

            if(connectedPin){
                p_config->Write(m_item->m_description.c_str(),
                    wxString("1::")+
                    wxString(connectedPin->m_moduleInstance->getInstanceName().c_str())+
                    wxString("::")+
                    wxString(connectedPin->m_description.c_str()));
            } else {
                p_config->Write(m_item->m_description.c_str(),
                    wxString("0::NotConnected::NotConnected"));
            }
        }

        virtual bool read()
        {
            wxString tmp;
            bool readSuccess = p_config->Read(m_item->m_description.c_str(),&tmp);
            if( readSuccess ){
                readSuccess = false;
                wxStringTokenizer tkz(tmp, ":",wxTOKEN_STRTOK);
                wxString connected      = tkz.GetNextToken();
                wxString moduleInstance = tkz.GetNextToken();
                wxString description    = tkz.GetNextToken();

                if(connected == wxString("1")){
                    configManager::TOutputPins outputPins = configManager::getManager().
                        getMatchingOutputPins(m_item->m_selections.front(),m_targetModule);

                    configManager::TOutputPins::iterator currentpin;

                    int slot = -1;

                    for(currentpin=outputPins.begin();
                        currentpin != outputPins.end() && !readSuccess;
                        ++currentpin){
                            if(wxString((*currentpin)->m_description.c_str()) == description &&
                                wxString((*currentpin)->m_moduleInstance->getInstanceName().c_str()) ==
                                    moduleInstance){
                                    slot = (*currentpin)->m_slot;
                                    readSuccess = true;
                                }

                                *((int *) m_item->m_target.data) = slot;
                        }
                }
            } else {
                *((int *) m_item->m_target.data) = -1;
            }

            return true;
        }
    };

    /***************************************************************************/
    //! Child for configurable multiple ints
    /*!
     *
     *  \author  Fabian Wenzel
     *
     ***************************************************************************/
    class configFileMultipleInts : public configFileItemBase{
    public:
        configFileMultipleInts(const configItem *item,
            configurableObject *targetModule,
            wxFileConfig *config) :
        configFileItemBase(item,targetModule,config) {}

        virtual void save()
        {
            p_config->SetPath(m_item->m_description.c_str());
            vector< int >::iterator currentIntItem;
            p_config->Write(m_itemSectionIdentifier,(long int) target().size());
            int i=0;
            for(currentIntItem  = target().begin();
                currentIntItem != target().end();
                ++currentIntItem,++i)
                p_config->Write(wxString::Format(m_itemCounter,i),
                (long int) (*currentIntItem));
            p_config->SetPath("..");
        }

        virtual bool read()
        {
            long tmp;
            bool okay;
            p_config->SetPath(m_item->m_description.c_str());
            target().clear();
            int i;
            if ( okay =  p_config->Read(m_itemSectionIdentifier,&tmp) ){
                int itemCount = (int) tmp;
                for(i=0;i<itemCount && okay ;i++)
                    if (okay = p_config->Read(wxString::Format(m_itemCounter,i),&tmp))
                        target().push_back((int) tmp);
            }
            p_config->SetPath("..");
            return okay;
        }
    private:
        static const wxString m_itemSectionIdentifier;
        static const wxString m_itemCounter;
        vector< int > &target() const
        {
            return *((vector< int > *) m_item->m_target.data);
        }
    };

    const wxString configFileMultipleInts::m_itemSectionIdentifier("items");
    const wxString configFileMultipleInts::m_itemCounter("Item %d");


    /***************************************************************************/
    //! Child for configurable multiple strings
    /*!
     *
     *  \author  Fabian Wenzel
     *
     ***************************************************************************/
    class configFileMultipleStrings : public configFileItemBase{
    public:
        configFileMultipleStrings(const configItem *item,
            configurableObject *targetModule,
            wxFileConfig *config) :
        configFileItemBase(item,targetModule,config) {}
        virtual void save()
        {
            p_config->SetPath(m_item->m_description.c_str());
            vector< string >::iterator currentStringItem;
            p_config->Write(m_itemSectionIdentifier,(long int) target().size());
            int i=0;
            for(currentStringItem = target().begin();
                currentStringItem!= target().end();
                ++currentStringItem,++i){
                    p_config->Write(wxString::Format(m_itemCounter,i),
                        currentStringItem->c_str() );
                }
                p_config->SetPath("..");
        }
        virtual bool read()
        {
            long tmp;
            bool okay;
            p_config->SetPath(m_item->m_description.c_str());
            target().clear();
            int i;
            if ( okay = p_config->Read(m_itemSectionIdentifier,&tmp) ){
                int itemCount = (int) tmp;
                for(i=0;i<itemCount && okay ;i++){
                    wxString input;
                    if (okay = p_config->Read(wxString::Format(m_itemCounter,i),&input))
                        target().push_back(input.c_str());
                }
            }
            p_config->SetPath("..");
            return okay;
        }
    private:
        static const wxString m_itemSectionIdentifier;
        static const wxString m_itemCounter;
        vector< string > &target() const
        {
            return *((vector< string > *) m_item->m_target.data);
        }
    };

    const wxString configFileMultipleStrings::m_itemSectionIdentifier("items");
    const wxString configFileMultipleStrings::m_itemCounter("Item %d");


}

using namespace vistaLabFileItems;



//////////////////// Public methods for the class ///////////////////////////



/***************************************************************************/
/*!
 *  Constructor adds configurable items to the internal list
 *
 ***************************************************************************/
moduleConfigFile::moduleConfigFile(wxFileConfig *config):
p_config(config)
{
}

moduleConfigFile::~moduleConfigFile()
{
    for_each(m_fileConfigItems.begin(),m_fileConfigItems.end(),
        boost::checked_deleter< vistaLabFileItems::configFileItemBase >());
}

void moduleConfigFile::attachModule(configurableObject *module)
{
    moduleConfigFileBase::attachModule(module);
    addConfigItems();
}

/***************************************************************************/
/*!
 *  Configure function tries to read all items from all lists
 *  and aborts if okay gets false
 *
 ***************************************************************************/
bool moduleConfigFile::configure()
{
    bool okay = true;

    TConfigFileItems::const_iterator currentConfigItem;
    for(currentConfigItem  = m_fileConfigItems.begin();
        currentConfigItem != m_fileConfigItems.end() && okay;
        ++currentConfigItem)
        okay = (*currentConfigItem)->read();

    if(okay)
        okay = m_targetModule->initChoices();

    return okay;
}


/***************************************************************************/
/*!
 *  Adds configurable items to the internal list. They get fetched from
 *  the configManager and are wrapped in a configfile - item for
 *  accessing wxConfig in read() and save()
 *
 ***************************************************************************/
void moduleConfigFile::addConfigItems()
{
    for_each(m_fileConfigItems.begin(),m_fileConfigItems.end(),
        boost::checked_deleter< vistaLabFileItems::configFileItemBase >());
    m_fileConfigItems.clear();

    const TConfigItems *configitems = m_targetModule->getConfigItems();

    TConfigItems::const_iterator currentItem;

    for(currentItem  = configitems->begin();
        currentItem != configitems->end();
        ++currentItem)
        switch(currentItem->m_type){

           case configItem::STRING       :
           case configItem::FILENAME     :
           case configItem::SAVEFILENAME :
           case configItem::CUSTOMCHOICE :
               m_fileConfigItems.push_back(new configFileString(&*currentItem,
                   m_targetModule,
                   p_config));
               break;

           case configItem::INT          :
           case configItem::SINGLECHOICE :
               m_fileConfigItems.push_back(new configFileItem< int >(&*currentItem,
                   m_targetModule,
                   p_config));
               break;

           case configItem::INPUTPIN     :
               m_fileConfigItems.push_back(new configFileInputPin(&*currentItem,
                   m_targetModule,
                   p_config));
               break;

           case configItem::MULTIPLEFILENAMES:
               m_fileConfigItems.push_back(new configFileMultipleStrings(
                   &*currentItem, m_targetModule,
                   p_config));
               break;

           case configItem::BOOL:
           case configItem::ACTIVEBOOL:
               m_fileConfigItems.push_back(new configFileItem< bool >(&*currentItem,
                   m_targetModule,
                   p_config));
               break;

           case configItem::DOUBLE:
               m_fileConfigItems.push_back(new configFileItem< double >(&*currentItem,
                   m_targetModule,
                   p_config));
               break;

           case configItem::MULTIPLECHOICE:
           case configItem::RENDERINPUTPIN:
           case configItem::LOGINPUTPIN:
               m_fileConfigItems.push_back(new configFileMultipleInts(&*currentItem,
                   m_targetModule,
                   p_config));
           default:
               break;
        }
}


/***************************************************************************/
/*!
 *  Save function saves all items to the config-file
 *
 ***************************************************************************/
void moduleConfigFile::save()
{
    TConfigFileItems::const_iterator currentConfigItem;
    for(currentConfigItem  = m_fileConfigItems.begin();
        currentConfigItem != m_fileConfigItems.end();
        ++currentConfigItem){
            (*currentConfigItem)->save();
        }
}
