/***************************************************************************
                    vlmoduleconfigxml.cpp  -  description
                            -------------------
    begin                : Tue Aug 3 2004
    copyright            : (C) 2004 by Fabian Wenzel
    email                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include <wx/filename.h>

#include "vlmoduleconfigxml.h"
#include "vlconfigmanager.h"
#include "vlapp.h"
#include "vlprocessor.h"
#include "vlconfigitem.h"
#include "vlpin.h"

#include <boost/checked_delete.hpp>

using namespace std;
using namespace vistaLab;

const char*
convertToUTF8 (const char *inputString, char *encoding)
{
    const char *retval=NULL;
    xmlCharEncodingHandlerPtr coder = xmlFindCharEncodingHandler(encoding);
    xmlBufferPtr in = NULL, out = NULL;

    if(!coder){
        printf("No coder\n");
        return NULL;
    }

    in    = xmlBufferCreate();
    out   = xmlBufferCreate();

    xmlBufferCCat( in, inputString );
    if ( xmlCharEncInFunc( coder, out, in ) >= 0 ) {
        retval = (const char *) xmlStrdup( out->content );
    } else {
        printf("Error converting to UTF8\n");
    }

    xmlBufferFree( in );
    xmlBufferFree( out );
    xmlCharEncCloseFunc( coder );

    return retval;
}

const char*
convertFromUTF8 (const char *inputString, char *encoding)
{
    const char *retval=NULL;
    xmlCharEncodingHandlerPtr coder = xmlFindCharEncodingHandler(encoding);
    xmlBufferPtr in = NULL, out = NULL;

    if(!coder){
        printf("No coder\n");
        return NULL;
    }

    in    = xmlBufferCreate();
    out   = xmlBufferCreate();

    xmlBufferCCat( in, inputString );
    if ( xmlCharEncOutFunc( coder, out, in ) >= 0 ) {
        retval = (const char *) xmlStrdup( out->content );
    } else {
        printf("Error converting from UTF8\n");
    }

    xmlBufferFree( in );
    xmlBufferFree( out );
    xmlCharEncCloseFunc( coder );

    return retval;
}


//////////////////// Baseclass for the config-dialog items //////////////////

// ConfigDialogItems get an own namespace for better sorting in KDevelop
namespace vistaLabXMLItems{


    /***************************************************************************/
    //! Abstract Base File Item Class
    /*!
     *  \author  Fabian Wenzel
     *
     ***************************************************************************/
    class configXMLItemBase{
    public:
        virtual void save()=0;
        virtual bool read()=0;
        virtual ~configXMLItemBase() {}

    protected:
        configXMLItemBase(const configItem *item,
            configurableObject *targetModule,
            xmlNodePtr xmlNode) :
        m_item(item),
            m_targetModule(targetModule),
            m_XMLNode(xmlNode)
        {
        }

        const configItem         *m_item;
        const configurableObject *m_targetModule;
        xmlNodePtr                m_XMLNode;
    };




    /***************************************************************************/
    //! Template Child file Item middle-layer for matching wxConfig-calls
    /*!
     *  \author  Fabian Wenzel
     *
     ***************************************************************************/
    template< class T > class configXMLItem : public configXMLItemBase{
    public:
        configXMLItem(const configItem *item,
            configurableObject *targetModule,
            xmlNodePtr xmlNode) :
        configXMLItemBase(item,targetModule,xmlNode) {}
    protected:
        T &target() const
        {
            return *((T *) m_item->m_target.data);
        }
        virtual void save()
        {
            stringstream str;
            str << target();
            xmlNodePtr myParam = xmlNewChild(m_XMLNode, NULL, BAD_CAST "Parameter",NULL);
            xmlNewProp(myParam, BAD_CAST "Id", BAD_CAST m_item->m_description.c_str());
            xmlNewProp(myParam, BAD_CAST "Value", BAD_CAST BAD_CAST str.str().c_str());
        }

        virtual bool read()
        {
            xmlNodePtr currentNode=m_XMLNode->children;
            while(currentNode && !xmlStrEqual(xmlGetProp(currentNode,BAD_CAST "Id"),
                BAD_CAST m_item->m_description.c_str())){
                currentNode=currentNode->next;
            }

            if(currentNode){
                stringstream str;
                str << xmlGetProp(currentNode, BAD_CAST "Value");
                T tmp;
                str >> tmp;

                target() = tmp;
            }
            return currentNode != NULL;
        }
    };


    /***************************************************************************/
    //! Child for configurable input pins
    /*!
     *
     *  \author  Fabian Wenzel
     *
     ***************************************************************************/
    class configXMLInputPin : public configXMLItem<int>{
    public:
        configXMLInputPin(const configItem *item,
            configurableObject *targetModule,
            xmlNodePtr xmlNode) :
        configXMLItem<int>(item,targetModule,xmlNode) {}

        virtual void save()
        {
            outputPin *connectedPin =
                configManager::getManager().getOutputPin(target());
            if(connectedPin){
                xmlNodePtr inputPinNode = xmlNewChild(m_XMLNode, NULL, BAD_CAST "Pin",NULL);
                xmlNewProp(inputPinNode, BAD_CAST "Id", BAD_CAST m_item->m_description.c_str());
                xmlNewProp(inputPinNode, BAD_CAST "Connected", BAD_CAST "1");
                xmlNewProp(inputPinNode, BAD_CAST "ConnectedInstance", BAD_CAST
                    convertToUTF8((const char *) connectedPin->m_moduleInstance->getInstanceName().c_str(),
                        "ISO-8859-1"));
                xmlNewProp(inputPinNode, BAD_CAST "ConnectedOutputPin", BAD_CAST
                    connectedPin->m_description.c_str());
            } else {
                xmlNodePtr inputPinNode = xmlNewChild(m_XMLNode, NULL, BAD_CAST "Pin",NULL);
                xmlNewProp(inputPinNode, BAD_CAST "Id", BAD_CAST m_item->m_description.c_str());
                xmlNewProp(inputPinNode, BAD_CAST "Connected", BAD_CAST "0");
                xmlNewProp(inputPinNode, BAD_CAST "ConnectedInstance", BAD_CAST "Not Connected");
                xmlNewProp(inputPinNode, BAD_CAST "ConnectedOutputPin", BAD_CAST "Not Connected");
            }
        }

        virtual bool read()
        {
            bool readSuccess=false;

            xmlNodePtr currentNode=m_XMLNode->children;
            while(currentNode && !xmlStrEqual(BAD_CAST m_item->m_description.c_str(),xmlGetProp(currentNode, BAD_CAST "Id"))){
                currentNode=currentNode->next;
            }

            string moduleInstance;
            string pinDescription;

            if(currentNode){
                string connected = std::string((const char *) xmlGetProp(currentNode, BAD_CAST "Connected"));
                if(connected == std::string("1")){
                    moduleInstance = std::string((const char *) convertFromUTF8((const char *)
                        xmlGetProp(currentNode, BAD_CAST "ConnectedInstance"),"ISO-8859-1"));
                    pinDescription = std::string((const char *) xmlGetProp(currentNode,
                        BAD_CAST "ConnectedOutputPin"));

                    configManager::TOutputPins outputPins = configManager::getManager().
                        getMatchingOutputPins(m_item->m_selections.front(),m_targetModule);

                    configManager::TOutputPins::iterator currentPin;

                    // Initialize slot with "Not Connected".
                    int slot = -1;

                    for(currentPin=outputPins.begin();
                        currentPin != outputPins.end() && !readSuccess;
                        ++currentPin){

                        if((*currentPin)->m_description == pinDescription &&
                            (*currentPin)->m_moduleInstance->getInstanceName() == moduleInstance){
                                slot = (*currentPin)->m_slot;
                                readSuccess = true;
                            }

                            target() = slot;
                        }
                } else {
                    target() = -1;
                }

            }

            return true;
        }
    };




    /***************************************************************************/
    //! Child for configurable multiple ints
    /*!
     *
     *  \author  Fabian Wenzel
     *
     ***************************************************************************/
    class configXMLMultipleInts : public configXMLItemBase{
    public:
        configXMLMultipleInts(const configItem *item,
            configurableObject *targetModule,
            xmlNodePtr xmlNode) :
        configXMLItemBase(item,targetModule,xmlNode) {}

        virtual void save()
        {
            xmlNodePtr multipleFiles = xmlNewChild(m_XMLNode, NULL, BAD_CAST "Parameter",NULL);
            xmlNewProp(multipleFiles, BAD_CAST "Id", BAD_CAST m_item->m_description.c_str());

            vector< int >   ::iterator currentIntItem;

            for(currentIntItem  = target().begin();
                currentIntItem != target().end();
                ++currentIntItem){
                    stringstream str;
                    str << *currentIntItem;
                    xmlNewChild(multipleFiles,NULL, BAD_CAST "Item",BAD_CAST str.str().c_str());
                }
        }

        virtual bool read()
        {
            target().clear();
            xmlNodePtr currentNode=m_XMLNode->children;
            while(currentNode && !xmlStrEqual(xmlGetProp(currentNode,BAD_CAST "Id"),
                BAD_CAST m_item->m_description.c_str())){
                currentNode=currentNode->next;
            }
            if(currentNode){
                xmlNodePtr currentChild = currentNode->children;
                while(currentChild){
                    if(xmlStrEqual(currentChild->name,BAD_CAST "Item")){
                        stringstream str(std::string((const char *) currentChild->children->content));
                        int tmpTarget;
                        str >> tmpTarget;
                        target().push_back(tmpTarget);
                    }
                    currentChild=currentChild->next;
                }
            }
            return true;
        }
    protected:
        vector< int > &target() const
        {
            return *((vector< int > *) m_item->m_target.data);
        }

    };


    /***************************************************************************/
    //! Child for Render pin
    /*!
     *
     *  \author  Fabian Wenzel
     *
     ***************************************************************************/
    class configXMLRenderInputPin : public configXMLMultipleInts{
    public:
        configXMLRenderInputPin(const configItem *item,
            configurableObject *targetModule,
            xmlNodePtr xmlNode) :
        configXMLMultipleInts(item,targetModule,xmlNode) {}

        virtual void save()
        {
            xmlNodePtr renderPins = xmlNewChild(m_XMLNode, NULL, BAD_CAST "Parameter",NULL);
            xmlNewProp(renderPins, BAD_CAST "Id", BAD_CAST m_item->m_description.c_str());

            vector< int >   ::iterator currentIntItem;

            configManager::TRenderPins &pins = configManager::getManager().getRenderPins();

            for(currentIntItem  = target().begin();
                currentIntItem != target().end();
                ++currentIntItem){

            renderPin *pin=pins[*currentIntItem];

            xmlNodePtr inputPinNode = xmlNewChild(renderPins, NULL, BAD_CAST "Pin",NULL);
            xmlNewProp(inputPinNode, BAD_CAST "ConnectedInstance", BAD_CAST
                convertToUTF8((const char *) pin->m_moduleInstance->getInstanceName().c_str(),"ISO-8859-1"));
            xmlNewProp(inputPinNode, BAD_CAST "ConnectedRenderPin", BAD_CAST
                pin->m_description.c_str());
            }
        }

        virtual bool read()
        {
            target().clear();
            bool readSuccess=false;

            xmlNodePtr currentNode=m_XMLNode->children;
            while(currentNode && !xmlStrEqual(BAD_CAST m_item->m_description.c_str(),xmlGetProp(currentNode, BAD_CAST "Id"))){
                currentNode=currentNode->next;
            }

            string moduleInstance;
            string pinDescription;

            if(currentNode){

                xmlNodePtr currentChild = currentNode->children;
                while(currentChild){
                    const char *tmpInstance = (const char *) xmlGetProp(currentChild, BAD_CAST "ConnectedInstance");
                    const char *tmpRenderPin= (const char *) xmlGetProp(currentChild, BAD_CAST "ConnectedRenderPin");
                    if(tmpInstance && tmpRenderPin){

                        readSuccess=false;

                        string instance  = std::string(tmpInstance);
                        string renderPin = std::string(tmpRenderPin);

                        configManager::TRenderPins renderPins = configManager::getManager().getRenderPins();
                        configManager::TRenderPins::iterator currentPin;

                        int i;
                        for(i=0,currentPin=renderPins.begin();
                            currentPin != renderPins.end() && !readSuccess;
                            ++currentPin,++i){

                            if((*currentPin)->m_description == renderPin &&
                               (*currentPin)->m_moduleInstance->getInstanceName() == instance){
                                target().push_back(i);
                                readSuccess = true;
                            }
                        }
                        if(!readSuccess)
                            // There is an instance and a pin description which cannot be found, so there is an error
                            return false;
                    }
                    currentChild=currentChild->next;
                }
            }

            return true;
        }
    };


    /***************************************************************************/
    //! Child for Log pin
    /*!
     *
     *  \author  Fabian Wenzel
     *
     ***************************************************************************/
    class configXMLLogInputPin : public configXMLMultipleInts{
    public:
        configXMLLogInputPin(const configItem *item,
            configurableObject *targetModule,
            xmlNodePtr xmlNode) :
        configXMLMultipleInts(item,targetModule,xmlNode) {}

        virtual void save()
        {
            xmlNodePtr logPins = xmlNewChild(m_XMLNode, NULL, BAD_CAST "Parameter",NULL);
            xmlNewProp(logPins, BAD_CAST "Id", BAD_CAST m_item->m_description.c_str());

            configManager::TLogPins  &pins = configManager::getManager().getLogPins();

            vector< int >   ::iterator currentIntItem;

            for(currentIntItem  = target().begin();
                currentIntItem != target().end();
                ++currentIntItem){

                logPin *pin=pins[*currentIntItem];

                xmlNodePtr inputPinNode = xmlNewChild(logPins, NULL, BAD_CAST "Pin",NULL);
                xmlNewProp(inputPinNode, BAD_CAST "ConnectedInstance", BAD_CAST
                    convertToUTF8((const char *) pin->m_moduleInstance->getInstanceName().c_str(),"ISO-8859-1"));
                xmlNewProp(inputPinNode, BAD_CAST "ConnectedLogPin", BAD_CAST
                    pin->m_description.c_str());
            }
        }

        virtual bool read()
        {
            target().clear();

            xmlNodePtr currentNode=m_XMLNode->children;
            while(currentNode && !xmlStrEqual(BAD_CAST m_item->m_description.c_str(),
                xmlGetProp(currentNode, BAD_CAST "Id"))){
                currentNode=currentNode->next;
            }

            string moduleInstance;
            string pinDescription;

            if(currentNode){

                bool readSuccess = false;

                xmlNodePtr currentChild = currentNode->children;
                while(currentChild){
                    const char *tmpInstance = (const char *) xmlGetProp(currentChild, BAD_CAST "ConnectedInstance");
                    const char *tmpLogPin   = (const char *) xmlGetProp(currentChild, BAD_CAST "ConnectedLogPin");

                    if(tmpInstance && tmpLogPin){

                        readSuccess=false;
                        string instance  = std::string(tmpInstance);
                        string logPin    = std::string(tmpLogPin);

                        configManager::TLogPins logPins = configManager::getManager().getLogPins();
                        configManager::TLogPins::iterator currentPin;

                        int i=0;
                        for(currentPin=logPins.begin();
                            currentPin != logPins.end() && !readSuccess;
                            ++currentPin,++i){

                            if((*currentPin)->m_description == logPin &&
                               (*currentPin)->m_moduleInstance->getInstanceName() == instance){
                                target().push_back(i);
                                readSuccess = true;
                            }
                        }
                        if(!readSuccess)
                            // There is an instance and a pin description which cannot be found, so there is an error
                            return false;
                    }

                    currentChild=currentChild->next;
                }
            }

            return true;
        }
    };


    /***************************************************************************/
    //! Child for configurable multiple strings
    /*!
     *
     *  \author  Fabian Wenzel
     *
     ***************************************************************************/
    class configXMLMultipleFilenames : public configXMLItemBase{
    public:
        configXMLMultipleFilenames(const configItem *item,
            configurableObject *targetModule,
            xmlNodePtr xmlNode) :
        configXMLItemBase(item,targetModule,xmlNode) {}

        virtual void save(){
            xmlNodePtr multipleFiles = xmlNewChild(m_XMLNode, NULL, BAD_CAST "Parameter",NULL);
            xmlNewProp(multipleFiles, BAD_CAST "Id", BAD_CAST m_item->m_description.c_str());

            vector< string >   ::iterator currentItem;

            for(currentItem  = target().begin();
                currentItem != target().end();
                ++currentItem){
                // Save relative path (suggested by ralph)
                wxFileName fileName(currentItem->c_str());
                fileName.MakeRelativeTo(wxGetCwd());

                xmlNewChild(multipleFiles,NULL, BAD_CAST "Item",
                    BAD_CAST convertToUTF8((const char *)fileName.GetFullPath().c_str(),"ISO-8859-1"));
            }
        }

        virtual bool read()
        {
            target().clear();
            xmlNodePtr currentNode=m_XMLNode->children;
            while(currentNode && !xmlStrEqual(xmlGetProp(currentNode,BAD_CAST "Id"),
                BAD_CAST m_item->m_description.c_str())){
                currentNode=currentNode->next;
            }

            if(currentNode){
                xmlNodePtr currentChild = currentNode->children;

                while(currentChild){
                    if(xmlStrEqual(currentChild->name,BAD_CAST "Item")){
                        wxFileName fileName(convertFromUTF8(
                            (const char *) currentChild->children->content,"ISO-8859-1"));
                        fileName.MakeAbsolute();
                        target().push_back(fileName.GetFullPath().c_str());
                    }
                    currentChild=currentChild->next;
                }
            }
            return true;
        }
    private:
        vector< string >  &target() const
        {
            return *(( vector< string > *) m_item->m_target.data);
        }
    };

    class configXMLString: public configXMLItemBase{
    public:
        configXMLString(const configItem *item,
            configurableObject *targetModule,
            xmlNodePtr xmlNode) :
        configXMLItemBase(item,targetModule,xmlNode) {}

        virtual void save()
        {
            xmlNodePtr myParam = xmlNewChild(m_XMLNode, NULL, BAD_CAST "Parameter",NULL);
            xmlNewProp(myParam, BAD_CAST "Id", BAD_CAST m_item->m_description.c_str());
            xmlNewProp(myParam, BAD_CAST "Value", BAD_CAST convertToUTF8(target().c_str(),"ISO-8859-1"));
        }

        virtual bool read()
        {
            xmlNodePtr currentNode=m_XMLNode->children;
            while(currentNode && !xmlStrEqual(xmlGetProp(currentNode,BAD_CAST "Id"),
                BAD_CAST m_item->m_description.c_str())){
                currentNode=currentNode->next;
            }

            if(currentNode){
                const char * convertedString =
                    convertFromUTF8((const char *) xmlGetProp(currentNode, BAD_CAST "Value"),"ISO-8859-1");
                if(convertedString)
                    target() = convertedString;
                else return false;
            }
            return currentNode != NULL;
        }

    private:
        string &target() const
        {
            return *((string *) m_item->m_target.data);
        }

    };

    class configXMLFilename: public configXMLItemBase{
    public:
        configXMLFilename(const configItem *item,
            configurableObject *targetModule,
            xmlNodePtr xmlNode) :
        configXMLItemBase(item,targetModule,xmlNode) {}

        virtual void save()
        {
            wxFileName fileName(target().c_str());
            fileName.MakeRelativeTo(wxGetCwd());
            xmlNodePtr myParam = xmlNewChild(m_XMLNode, NULL, BAD_CAST "Parameter",NULL);
            xmlNewProp(myParam, BAD_CAST "Id", BAD_CAST m_item->m_description.c_str());
            xmlNewProp(myParam, BAD_CAST "Value", BAD_CAST
                convertToUTF8(fileName.GetFullPath().c_str(),"ISO-8859-1"));
        }

        virtual bool read()
        {
            xmlNodePtr currentNode=m_XMLNode->children;
            while(currentNode && !xmlStrEqual(xmlGetProp(currentNode,BAD_CAST "Id"),
                BAD_CAST m_item->m_description.c_str())){
                currentNode=currentNode->next;
            }

            if(currentNode){
                const char * convertedString =
                    convertFromUTF8((const char *) xmlGetProp(currentNode, BAD_CAST "Value"),"ISO-8859-1");
                if(convertedString){
                    wxFileName fileName(convertedString);
                    fileName.MakeAbsolute();
                    target() = fileName.GetFullPath().c_str();
                }
                else return false;
            }
            return currentNode != NULL;
        }

    private:
        string &target() const
        {
            return *((string *) m_item->m_target.data);
        }

    };
}


using namespace vistaLabXMLItems;



//////////////////// Public methods for the class ///////////////////////////



/***************************************************************************/
/*!
 *  Constructor adds configurable items to the internal list
 *
 ***************************************************************************/
moduleConfigXML::moduleConfigXML(xmlNodePtr XMLNode) :
p_XMLNode(XMLNode)
{
}

moduleConfigXML::~moduleConfigXML()
{
    clearConfigItems();
}

void moduleConfigXML::attachModule(configurableObject *module)
{
    clearConfigItems();
    moduleConfig::attachModule(module);
    addConfigItems();
}

void moduleConfigXML::clearConfigItems()
{
    for_each(m_XMLConfigItems.begin(),m_XMLConfigItems.end(),
            boost::checked_deleter< configXMLItemBase >());
}

/***************************************************************************/
/*!
 *  Configure function tries to read all items from all lists
 *  and aborts if okay gets false
 *
 ***************************************************************************/
bool moduleConfigXML::configure()
{
    bool okay = true;

    TConfigXMLItems::const_iterator currentConfigItem;
    for(currentConfigItem  = m_XMLConfigItems.begin();
        currentConfigItem != m_XMLConfigItems.end() && okay;
        ++currentConfigItem)
        okay = (*currentConfigItem)->read();

    if(okay)
        okay = m_targetModule->initChoices();

    return okay;
}


/***************************************************************************/
/*!
 *  Adds configurable items to the internal list. They get fetched from
 *  the configManager and are wrapped in a configfile - item for
 *  accessing wxConfig in read() and save()
 *
 ***************************************************************************/
void moduleConfigXML::addConfigItems()
{
    const TConfigItems *configItems = m_targetModule->getConfigItems();

    TConfigItems::const_iterator currentItem;

    for(currentItem  = configItems->begin();
        currentItem != configItems->end();
        ++currentItem)

        switch(currentItem->m_type){

           case configItem::STRING       :
           case configItem::CUSTOMCHOICE :
               m_XMLConfigItems.push_back(new configXMLString(&*currentItem,
                   m_targetModule,
                   p_XMLNode));
               break;

           case configItem::FILENAME     :
           case configItem::SAVEFILENAME :
               m_XMLConfigItems.push_back(new configXMLFilename(&*currentItem,
                   m_targetModule,
                   p_XMLNode));
               break;

           case configItem::INT          :
           case configItem::SINGLECHOICE :
               m_XMLConfigItems.push_back(new configXMLItem< int >(&*currentItem,
                   m_targetModule,
                   p_XMLNode));
               break;

           case configItem::INPUTPIN     :
               m_XMLConfigItems.push_back(new configXMLInputPin(&*currentItem,
                   m_targetModule,
                   p_XMLNode));
               break;

           case configItem::MULTIPLEFILENAMES:
               m_XMLConfigItems.push_back(new configXMLMultipleFilenames(
                   &*currentItem, m_targetModule,p_XMLNode));
               break;

           case configItem::BOOL:
           case configItem::ACTIVEBOOL:
               m_XMLConfigItems.push_back(new configXMLItem< bool >(&*currentItem,
                   m_targetModule,p_XMLNode));
               break;

           case configItem::DOUBLE:
               m_XMLConfigItems.push_back(new configXMLItem< double >(&*currentItem,
                   m_targetModule,p_XMLNode));
               break;

           case configItem::MULTIPLECHOICE:
               m_XMLConfigItems.push_back(new configXMLMultipleInts(&*currentItem,
                   m_targetModule,p_XMLNode));
               break;

           case configItem::RENDERINPUTPIN:
               m_XMLConfigItems.push_back(new configXMLRenderInputPin(&*currentItem,
                   m_targetModule,p_XMLNode));
               break;

           case configItem::LOGINPUTPIN:
               m_XMLConfigItems.push_back(new configXMLLogInputPin(&*currentItem,
                   m_targetModule,p_XMLNode));
               break;

           default:
               break;
        }
}


/***************************************************************************/
/*!
 *  Save function saves all items to the config-file
 *
 ***************************************************************************/
void moduleConfigXML::save()
{
    TConfigXMLItems::const_iterator currentConfigItem;
    for(currentConfigItem  = m_XMLConfigItems.begin();
        currentConfigItem != m_XMLConfigItems.end();
        ++currentConfigItem)
        (*currentConfigItem)->save();
}
