/***************************************************************************
                        vlprocessor.cpp  -  description
                            -------------------
    begin                : Fri Nov 16 2001
    copyright            : (C) 2001 by Fabian Wenzel
    email                : wenzel@tu-harburg.de
 
***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include "vlprocessor.h"
#include "vlinputthread.h"
#include "vlapp.h"
#include "vldocument.h"
#include "vlmodule.h"
#include "vlconfigmanager.h"
#include "vlframe.h"
#include "vlframecontrols.h"

#include <wx/msgdlg.h>
#include <wx/regex.h>
#include <algorithm>
#include <functional>
#include <boost/checked_delete.hpp>
#include <vigra/stdimage.hxx>
#include <iostream>

// Exception handling: We handle standard exceptions

#include <wx/string.h>

#define FRAMERATETIMER 20000

using namespace std;
using namespace vistaLab;

///////////////////////// processCore Members ///////////////////////

wxMutex      processCore::m_processingWaitMutex;
int          processCore::m_triggeredFramesCounter;
bool         processCore::m_ready;

wxCondition
    processCore::m_processingWait (processCore::m_processingWaitMutex);

/***************************************************************************/
/*!
 *  Constructor creates the thread and gets pointers for fast processing
 *
 ***************************************************************************/
processCore::processCore(bool wait) :
wxThread(wxTHREAD_JOINABLE),
m_wait(wait)
{
    Create();
}

/***************************************************************************/
/*!
 *  The main process function waits for new data in vistaLabDocument
 *  This takes place as soon as there is a new slice ready to be processed
 *  and the function gets signaled or when there is already a new slice
 *  without having to wait at all.
 *
 ***************************************************************************/
wxThread::ExitCode processCore::Entry()
{
    // Gets set to false after each run
    module::m_processedFramesCounter=0;
    m_triggeredFramesCounter=0;

    // Main loop - run for every new slice until stopped
    while ( wxGetApp().getProcessor()->isRunning() ){

        waitIfNecessary();

        // The threads may have stopped the processor, so data may
        // be missing in this case. => Only continue, if still running
        if(wxGetApp().getProcessor()->isRunning()){

            wxGetApp().getDocument()->shiftBuffer();

            if(m_waited){
                m_processingWaitMutex.Unlock();
                m_waited=false;
            }

            m_ready=false;

            slice *currentSlice = wxGetApp().getDocument()->getCurrentSlice();

            // Termination - flag
            bool modulesSuccessful = true;

            TActiveModules::iterator currentModule;
            TActiveModules::iterator errorModule;

            wxString myErrorString = "";

            //////////////////////////////////////////////////////////////////

            // Apply pending parameter changes which are done in real-time
            // This code is partially taken from wxmoduleConfigdialog::Validate
            // Problem: Binding of validator-functions to parameters does not
            // happen, so either changes can be applied for all parameters of a
            // module or not
            // We must do this in GUI as labels are updated with colors

            if(((frame *) wxGetApp().GetTopWindow())->getActiveProjectTree()->hasParameterChanges()){
                    wxString errors = ((frame *)
                        wxGetApp().GetTopWindow())->getActiveProjectTree()->applyPendingChanges();

                if(errors!= wxEmptyString){
                    setError(errors);
                    // This can be done as we are having a GUI here
                    wxCommandEvent event(wxEVT_VISTALAB,DISPLAY_ERROR);
                    wxGetApp().GetTopWindow()->AddPendingEvent(event);
                }
            }




            //////////////////////////////////////////////////////////////////

            try{
                // Call process - functions of other input/processing/output/GUI

                for (currentModule  = wxGetApp().getProcessor()->m_inputModules.begin();
                    currentModule != wxGetApp().getProcessor()->m_inputModules.end()
                    && modulesSuccessful;
                ++currentModule){
                    modulesSuccessful = (*currentModule)->process(*currentSlice);
                    if(!modulesSuccessful)
                        errorModule = currentModule;
                }

                for (currentModule  = wxGetApp().getProcessor()->m_processingModules.begin();
                    currentModule != wxGetApp().getProcessor()->m_processingModules.end()
                    && modulesSuccessful;
                ++currentModule){
                    modulesSuccessful = (*currentModule)->process(*currentSlice);
                    if(!modulesSuccessful)
                        errorModule = currentModule;
                }

                for (currentModule  = wxGetApp().getProcessor()->m_outputModules.begin();
                    currentModule != wxGetApp().getProcessor()->m_outputModules.end()
                    && modulesSuccessful;
                ++currentModule){
                    modulesSuccessful = (*currentModule)->process(*currentSlice);
                    if(!modulesSuccessful)
                        errorModule = currentModule;
                }

                // block the whole GUI on GTK as Refresh() is not thread safe
                VISTALABENTERGUI
                    // The draw() - functions have to be inside the gui mutex
                    // As they might have own windows (OpenGL, Stopwatch etc.)
                    TActiveModules::iterator currentGuiModule;
                for (currentGuiModule  = wxGetApp().getProcessor()->m_guiModules.begin();
                    currentGuiModule != wxGetApp().getProcessor()->m_guiModules.end()
                    && modulesSuccessful;
                ++currentGuiModule){
                    modulesSuccessful = (*currentGuiModule)->process(*currentSlice);
                    if(!modulesSuccessful)
                        errorModule = currentModule;
                }
                VISTALABLEAVEGUI

            } catch (std::exception &e)
            {
                modulesSuccessful = false;
                errorModule = currentModule;
                setError(e.what());
            }
            catch (...)
            {
                modulesSuccessful = false;
                errorModule = currentModule;
                setError(_("Unknown Exception"));
            }
            if (!modulesSuccessful){
                // If a module was not successful: stop processing
                // This is what the threads to themselves in this case
                if(wxGetApp().singleRun()){
                    setError(wxString::Format(_("%s stopped processing: %s"),
                            (*errorModule)->getInstanceName().c_str(),getError().c_str()));
                } else {
                    setError(wxString::Format(_("%s stopped processing:\n\n%s"),
                            (*errorModule)->getInstanceName().c_str(),getError().c_str()));
                }

                wxGetApp().getProcessor()->stop();

                --m_triggeredFramesCounter;
                --module::m_processedFramesCounter;

                wxGetApp().getDocument()->shiftSlicePointer(true);
            }

            ++module::m_processedFramesCounter;

            wxCommandEvent event(wxEVT_VISTALAB,UPDATE_TICKS);
            wxGetApp().GetTopWindow()->AddPendingEvent(event);

        } else
            // Not running after having waited for processing - still have to unlock!
            if(m_wait)
                m_processingWaitMutex.Unlock();

    }

    // Not running anymore: Wait for threads to exit and then stop
    // modules

    for_each(wxGetApp().getProcessor()->m_activeInputThreads.begin(),
        wxGetApp().getProcessor()->m_activeInputThreads.end(),
        mem_fun(&inputThread::Wait));

    for_each(wxGetApp().getProcessor()->m_activeInputThreads.begin(),
        wxGetApp().getProcessor()->m_activeInputThreads.end(),
        boost::checked_deleter<inputThread>());

        return 0;
}

void processCore::waitIfNecessary()
{
    // Wait to be woken up - either by the document or by the processor
    // if we should wait - we don't need to if the user chooses
    // "fast processing" in global options
    if(m_wait || wxGetApp().getProcessor()->isPaused()){
        m_processingWaitMutex.Lock();
        m_waited=true;
        m_ready=true;
        do
            m_processingWait.Wait();
        while (wxGetApp().getProcessor()->isPaused());
    } else {
        // We are not waiting for an external trigger,
        // so increment number of triggered frames
        m_ready=true;
        m_wait=false;
        ++m_triggeredFramesCounter;
    }
}

// Triggers the processor core *externally*, i.e. from the
// timer (if no input threads are available) or from the
// document if all input threads have delivered data
void processCore::continueProcessing()
{
    m_processingWait.Signal();

    // Increment number of triggered frames if this
    // function has been called regularly (and not when
    // stopping to wake up processing thread in order
    // to exit)
    if(wxGetApp().getProcessor()->isRunning())
        ++m_triggeredFramesCounter;

    m_processingWaitMutex.Unlock();
    if(!m_ready){
        wxCommandEvent event(wxEVT_VISTALAB,UPDATE_TICKS);
        wxGetApp().GetTopWindow()->AddPendingEvent(event);
    }
}


/***************************************************************************
 *
 *  Event table
 *
 ***************************************************************************/
BEGIN_EVENT_TABLE(processor,wxEvtHandler)
EVT_TIMER(FRAMERATETIMER, processor::frameTick)
END_EVENT_TABLE()




////////////////////////////// Public functions ////////////////////////////


/***************************************************************************/
/*!
 *  Constructor sets the state to stopped and generates a new frameRateTimer
 *
 ***************************************************************************/
processor::processor():
m_running(false),
m_stopped(true),
m_paused(false),
m_needsNewSorting(true)
{
    m_frameRateTimer = new wxTimer(this,FRAMERATETIMER);
}


/***************************************************************************/
/*!
 *  Destructor deletes all active modules and the frameRateTimer
 *
 ***************************************************************************/
processor::~processor()
{
    delete m_frameRateTimer;
}


/***************************************************************************/
/*!
 *  Sorts all active modules. This function has two tasks: In the first
 *  one, the config manager is called for checking inputpins of all
 *  modules. If there are unconnected ones, the remainder of the
 *  data chain is invalid and has to be removed temporarily such that
 *  only fully connected modules will operate on data.
 *  The second step is to sort modules into categories such that
 *  threads can be started and that using the slider in the frame only
 *  triggers output modules (processing modules may not work anticausally)
 *
 ***************************************************************************/
void processor::sortModules()
{
    if(m_needsNewSorting){
        m_inputThreadModules.clear();
        m_inputModules.clear();
        m_processingModules.clear();
        m_outputModules.clear();
        m_guiModules.clear();


        // Let the config manager disable invalid modules
        configManager::getManager().setValidModules();

        // Use the remaining valid ones for sorting
        TActiveModules::const_iterator currentModule;
        for(currentModule  = m_activeModules.begin();
            currentModule != m_activeModules.end();
            ++currentModule){

                if( (*currentModule)->isActive() && (*currentModule)->isValid() )
                    switch((*currentModule)->getModuleClass()){

               case module::INPUTTHREAD:
                   m_inputThreadModules.push_back(*currentModule);
                   break;

               case module::INPUT:
                   m_inputModules.push_back(*currentModule);
                   break;

               case module::PROCESSING:
                   m_processingModules.push_back(*currentModule);
                   break;

               case module::OUTPUT:
                   m_outputModules.push_back(*currentModule);
                   break;

               case module::GUI:
                   m_guiModules.push_back(*currentModule);
                   break;
                    }
            }
            m_needsNewSorting = false;
    }
}

/***************************************************************************/
/*!
 *  Starts the engine. First processors and if this is successful
 *  the input streams by signaling them (such that they will initialize
 *  themselves). After having waited for all, the frameRateTimer is started
 *  and the input streams get signaled again for starting processing
 *
 ***************************************************************************/
void processor::start()
{
    m_paused = false;
    if (initModules()){

        wxGetApp().getDocument()->resetBuffers();

        m_running      = true;
        m_stopped      = false;

        ((frame *)wxGetApp().GetTopWindow())->updateMenu();

        m_frameCounter = 0;

        sortModules();

        // Create process core: The wait flag is disabled if we set it
        // in the global options and if there are no inputthread modules
        m_processorCore = new processCore(m_inputThreadModules.size() ||
            !wxGetApp().getDocument()->getWaitFlag());

        m_processorCore->Run();

        if(m_inputThreadModules.size()){
            m_activeInputThreads.clear();
            // Start input threads
            TActiveModules::iterator currentInputThread;
            for(currentInputThread  = m_inputThreadModules.begin();
                currentInputThread != m_inputThreadModules.end();
                ++currentInputThread){
                    inputThread *newThread = new inputThread(*currentInputThread);
                    newThread->Run();
                    m_activeInputThreads.push_back(newThread);
                }
        } else if (usingTimer())
            toggleTimer(true);
    } else {
        stopModules();
    }
}


/***************************************************************************/
/*!
 *  Stops the engine. This gets done by stopping the timer
 *
 ***************************************************************************/
void processor::stop()
{
    // This method gets called twice in a row when an input thread stops
    // (Once because of the immediate call, the second time from the
    // gui due to the emitted event which has to update the start/stop - menu)
    // So the next part has to happen only then the process is not stopped
    if( m_running ){
        m_paused  = false;
        m_running = false;
        processCore::prepareForProcessing();
        processCore::continueProcessing();

        wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED,PROCESS_STOP);
        wxGetApp().GetTopWindow()->AddPendingEvent(event);
        if(usingTimer())
            toggleTimer(false);

        ((frame *)wxGetApp().GetTopWindow())->updateMenu();
    } else {
        m_processorCore->Wait();
        delete m_processorCore;
        m_stopped = true;
        stopModules();
        ((frame *)wxGetApp().GetTopWindow())->updateMenu();
        // Display error messages here as we are in the main thread

        if(wxGetApp().singleRun()){
            wxString error = getError().c_str();

            if(error != wxEmptyString)
                printf("%s\n",error.c_str());

            wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED,wxID_EXIT);
            wxGetApp().GetTopWindow()->AddPendingEvent(event);
        } else {
            wxCommandEvent event(wxEVT_VISTALAB,DISPLAY_ERROR);
            wxGetApp().GetTopWindow()->AddPendingEvent(event);
        }
    }
}


/***************************************************************************/
/*!
 *  Removes a module
 *
 ***************************************************************************/
void processor::removeModule(module *moduleToRemove){
    wxGetApp().getDocument()->resetBuffers();

    m_needsNewSorting = true;

    TActiveModules::iterator currentModule =
        find(m_activeModules.begin(),m_activeModules.end(),moduleToRemove);

    delete *currentModule;
    m_activeModules.erase(currentModule);

    configManager::getManager().updatePins();
}


/***************************************************************************/
/*!
 *  Adds a module
 *
 ***************************************************************************/
void processor::addModule(module *newModule, module *predecessor)
{
    while(!checkInstanceName(newModule));

    wxGetApp().getDocument()->resetBuffers();
    m_needsNewSorting = true;
    module *newActiveModule = newModule->clone();
    newActiveModule->_initConfig(false);
    if(predecessor == NULL)
        m_activeModules.push_back(newActiveModule);
    else{
        TActiveModules::iterator currentModule;
        m_activeModules.insert(
            find(m_activeModules.begin(),m_activeModules.end(),predecessor),
            newActiveModule);
    }

    configManager::getManager().updatePins();
}


/***************************************************************************/
/*!
 *  Removes all (closes project)
 *
 ***************************************************************************/
void processor::removeAll()
{
    wxGetApp().getDocument()->resetBuffers();
    m_needsNewSorting = true;

    for_each(m_activeModules.begin(),m_activeModules.end(),
        boost::checked_deleter<module>());

    m_activeModules.clear();

    configManager::getManager().updatePins();
}

/***************************************************************************/
/*!
 *  The callback for the timer which just sync the streams
 *
 ***************************************************************************/
void processor::frameTick(wxTimerEvent &e)
{
    doFrameTick();
}

void processor::doFrameTick()
{
    processCore::prepareForProcessing();
    processCore::continueProcessing();
}

/***************************************************************************/
/*!
 *  Helper function: Gets called by the GUI when the slider has been moved.
 *  The document gets advised to shift the output pointer and the
 *  threads are notified
 *
 ***************************************************************************/
void processor::showNeighborSlice(bool downwards)
{
    wxGetApp().getDocument()->shiftSlicePointer(downwards);

    // The user may have (dis)connected modules in between
    sortModules();

    bool modulesSuccessful = true;

    slice *currentSlice = wxGetApp().getDocument()->getCurrentSlice();

    // The user may have used the slider beyond the first frame
    if(!currentSlice)
        return;


    // Initialize process modules as they might have log/renderPins and
    // want to use the data of the current slice 

    TActiveModules::iterator currentprocessingModule;
    for (currentprocessingModule  = m_processingModules.begin();
        currentprocessingModule != m_processingModules.end();
        ++currentprocessingModule)
        (*currentprocessingModule)->preProcess(*currentSlice);

    // Call process - functions of output and gui modules
    TActiveModules::iterator currentOutputmodule;
    for (currentOutputmodule  = m_outputModules.begin();
        currentOutputmodule != m_outputModules.end() && modulesSuccessful;
        ++currentOutputmodule)
        modulesSuccessful = (*currentOutputmodule)->process(*currentSlice);

    TActiveModules::iterator currentGuiModule;
    for (currentGuiModule  = m_guiModules.begin();
        currentGuiModule != m_guiModules.end() && modulesSuccessful;
        ++currentGuiModule)
        modulesSuccessful = (*currentGuiModule)->process(*currentSlice);
}


/***************************************************************************/
/*!
 *  Initializes the processors. All processors are initialized even if
 *  one fails as they are all stopped in processor::stopprocessors()
 *
 ***************************************************************************/
bool processor::initModules()
{
    bool initSuccess=true;
    TActiveModules::iterator currentModules;
    for (currentModules  = m_activeModules.begin();
        currentModules != m_activeModules.end();
        ++currentModules){
            if( (*currentModules)->isActive() && (*currentModules)->isValid() ){
                bool result = (*currentModules)->init();
                // No need to show a dialog here, this will be done in stopModules()
                if(!result)
                    setError(wxString::Format(_("%s could not be initialized:\n\n%s"),
                    (*currentModules)->getInstanceName().c_str(),
                    module::getError().c_str()));
                initSuccess &= result;
            }
        }

        return initSuccess;
}


/***************************************************************************/
/*!
 *  Stops the processors.
 *
 ***************************************************************************/
void processor::stopModules()
{
    TActiveModules::iterator currentModule;
    for (currentModule  = m_activeModules.begin();
        currentModule != m_activeModules.end();
        ++currentModule)
        if( (*currentModule)->isActive() && (*currentModule)->isValid() )
            (*currentModule)->done();
}


/***************************************************************************/
/*!
 *  Swaps two modules (if they are swapped by mouse in mList)
 *
 ***************************************************************************/
void processor::swapModules(module *first, module *second)
{
    wxGetApp().getDocument()->resetBuffers();

    TActiveModules::iterator firstIter = find(m_activeModules.begin(),
        m_activeModules.end(),first);

    TActiveModules::iterator secondIter = find(firstIter,
        m_activeModules.end(),second);

    if(secondIter == m_activeModules.end()){
        // The first module has either been dragged upwards
        // Or to the very end, so the search has to
        // take place across the whole list again
        firstIter = m_activeModules.erase(firstIter);
        m_activeModules.insert(
            find(m_activeModules.begin(),
            m_activeModules.end(),second),first);
    } else {
        firstIter=m_activeModules.erase(firstIter);
        if(*firstIter == second){
            m_activeModules.insert(firstIter+1,first);
        } else {
            m_activeModules.insert(find(firstIter,
                m_activeModules.end(),second),first);
        }
    }

    configManager::getManager().updatePins();
}

bool processor::checkInstanceName(module *moduleToCheck)
{
    // First, get configuration item for changing the name
    TConfigItems *configItems = moduleToCheck->getConfigItems();
    TConfigItems::iterator instanceNameItem = configItems->begin();

    while((*instanceNameItem).m_description != std::string("Instancename"))
        ++instanceNameItem;

    // Check if instance name is empty
    while(moduleToCheck->getInstanceName().empty()){
        wxString newName = wxGetTextFromUser(_("The instance name may not be empty!\nEnter a new\
 name"), _("Instance name is empty!"));
        if(newName == wxEmptyString)
            return false;
        *((string *)(instanceNameItem->m_target.data)) = newName.c_str();
    }

    // Check, if there is a module with the same instancename
    TActiveModules::iterator currentModule;
    for(currentModule=m_activeModules.begin();
        currentModule!=m_activeModules.end();
        ++currentModule){

            while(moduleToCheck != (*currentModule) && moduleToCheck->getInstanceName() == (*currentModule)->getInstanceName()){
                wxString newName;
                if(((frame *)wxGetApp().GetTopWindow())->getAutoRename()){
                    wxRegEx renumber("(.*[^0123456789]+)([0123456789]+)");
                    newName = moduleToCheck->getInstanceName().c_str();
                    if(renumber.Matches(moduleToCheck->getInstanceName().c_str())){
                        wxString oldname = renumber.GetMatch(moduleToCheck->getInstanceName().c_str(),1);
                        wxString numberext = renumber.GetMatch(moduleToCheck->getInstanceName().c_str(),2);
                        long tmp=1;
                        numberext.ToLong(&tmp);
                        numberext = wxString::Format("%ld",tmp+1);
                        newName = oldname+numberext;
                    } else {
                        newName = newName+" 2";
                    }
                } else {
                    newName = wxGetTextFromUser(wxString::Format(_("A module with the name %s\
 already exists.\nEnter a new name (or set the \"Auto Rename\" option"),
                        moduleToCheck->getInstanceName().c_str()), _("Instance name exists!"),
                        (*currentModule)->getInstanceName().c_str());
                    if(newName == wxEmptyString)
                        return false;
                }
                *((string *)(instanceNameItem->m_target.data)) = newName.c_str();
            }
        }
        return true;
}


void processor::toggleTimer(bool on)
{
    if(on){
        // No input threads and a timer should be used
        m_frameRateTimer->Start((int) (1000.0/
        (double) wxGetApp().getDocument()->getFramerate()) );
    } else {
        m_frameRateTimer->Stop();
    }
}

void processor::setPause(bool pause)
{
    m_paused = pause;

    if(usingTimer())
        toggleTimer(!pause);
    else if(!pause)
        doFrameTick();
}

bool processor::usingTimer()
{
    return (!m_inputThreadModules.size() && !wxGetApp().getDocument()->getWaitFlag());
}
