/***************************************************************************
                        vlprocessor.h  -  description
                            -------------------
    begin                : Fri Nov 16 2001
    copyright            : (C) 2001 by Fabian Wenzel
    email                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#ifndef _VLPROCESSOR_H_
#define _VLPROCESSOR_H_


#include <vector>
#include <wx/window.h>
#include <wx/thread.h>
#include <wx/timer.h>

#include "vlmodule.h"

namespace vistaLab {

class inputThread;
class channelWindow;
class processor;
class document;

typedef std::vector< module      *> TActiveModules;
typedef std::vector< inputThread *> TActiveInputThreads;


/***************************************************************************/
//! Abstract Base Parameter Change Class
/*!
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class parameterChangeBase {
public:
    virtual void save()                            =0;
    virtual void restore()                         =0;
    virtual ~parameterChangeBase() {}

    configurableObject *getModule()
    {
        return p_targetModule;
    }
protected:
    parameterChangeBase(const configItem *item,
        configurableObject *targetModule) :
    p_item(item),
        p_targetModule(targetModule)
    {
    }

    const configItem   *p_item;
    configurableObject *p_targetModule;
};



/***************************************************************************/
//! Template Parameter Change middle-layer Class for easy target access
/*!
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
template< class T, class BASE > class parameterChange : public BASE{
protected:
    //! Constructor initializes backup with target value (for restoration)
    parameterChange(const configItem *item,
        configurableObject *targetModule) :
    BASE(item,targetModule),
        m_backup(target())
    {
    }

    //! Get target value; Template mechanism simplifies castings
    const T  &target() const
    {
        return *((T *) this->p_item->m_target.data);
    }

    //! Restore original target value from backup
    virtual void restore()
    {
        *((T *) this->p_item->m_target.data)=m_backup;
    }

    //! Store new value in target such that validators can check afterwards
    void store(const T &value)
    {
        *((T *) this->p_item->m_target.data)=value;
    } 

private:
    T m_backup;
};


/***************************************************************************/
//! The processor core processes new slices.
/*!
 *
 *  It is implemented as a thread to be able to run in parallel to
 *  GUI updating (main program) and image I/O (inputThreads). This core
 *  is being created for a processing "session", i.e. when the user starts
 *  processing. When being stopped, the thread is terminated
 *  (and deletes itself as it is a detached thread)
 *
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class processCore: public wxThread {
public:
    processCore(bool wait);
    ExitCode Entry();

    //! Locks the mutex corresponding to the condition.
    /*! This is done whenever a new slice should be processed, either when
     *  running with input threads and all threads have produced some input,
     *  or when the main framework receives a timer event
     */
    static void prepareForProcessing()
    {
        m_processingWaitMutex.Lock();
    }


    //! Calls the condition and unlocks the mutex
    /*! This is done soon after processCore::prepareForProcessing() -
     *  see vistaLabDocument::addFrame() where main buffer shifts take place
     *  between the calls to avoid race conditions
     */
    static void continueProcessing();

    void waitIfNecessary();


    static int getTriggeredFrames(){
        return m_triggeredFramesCounter;
    }

    static void shiftFrameCounters(int offset){
        module::m_processedFramesCounter+=offset;
        m_triggeredFramesCounter+=offset;
    }

private:
    static int          m_triggeredFramesCounter;
    static bool         m_ready;

    //! Indicates whether the process core should wait for a new frame or not
    bool                m_shouldWait;
    bool                m_wait;
    bool                m_waited;

    //! Condition and Mutex for synchronization of new frames
    static wxCondition  m_processingWait;
    static wxMutex      m_processingWaitMutex;
};


/***************************************************************************/
//! The processor starts and stops and handles modules (adding, removing)
/*!
 *
 *  \author  Fabian Wenzel
 *
 ***************************************************************************/
class processor: public wxEvtHandler {
    friend class processCore;
public:
    processor();
    ~processor();

    // Functions for starting/stopping the processor
    void start();
    void stop();

    bool isRunning() const  { return m_running; }
    bool isPaused()  const  { return m_paused;  }
    bool isStopped() const  { return m_stopped; }
    void needsSorting()
    {
        m_needsNewSorting = true;
    }


    // Functions for adding/removing modules
    void addModule     (module *newModule, module *predecessor);
    void swapModules   (module *first , module *second);
    void removeModule  (module *moduleToRemove);
    void removeAll     ();
    bool checkInstanceName(module *moduleToCheck);


    // Interface for the frame which might want to configure modules
    TActiveModules  &getModules ()
    {
        return m_activeModules;
    }

    // Interface for the slice which keeps a counter for received thread data
    int getNumberOfInputThreads()
    {
        return m_inputThreadModules.size();
    }

    // Redirects slider calls to the document and calls all output modules
    void showNeighborSlice(bool downwards);

    void setPause(bool pause);

private:

    // The members (valid threads and agents which may contribute to processing)
    TActiveModules  m_inputThreadModules;
    TActiveModules  m_inputModules;
    TActiveModules  m_processingModules;
    TActiveModules  m_outputModules;
    TActiveModules  m_guiModules;

    // All modules which have been added by the user (but may not contribute)
    TActiveModules      m_activeModules;
    TActiveInputThreads m_activeInputThreads;

    // init / stop - functions for processors
    bool initModules();
    void stopModules();
    void sortModules();
    void toggleTimer(bool on);
    bool usingTimer();

    // Target for timer event when using a global framerate
    void frameTick(wxTimerEvent &e);
    void doFrameTick();

    wxTimer  *m_frameRateTimer;

    processCore *m_processorCore;

    // Misc members
    bool m_running;
    bool m_stopped;
    bool m_paused;
    int  m_frameCounter;
    bool m_needsNewSorting;

    DECLARE_EVENT_TABLE()
};

} // namespace vistaLab

#endif // #ifndef _VLPROCESSOR_H_
