/***************************************************************************
                        vlrepository.cpp  -  description
                            -------------------
    begin                : Fri Nov 23 2001
    copyright            : (C) 2001 by Fabian Wenzel
    email                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif


#include "vlrepository.h"

#include "vlapp.h"


#include <wx/dir.h>
#include <wx/dynlib.h>
#include <wx/intl.h>
#include <wx/msgdlg.h>
#include <algorithm>
#include <boost/checked_delete.hpp>


// BUILTINS


#include "modules/inputstreamimage.h"
#include "modules/inputstreamimagesequence.h"
#include "modules/emptyimagegenerator.h"
#include "modules/imagerenderer.h"
#include "modules/sobelfilter.h"
#include "modules/chronostart.h"
#include "modules/chronostop.h"
#include "modules/rotator.h"
#include "modules/deinterlacer.h"
#include "modules/delayline.h"
#include "modules/rgbtograyconverter.h"
#include "modules/graytorgbconverter.h"
#include "modules/cannyfilter.h"
#include "modules/thresholder.h"
#include "modules/imagemultiplier.h"
#include "modules/differenceimage.h"
#include "modules/imagechannel.h"
#include "modules/logwindow.h"
#include "modules/logfile.h"
#include "modules/grayimagenormalizer.h"
#include "modules/integerscaler.h"
#include "modules/imagesequencewriter.h"
#include "modules/harriscorners.h"
#include "modules/autostopper.h"

using namespace std;
using namespace vistaLab;

typedef module   * (*moduleFunc)();


////////////////////////////// Public functions ////////////////////////////


/***************************************************************************/
/*!
 *  Constructor registers builtins and plugins
 *
 ***************************************************************************/
repository::repository()
{
    registerModules();
}


/***************************************************************************/
/*!
 *  Destructor deletes all items in the repository and closes all plugins
 *
 ***************************************************************************/
repository::~repository()
{
    cleanModules();
}





////////////////////////////// Private functions ///////////////////////////



/***************************************************************************/
/*!
 *  register builtins
 *
 ***************************************************************************/
void repository::registerBuiltins()
{
    // register builtins
    registerModule(new inputStreamImage()         );
    registerModule(new emptyImageGenerator()      );
    registerModule(new inputStreamImageSequence() );
    registerModule(new imageRenderer()            );
    registerModule(new sobelFilter()              );
    registerModule(new rotator()                  );
    registerModule(new deinterlacer()             );
    registerModule(new chronoStart()              );
    registerModule(new chronoStop()               );
    registerModule(new delayLine()                );
    registerModule(new RGBToGrayConverter()       );
    registerModule(new grayToRGBConverter()       );
    registerModule(new cannyFilter()              );
    registerModule(new thresholder()              );
    registerModule(new imageMultiplier()          );
    registerModule(new differenceImage()          );
    registerModule(new imageChannel()             );
    registerModule(new logWindow()                );
    registerModule(new logFile()                  );
    registerModule(new grayImageNormalizer()      );
    registerModule(new integerScaler()            );
    registerModule(new imageSequenceWriter()      );
    registerModule(new harrisCorners()            );
    registerModule(new autoStopper()              );
}


/***************************************************************************/
/*!
 *  register plugins. First, the plugin directory based on the
 *  called program name gets evaluated and then all plugins are
 *  loaded and added
 *
 ***************************************************************************/
void repository::registerPlugins()
{
    int i;

    // First, determine pluginDirectory. The program directory of the
    // installed program is stored in config.h and the program name of the
    // actually running instance has been determined in vistaLabApp::OnInit()
    wxString pluginDirName = wxGetApp().getPluginDir();

    // Now register input streams which are in the subdirectory "inputstreams"
    {
        wxString pluginDir(pluginDirName);
        wxArrayString plugins;
        if(wxDirExists(pluginDir)){
            wxString dllPattern = wxDynamicLibrary::CanonicalizeName("*");
            wxDir::GetAllFiles(pluginDir,&plugins,dllPattern,
                wxDIR_FILES|wxDIR_DIRS);

            for (i=0;i < (int) plugins.GetCount();++i)
                addPlugin(plugins[i]);
        }
    }
}

void repository::registerModules()
{
    registerBuiltins();
    registerPlugins();
    initConfigs();
}

void repository::cleanModules()
{
    // Delete available modules

    TModuleRepository::iterator currentModule;
    for_each(m_availableModules.begin(),m_availableModules.end(),
        boost::checked_deleter<module>());

    // Delete plugins
    for_each (m_plugins.begin(),m_plugins.end(),
        boost::checked_deleter<wxDynamicLibrary>());

    m_availableModules.clear();
    m_plugins.clear();
}

/***************************************************************************/
/*!
 *  Helper functions for adding plugins: Opens a DLL and adds module
 *
 ***************************************************************************/
void repository::addPlugin(const wxString &name)
{
    wxString path, basename, ext;
    wxSplitPath(name, &path, &basename, &ext);
    wxGetApp().addLocaleCatalog(basename);

    wxDynamicLibrary *plugin = new wxDynamicLibrary(name);
    if(plugin->IsLoaded()){
        moduleFunc getModule = (moduleFunc) plugin->GetSymbol("getModule");

        if(getModule != NULL){
            registerModule(getModule());
            m_plugins.push_back(plugin);
        } else {
            delete plugin;
        }
    } else
        delete plugin;
}


/***************************************************************************/
/*!
 *  registers a new module
 *
 ***************************************************************************/
void repository::registerModule( module *newmodule )
{
    m_availableModules.push_back(newmodule);
}


/***************************************************************************/
/*!
 *  Initializes configuration stuff for all modules
 *
 ***************************************************************************/
void repository::initConfigs()
{
    bool success=true;

    TModuleRepository::iterator currentModule;
    for (currentModule = m_availableModules.begin();
        currentModule != m_availableModules.end() && success;
        ++currentModule){
            success = (*currentModule)->_initConfig(true);
            if(!success){
                wxLogError(wxString::Format(_("Caution: \
Could not initialize configuration elements for %s"),
                                              _((*currentModule)->getName().c_str())));
            }
        }
}
